@@ -5,6 +5,7 @@
 
 #include "mined.h"
 #include "charcode.h"
+#include "charprop.h"
 #include "termprop.h"
 
 
@@ -82,6 +83,63 @@ iswhitespace (c)
 	|| c == 0x202F || c == 0x205F;
 }
 
+struct interval {
+    unsigned long first;
+    unsigned long last;
+};
+/*
+static struct interval list_Quotation_Mark [] =
+static struct interval list_Dash [] =
+*/
+#include "typoprop.t"
+
+static
+int
+lookup (ucs, table, length)
+  unsigned long ucs;
+  struct interval * table;
+  int length;
+{
+  int min = 0;
+  int mid;
+  int max = length - 1;
+
+  if (ucs < table [0].first || ucs > table [max].last) {
+	return 0;
+  }
+  while (max >= min) {
+	mid = (min + max) / 2;
+	if (ucs > table [mid].last) {
+		min = mid + 1;
+	} else if (ucs < table [mid].first) {
+		max = mid - 1;
+	} else {
+		return 1;
+	}
+  }
+
+  return 0;
+}
+
+/**
+   Check if character is a quotation mark
+ */
+int
+isquotationmark (unichar)
+  unsigned long unichar;
+{
+  return lookup (unichar, list_Quotation_Mark, arrlen (list_Quotation_Mark));
+}
+
+/**
+   Check if character is a dash
+ */
+int
+isdash (unichar)
+  unsigned long unichar;
+{
+  return lookup (unichar, list_Dash, arrlen (list_Dash));
+}
 
 /**
    Return display indication for a control character.
@@ -459,12 +517,12 @@ struct charmap_table_entry {
 };
 
 static struct charmap_table_entry charmaps_table [] = {
-#ifdef __TURBOC__
+# ifdef __TURBOC__
 	{cp437_table, & cp437_table_len, "CP437", "PC", 'p'},
 	{cp850_table, & cp850_table_len, "CP850", "PL", 'P'},
-#else
-#include "charmaps.t"
-#endif
+# else
+# include "charmaps.t"
+# endif
 };
 
 
@@ -530,8 +588,8 @@ matchprefix (s, m)
 
 /**
    matchpart determines whether its first parameter contains its 
-   second parameter matching approximately as an initial prefix or 
-   as a prefix of any part after a '/' or '>' separator,
+   non-empty second parameter matching approximately as an initial 
+   prefix or as a prefix of any part after a '/' or '>' separator,
    at word boundaries!
    The match ignores separating '-', '_', and space characters, 
    and does not match case.
@@ -545,6 +603,9 @@ matchpart (s, m)
   char * m;
 {
   char * p;
+  if (! * m) {
+	return False;
+  }
   if (matchprefix (s, m)) {
 	return True;
   } else {
@@ -575,18 +636,6 @@ static unsigned int terminal_table_len = 0;
 static struct encoding_table_entry * text_table = (struct encoding_table_entry *) 0;
 static unsigned int text_table_len = 0;
 
-#ifndef use_cjk_tables
-/* define dummy tables referred in code */
-#if defined(__TURBOC__) || defined(vms)
-#define EMPTY 1
-#else
-#define EMPTY 0
-#endif
-static struct encoding_table_entry gb_table [EMPTY];
-static struct encoding_table_entry ejis_table [EMPTY];
-static struct encoding_table_entry sjis_table [EMPTY];
-#endif
-
 
 /**
    Are mapped text and terminal encodings different?
@@ -700,6 +749,7 @@ decode_cjk (entrypoi, map_table)
   struct encoding_table_entry * entrypoi;
   struct encoding_table_entry * map_table;
 {
+#ifdef use_CJKcharmaps
   if (map_table == gb_table) {
 	if ((unsigned int) entrypoi->cjk_ext == 0xFF) {
 		return entrypoi->cjk_base;
@@ -710,12 +760,13 @@ decode_cjk (entrypoi, map_table)
 			| ((((unsigned int) entrypoi->cjk_ext) & 0xF0) << 12)
 			| (((unsigned int) entrypoi->cjk_ext) & 0x0F);
 	}
+  }
+#endif
+
+  if ((unsigned int) entrypoi->cjk_ext >= 0x90) {
+	return 0x8E000000 | (((unsigned int) entrypoi->cjk_ext) << 16) | entrypoi->cjk_base;
   } else {
-	if ((unsigned int) entrypoi->cjk_ext >= 0x90) {
-		return 0x8E000000 | (((unsigned int) entrypoi->cjk_ext) << 16) | entrypoi->cjk_base;
-	} else {
-		return (((unsigned int) entrypoi->cjk_ext) << 16) | entrypoi->cjk_base;
-	}
+	return (((unsigned int) entrypoi->cjk_ext) << 16) | entrypoi->cjk_base;
   }
 }
 #endif
@@ -793,6 +844,9 @@ set_char_encoding (term, charmap, tag)
   char * charmap;
   char tag;
 {
+  if (term) {
+	ascii_screen = False;
+  }
   if (charmap && ! term
 	&& (streq (":16", charmap) || matchpart ("UTF-16BE", charmap))) {
 	utf8_text = True;
@@ -869,6 +923,9 @@ set_char_encoding (term, charmap, tag)
 			utf8_screen = False;
 			utf8_input = False;
 			term_encoding = charmaps_table [i].charmap;
+			if (streq (term_encoding, "ASCII")) {
+				ascii_screen = True;
+			}
 		} else {
 			utf8_text = False;
 			utf16_file = False;
@@ -1128,9 +1185,11 @@ mapped_char (unichar, map_table, map_table_len)
 {
 	unsigned long cjkchar;
 
+#ifdef use_CJKcharmaps
 	if (map_table == gb_table && unichar >= 0x10000) {
 		return unicode_to_gb (unichar);
 	}
+#endif
 
 	cjkchar = unmap_char (unichar, map_table, map_table_len);
 	if (cjkchar != CHAR_INVALID) {
@@ -1216,9 +1275,11 @@ lookup_mapped_char (cjk, map_table, map_table_len)
 {
 	unsigned long unichar;
 
+#ifdef use_CJKcharmaps
 	if (map_table == gb_table && cjk >= 0x90000000) {
 		return gb_to_unicode (cjk);
 	}
+#endif
 
 	unichar = map_char (cjk, map_table, map_table_len);
 	if (! no_unichar (unichar)) {
