@@ -4,6 +4,7 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "charprop.h"
 #include "io.h"	/* for modifier _mask definitions */
 
 
@@ -211,7 +212,7 @@ struct {
 	char * mnemo;
 	int uni;
 } unimnemo3 [] = {
-#include "mnemos.h"
+#include "mnemos.t"
 	{NIL_PTR}
 };
 
@@ -223,7 +224,7 @@ char *
 mnemos (ucs)
   int ucs;
 {
-  static char mnemos [MAX_CHARS];
+  static char mnemos [maxMSGlen];
   char * mpoi = mnemos;
   int i = 0;
   strcpy (mpoi, "");
@@ -231,7 +232,7 @@ mnemos (ucs)
   while (unimnemo3 [i].mnemo != NIL_PTR) {
 	if (unimnemo3 [i].uni == ucs) {
 		int l = strlen (unimnemo3 [i].mnemo);
-		if (mpoi + 1 + l < & mnemos [MAX_CHARS - 1]) {
+		if (mpoi + 1 + l < & mnemos [maxMSGlen - 1]) {
 			strcpy (mpoi, " ");
 			mpoi += 1;
 			strcpy (mpoi, unimnemo3 [i].mnemo);
@@ -301,8 +302,9 @@ compose_mnemo (mnemo, variable_length, return_unicode)
 	return CHAR_UNKNOWN;
   }
 
-  if (return_unicode) {
+  if (return_unicode || utf8_text) {
 	return ret;
+	/* need to guard (above) with ! utf8_text in case of temp. utf8_text */
   } else if (cjk_text || mapped_text) {
 	return encodedchar (ret);
   } else {
@@ -546,7 +548,7 @@ compose_three_patterns (base, pat1a, pat1b, pat2a, pat2b, pat3a, pat3b)
   char * pat3a;
   char * pat3b;
 {
-  char mnemo [MAX_CHARS];
+  char mnemo [maxMSGlen];
   unsigned long composed;
 
   (void) utfencode (base, mnemo);
@@ -661,7 +663,7 @@ compose_base_patterns (base, ps, ps2, ps3)
   struct prefixspec * ps3;
 {
   char * patpoi = ps->pat1;
-  char mnemo [MAX_CHARS];
+  char mnemo [maxMSGlen];
   char * mnemopoi = mnemo;
   unsigned long composed;
 
