@@ -42,28 +42,39 @@
 #include "encoding.h"
 
 #include <string.h>
+#include <stdlib.h>
 
 extern char * getenv ();
 
 char * language_code;
 char * country_code;
+int language_preference;
 
 
 static
 char *
-locale_encoding (var1, var2, var3, var4)
+locale_encoding (var1, var2, var3, var4, var5)
   char * var1;
   char * var2;
   char * var3;
   char * var4;
+  char * var5;
 {
   char * locale = getenv (var1);
+  int varchecked = 1;
   if (! locale || ! * locale) {
 	locale = getenv (var2);
+	varchecked = 2;
 	if (! locale || ! * locale) {
 		locale = getenv (var3);
+		varchecked = 3;
 		if ((! locale || ! * locale) && var4) {
 			locale = getenv (var4);
+			varchecked = 4;
+			if ((! locale || ! * locale) && var5) {
+				locale = getenv (var5);
+				varchecked = 5;
+			}
 		}
 	}
   }
@@ -72,6 +83,11 @@ locale_encoding (var1, var2, var3, var4)
 	char * encoding;
 
 	language_code = locale;
+	if (varchecked <= 2) {
+		language_preference = 1;
+	} else {
+		language_preference = 0;
+	}
 
 	country_code = (char *) strchr (locale, '_');
 	if (! country_code) {
@@ -88,14 +104,15 @@ locale_encoding (var1, var2, var3, var4)
 		if (encoding) {
 			/* "@euro" */
 		} else {
-			/* "_TH" */
-			encoding = country_code;
+			/* "_CC" */
+			encoding = "";
 		}
 	}
 	return encoding;
   } else {
 	language_code = "";
 	country_code = "";
+	language_preference = 0;
 
 	return "";
   }
@@ -104,12 +121,12 @@ locale_encoding (var1, var2, var3, var4)
 char *
 locale_terminal_encoding ()
 {
-  return locale_encoding ("LC_ALL", "LC_CTYPE", "LANG", 0);
+  return locale_encoding ("LC_ALL", "LC_CTYPE", "LANG", 0, 0);
 }
 
 char *
 locale_text_encoding ()
 {
-  return locale_encoding ("TEXTLANG", "LC_ALL", "LC_CTYPE", "LANG");
+  return locale_encoding ("LANGUAGE", "TEXTLANG", "LC_ALL", "LC_CTYPE", "LANG");
 }
 
