@@ -6,15 +6,28 @@
 #include "mined.h"
 
 
+/* disable Han info on VAX (VMS and Ultrix), MSDOS, AIX */
+#if defined (msdos) || defined (vax) || defined (_AIX)
+#define noHANinfo
+#endif
+
+
+#define use_HANinfo
+
+#if defined (noHANinfo) || defined (NOCJK)
+#undef use_HANinfo
+#endif
+
+
 /*======================================================================*\
 |*		Han description table					*|
 \*======================================================================*/
 
 struct raw_hanentry hantable [] = {
-#ifdef use_cjk_tables
+#ifdef use_HANinfo
 #include "handescr.t"
 #else
-#if defined(__TURBOC__) || defined(vms)
+#if defined (__TURBOC__) || defined (vms)
 	{-1}
 #endif
 #endif
