@@ -1,3 +1,10 @@
+/* terminal control mode */
+#ifdef msdos
+#define conio
+#define dosmouse
+#endif
+
+
 /* debug options */
 #define dont_debug_keytrack
 
@@ -11,13 +18,6 @@ extern void do_trace_keytrack _((char * tag, unsigned long c));
 
 #define dont_debug_mouse
 
-#define dont_dynamic_hilite_tracking
-#define dont_debug_hilite_tracking
-
-#ifdef debug_hilite_tracking
-# define debug_mouse
-#endif
-
 #ifdef debug_mouse
 #define trace_mouse(tag)	do_trace_mouse (tag)
 extern void do_trace_mouse _((char * tag));
@@ -30,9 +30,6 @@ extern void do_trace_mouse _((char * tag));
 
 
 /* options */
-extern void mouse_hilite_mode _((FLAG hilite));
-extern void mouse_hilite_tracking_on _((int startx, int starty, int firstrow, int lastrow));
-extern void mouse_hilite_tracking_off _((void));
 extern void mouse_button_event_mode _((FLAG report));
 
 
@@ -56,6 +53,7 @@ extern void getwinsize _((void));
 
 /* output handling */
 extern void flush _((void));
+extern void _putescape _((char * s));
 extern void putescape _((char * s));
 
 /* terminal encoding detection */
@@ -78,7 +76,7 @@ extern void set_font_height _((int r));
 	/* set font height in character pixels, <= 32 */
 extern void set_screen_mode _((int mode));
 extern void maximize_restore_screen _((void));
-extern void font_size _((int inc));
+extern void resize_font _((int inc));
 extern void resize_the_screen _((FLAG sb, FLAG keep_columns));
 extern void switch_textmode_height _((FLAG cycle));
 	/* True: cycle through font heights 8/14/16 
@@ -94,8 +92,6 @@ extern void scroll_reverse _((void));
 extern void add_line _((int));
 extern void delete_line _((int));
 extern void set_cursor _((int, int));
-extern int current_cursor_y;
-#define in_status_line	(current_cursor_y == YMAX)
 
 /* terminal screen attributes handling */
 extern void disp_normal _((void));
@@ -111,12 +107,18 @@ extern void mark_on _((void));
 extern void mark_off _((void));
 extern void bold_on _((void));
 extern void bold_off _((void));
+extern void underline_on _((void));
+extern void underline_off _((void));
+extern void blink_on _((void));
+extern void blink_off _((void));
 extern void unidisp_on _((void));
 extern void unidisp_off _((void));
 extern void ctrldisp_on _((void));
 extern void ctrldisp_off _((void));
 extern void menudisp_on _((void));
 extern void menudisp_off _((void));
+extern void dispXML_attrib _((void));
+extern void dispXML_value _((void));
 extern void dispHTML_code _((void));
 extern void dispHTML_jsp _((void));
 extern void dispHTML_comment _((void));
@@ -165,15 +167,18 @@ extern char * specialansi;	/* Unicode (lineend) marker display */
 extern char * combiningansi;	/* combining character display */
 extern char * menuansi;		/* menu line */
 extern char * HTMLansi;		/* HTML display */
+extern char * XMLattribansi;	/* HTML/XML attribute display */
+extern char * XMLvalueansi;	/* HTML/XML value display */
 extern char * diagansi;		/* dialog (bottom status) line */
 extern char * scrollfgansi;	/* scrollbar foreground */
 extern char * scrollbgansi;	/* scrollbar background */
 
 
 /* input functions */
-extern int get_digits _((int *));
 extern int __readchar _((void));
 extern int __readchar_report_winchg _((void));
+extern int get_digits _((int *));
+extern int get_string_nokeymap _((char *, char *, FLAG, char *));
 
 /* basic keyboard to function mappings */
 extern voidfunc key_map [256];
@@ -207,6 +212,7 @@ extern int char_ready_within _((int msec, char * debug_tag));
 
 /* mouse handling */
 extern void DIRECTcrttool _((void));
+extern void DIRECTvtlocator _((void));
 extern void DIRECTxterm _((void));
 extern void TRACKxterm _((void));
 extern void TRACKxtermT _((void));
@@ -222,18 +228,21 @@ extern int mouse_buttons_pressed;	/* count pressed buttons */
 extern void menu_mouse_mode _((FLAG));
 extern FLAG in_menu_border;
 
+/* switch normal/alternate screen buffer */
+extern void screen_buffer _((FLAG alternate));
+
 
 /* terminal feature indications */
 extern FLAG can_scroll_reverse;
 extern FLAG can_add_line;
 extern FLAG can_delete_line;
 extern FLAG can_clear_eol;
+extern FLAG can_clear_eos;
 extern FLAG can_erase_chars;
 extern FLAG can_reverse_mode;
 extern FLAG can_hide_cursor;
 extern FLAG can_dim;
 extern FLAG can_alt_cset;
-extern FLAG use_mouse_hilite_tracking;
 extern FLAG use_mouse_button_event_tracking;
 
 /* terminal feature usage flags */
@@ -253,13 +262,16 @@ extern FLAG explicit_scrollbar_style;
 extern FLAG explicit_keymap;
 extern FLAG use_script_colour;
 extern FLAG use_mouse;
+extern FLAG use_mouse_release;
 extern FLAG use_mouse_anymove_inmenu;
 extern FLAG use_mouse_anymove_always;
 extern FLAG use_mouse_extended;	/* UTF-8 encoding of mouse coordinates */
+extern FLAG use_mouse_1015;	/* numeric encoding of mouse coordinates */
 extern FLAG use_bold;
 extern FLAG avoid_reverse_colour;
 extern FLAG use_bgcolor;
 extern FLAG use_modifyOtherKeys;
+extern int cursor_style;
 extern void disable_modify_keys _((FLAG dis));
 extern FLAG colours_256;
 extern FLAG colours_88;
@@ -267,14 +279,6 @@ extern FLAG ansi_esc;
 extern FLAG standout_glitch;
 
 
-/* keyboard mapping setup functions */
-extern void toggleKEYMAP _((void));
-extern void setupKEYMAP _((void));
-extern FLAG setKEYMAP _((char *));
-extern int get_string_nokeymap _((char *, char *, FLAG, char *));
-extern char * keyboard_mapping;
-extern char * last_keyboard_mapping;	/* for write_open_pos */
-
 /* keyboard input special results (additional information) */
 extern voidfunc keyproc;	/* function addressed by entered function key */
 
@@ -301,9 +305,32 @@ ctrl	alt	shift	ESC code/ -1	keyshift code
 #define altctrlshift_mask	(alt_mask | ctrl_mask | shift_mask)
 #define applkeypad_mask	0x08
 /* skip 0x30 (may be left in from keyboard.c) */
-#define report_mask	0x40
+#define report_mask	0x40	/* indicate CSI ? sequence */
 
 extern int ansi_params;
 extern int ansi_param [];
+extern char ansi_ini;
 extern char ansi_fini;
 
+
+/* prompt.c: callback for file menu */
+extern int is_directory _((menuitemtype * item));
+extern void check_slow_hint _((void));
+
+/* prompt.c, textfile.c: check file "modes" */
+#ifndef S_ISDIR
+#define S_ISDIR(mode)	(mode & S_IFMT) == S_IFDIR
+#endif
+
+#ifndef S_ISFIFO
+#define S_ISFIFO(mode)	(mode & S_IFMT) == S_IFIFO
+#endif
+
+#ifndef S_ISCHR
+#define S_ISCHR(mode)	(mode & S_IFMT) == S_IFCHR
+#endif
+
+#ifndef S_ISBLK
+#define S_ISBLK(mode)	(mode & S_IFMT) == S_IFBLK
+#endif
+
