@@ -5,6 +5,8 @@
 
 #include "mined.h"
 
+#include "charprop.h"
+
 
 int first_left_margin = 0;
 int next_left_margin = 0;
@@ -34,7 +36,7 @@ int right_margin = 71;
 /*
  * Return string without '\n' for tracing
  */
-static char tracebuf [MAX_CHARS];
+static char tracebuf [maxLINElen];
 
 static
 char *
@@ -173,7 +175,7 @@ try:
 		if (* poi == '\n') {
 			move_address (poi, y);
 			trace_justi ((" moving right beyond newline\n"));
-			MRT ();
+			MOVRT ();
 		} else {
 			advance_char (& poi);
 			move_address (poi, y);
@@ -182,7 +184,7 @@ try:
 				S0 ('\n');
 			} else {
 				trace_justi ((" moving right\n"));
-				MRT ();
+				MOVRT ();
 			}
 			justi_line (next_left_margin, jushop, False, justi_tabs, space_entered, auto_jus);
 		}
@@ -217,12 +219,12 @@ try:
 			poi = cur_text;
 			precede_char (& poi, cur_line->text);
 		}
-		MRT ();
+		MOVRT ();
 	}
   } else {
 	trace_justi ((" moving right finally\n"));
 	move_address (poi, y);
-	MRT ();
+	MOVRT ();
   }
 }
 
@@ -297,7 +299,9 @@ JUSclever ()
   }
   if (lastc != ' ') {
 	/* current line begins a paragraph */
+	FLAG withdigits = False;
 	unsigned long unichar = unicodevalue (poi);
+
 	trace_jc (("skip letters from %d\n", column));
 	if (  unichar == '@'
 	   || (unichar == '/' && * (poi + 1) == '*')
@@ -326,20 +330,21 @@ JUSclever ()
 	{
 		javadoc = True;
 	}
-	while   (  (* poi < '#' && * poi > ' ')
-		|| * poi == '*'
-		|| (* poi >= '/' && * poi <= '9')
-		|| * poi > '@'
+
+	while   (  (* poi >= '/' && * poi <= '9')
+		|| isLetter (unichar)
 		)
 	{
+		if (* poi >= '0' && * poi <= '9') {
+			withdigits = True;
+		}
 		/* skip letters ... */
 		advance_char_scr (& poi, & column, cur_line->text);
+		unichar = unicodevalue (poi);
 	}
+
 	trace_jc (("check marker at %d\n", column));
-	if (javadoc == False && * poi != '.' && * poi != ')') {
-		/* normal text */
-		next_left_margin = first_left_margin;
-	} else {
+	if (javadoc || * poi == ')' || (* poi == '.' && withdigits)) {
 		trace_jc (("skip non-space from %d\n", column));
 		/* numbered text / listing */
 		while (! white_space (* poi)) {
@@ -357,6 +362,9 @@ JUSclever ()
 		if (next_left_margin > right_margin - 2) {
 			right_margin = next_left_margin + 2;
 		}
+	} else {
+		/* normal text */
+		next_left_margin = first_left_margin;
 	}
   }
   trace_jc (("margins %d-%d.\n", first_left_margin, next_left_margin));
@@ -401,7 +409,7 @@ JUSandreturn ()
   move_address (prevline->next->text, find_y (prevline->next));
   for (i = 0; i < prev_offset; i ++) {
 	if (* cur_text == '\n') {
-		MRT ();
+		MOVRT ();
 		/* skip indentation */
 		column = 0;
 		poi = cur_text;
@@ -412,15 +420,15 @@ JUSandreturn ()
 			move_address (poi, y);
 		}
 	}
-	MRT ();
+	MOVRT ();
   }
 #ifdef garble
   if (not_at_lineend && * cur_text == '\n') {
 	printf (" moving right 1 more\n");
-	MRT ();
+	MOVRT ();
 	while (white_space (* cur_text)) {
 		printf (" moving right 1 more\n");
-		MRT ();
+		MOVRT ();
 	}
   }
 #endif
