@@ -3,19 +3,20 @@
 |*		keyboard input						*|
 \*======================================================================*/
 
-#ifdef vms
-#define NOSTRINGTABLES
-#endif
-
-
 #include "mined.h"
+#include "charprop.h"
 #include "io.h"
 #include "termprop.h"
 
 
-#ifndef __TURBOC__
+#if ! defined (__TURBOC__) && ! defined (VAXC)
 #include <sys/time.h>
 #endif
+
+#ifdef vms
+#undef CURSES
+#endif
+
 #ifdef CURSES
 #include <curses.h>
 #include <strings.h>
@@ -28,6 +29,7 @@ static unsigned char fkeyshift;
 #define max_ansi_params 10
 int ansi_params;
 int ansi_param [max_ansi_params];
+char ansi_ini;
 char ansi_fini;
 
 static
@@ -83,7 +85,7 @@ static unsigned long _readchar _((void));
 
 
 /*======================================================================*\
-|*		Handling function key and keyboard mapping tables	*|
+|*		Function key tables					*|
 \*======================================================================*/
 
 #define dont_debug_findkey
@@ -91,6 +93,10 @@ static unsigned long _readchar _((void));
 
 #ifndef msdos
 
+#ifdef unix
+#define TERMINFO
+#endif
+
 static
 void
 dummyfunc ()
@@ -109,7 +115,7 @@ extern struct fkeyentry fkeymap_scoansi [];
 extern struct fkeyentry fkeymap_interix [];
 
 static struct fkeyentry * fkeymap_spec = fkeymap_xterm;
-#ifndef vms
+#ifdef TERMINFO
 extern struct fkeyentry fkeymap_terminfo [];
 #endif
 
@@ -120,6 +126,7 @@ set_fkeymap (term)
 #ifdef debug_findkey
   printf ("set_fkeymap %s\n", term);
 #endif
+
   if (term == NIL_PTR) {
 	fkeymap_spec = fkeymap;
   } else if (* term == 'x') {
@@ -158,192 +165,8 @@ set_fkeymap (term)
 #endif
 
 
-char * keyboard_mapping = "--";
-char * last_keyboard_mapping = "--";
-static char * next_keyboard_mapping = "";
-
-static keymaptabletype keyboard_map = (keymaptabletype) NIL_PTR;
-static keymaptabletype last_keyboard_map = (keymaptabletype) NIL_PTR;
-
-/**
-   Notify user of changed input method
- */
-static
-void
-notify_input_method ()
-{
-  menuitemtype * km = lookup_Keymap_menuitem (keyboard_mapping);
-  char * source = NIL_PTR;
-
-  switch (km->extratag ? km->extratag [0] : ' ') {
-	case 'U':	source = "UnicodeData"; break;
-	case 'H':	source = "Unihan"; break;
-	case 'C':	source = "cxterm"; break;
-	case 'M':	source = "m17n"; break;
-	case 'Y':	source = "yudit"; break;
-	case 'V':	source = "vim"; break;
-	case 'X':	source = "X"; break;
-  }
-  if (source) {
-	build_string (text_buffer, "Input method: %s (source: %s)", 
-			km->itemname, source);
-  } else {
-	build_string (text_buffer, "Input method: %s", km->itemname);
-  }
-  status_uni (text_buffer);
-}
-
-/**
-   Set keymap.
- */
-static
-void
-set_keymap (new_keymap, current, next)
-  keymaptabletype new_keymap;
-  char * current;
-  char * next;
-{
-  last_keyboard_map = keyboard_map;
-  last_keyboard_mapping = keyboard_mapping;
-  keyboard_map = new_keymap;
-  keyboard_mapping = current;
-  next_keyboard_mapping = next;
-
-  if (! loading && ! in_status_line && ! input_active) {
-	notify_input_method ();
-  }
-}
-
-/**
-   Exchange current and previous keyboard mapping; set previous active.
-   With HOP, reset keyboard mapping to none.
- */
-void
-toggleKEYMAP ()
-{
-  if (allow_keymap) {
-	if (hop_flag > 0) {
-		set_keymap ((keymaptabletype) NIL_PTR, "--", keymappingtable [0].shortcut);
-	} else {
-		set_keymap (last_keyboard_map, last_keyboard_mapping, last_keyboard_mapping);
-	}
-	flags_changed = True;
-  } else {
-	error ("Keyboard mapping not active");
-  }
-}
-
-/**
-   Set keyboard mapping.
- */
-static
-FLAG
-selectKEYMAP (script)
-  char * script;
-{
-  if (script == NIL_PTR || * script == '\0') {
-	set_keymap ((keymaptabletype) NIL_PTR, "--", keymappingtable [0].shortcut);
-	flags_changed = True;
-	return True;
-  } else {
-	/* map alternative shortcuts */
-	int i;
-	if (strisprefix ("el", script)) {
-		script = "gr";
-	} else if (strisprefix ("ru", script)) {
-		script = "cy";
-	}
-	/* map shortcut to keymap */
-	for (i = 0; i < keymappingtable_len; i ++) {
-	    if (strisprefix (keymappingtable [i].shortcut, script)) {
-		set_keymap (keymappingtable [i].table,
-			    keymappingtable [i].shortcut,
-			    i + 1 == keymappingtable_len ?
-			      "" : keymappingtable [i + 1].shortcut);
-		flags_changed = True;
-		return True;
-	    }
-	}
-	return False;
-  }
-}
-
-/**
-   Set keyboard mapping.
- */
-FLAG
-setKEYMAP (script)
-  char * script;
-{
-  if (script != NIL_PTR && * script == '-') {
-	/* set keymap as secondary map */
-	script ++;
-	(void) selectKEYMAP (script);
-	(void) selectKEYMAP (NIL_PTR);
-  } else if (script != NIL_PTR) {
-	char * secondary = strchr (script, '-');
-	if (secondary != NIL_PTR) {
-		secondary ++;
-		(void) selectKEYMAP (secondary);
-	}
-	return selectKEYMAP (script);
-  }
-  return True;
-}
-
-#ifdef not_used
-/**
-   Cycle through keyboard mappings.
- */
-void
-cycleKEYMAP ()
-{
-  (void) selectKEYMAP (next_keyboard_mapping);
-}
-#endif
-
-/**
-   Enable keyboard mapping interactively.
-   With HOP, cycle through keyboard mappings.
- */
-void
-setupKEYMAP ()
-{
-  if (allow_keymap) {
-	if (hop_flag > 0) {
-		hop_flag = 0;
-		(void) selectKEYMAP (next_keyboard_mapping);
-	} else {
-		handleKeymapmenu ();
-	}
-  } else {
-	error ("Keyboard mapping not active");
-  }
-}
-
-/**
-   get_string_nokeymap prompts a string like get_string
-   but doesn't apply international keyboard mapping
- */
-int
-get_string_nokeymap (prompt, inbuf, statfl, term_input)
-  char * prompt;
-  char * inbuf;
-  FLAG statfl;
-  char * term_input;
-{
-  int ret;
-
-  dont_suppress_keymap = 0;
-  ret = get_string (prompt, inbuf, statfl, term_input);
-  dont_suppress_keymap = 1;
-
-  return ret;
-}
-
-
 /*======================================================================*\
-|*		Digits input						*|
+|*		String input functions					*|
 \*======================================================================*/
 
 /**
@@ -353,7 +176,7 @@ get_string_nokeymap (prompt, inbuf, statfl, term_input)
    The last character typed in is returned.
    The resulting number is put into the integer the arguments points to.
    It is for use within I/O routines (e.g. reading mouse escape sequences 
-   or cursor position report) and thus calls _readchar ().
+   or cursor position report) and thus calls read1byte / __readchar.
  */
 int
 get_digits (result)
@@ -386,6 +209,27 @@ get_digits (result)
 }
 
 
+/**
+   get_string_nokeymap prompts a string like get_string
+   but doesn't apply international keyboard mapping
+ */
+int
+get_string_nokeymap (prompt, inbuf, statfl, term_input)
+  char * prompt;
+  char * inbuf;
+  FLAG statfl;
+  char * term_input;
+{
+  int ret;
+
+  dont_suppress_keymap = 0;
+  ret = get_string (prompt, inbuf, statfl, term_input);
+  dont_suppress_keymap = 1;
+
+  return ret;
+}
+
+
 /*======================================================================*\
 |*				Terminal input				*|
 \*======================================================================*/
@@ -430,8 +274,7 @@ static
 void
 timestamp_readchar ()
 {
-#ifndef __TURBOC__
-#ifndef VAXC
+#if ! defined (__TURBOC__) && ! defined (VAXC)
   struct timeval now;
   long sec;
   long msec;
@@ -463,7 +306,6 @@ timestamp_readchar ()
 
   average_delta_readchar = total_deltas / n_stamps;
 #endif
-#endif
 }
 
 /*
@@ -705,6 +547,11 @@ readcharacter_mapping (map_keyboard, return_unicode)
 			printf ("rc_m mapped (ret_u %d) -> %04X\n", return_unicode, unichar);
 #endif
 			return unichar;
+		} else if (ascii_screen) {
+			/* skip validity check to allow Latin-1 input 
+			   if ASCII screen is set (reason may be just 
+			   missing display capability)
+			 */
 		} else {
 #ifdef debug_mapped_keyboard
 			printf ("rc_m mapped (ret_u %d) -> %04X -> %04X ...\n", return_unicode, unichar, lookup_mappedtermchar (unichar));
@@ -791,7 +638,7 @@ readcharacter_allbuttons (map_keyboard)
 #endif
   report_release = True;
 
-  c = readcharacter_mapping (map_keyboard, False);
+  c = readcharacter_mapping (map_keyboard, True);
 
   report_winchg = prev_report_winchg;
 
@@ -891,12 +738,44 @@ DIRECTxtermgetpos ()
 		pos &= 0x1F;
 		pos = (pos << 6) | (_readchar_nokeymap () & 0x3F);
 	}
-  } else if (pos == '\0') {
-	pos = 256;
+	if (pos == 0) {
+		pos = 0x800;	/* out of range */
+	}
+  } else if (pos == 0) {
+	pos = 256;	/* out of range */
   }
   return pos - 33;
 }
 
+static
+void
+get_mouse_button (button)
+  unsigned int button;
+{
+  mouse_shift = button & 0x1C;
+  if (button & 32) {
+	mouse_button = movebutton;
+  } else if (button & 64) {
+	button &= 0x03;
+	if (button == 0x00) {
+		mouse_button = wheelup;
+	} else {
+		mouse_button = wheeldown;
+	}
+  } else {
+	button &= 0x03;
+	if (button == 0x00) {
+		mouse_button = leftbutton;
+	} else if (button == 0x01) {
+		mouse_button = middlebutton;
+	} else if (button == 0x02) {
+		mouse_button = rightbutton;
+	} else {
+		mouse_button = releasebutton;
+	}
+  }
+}
+
 static
 void
 DIRECTxtermgetxy (code)
@@ -925,34 +804,12 @@ DIRECTxtermgetxy (code)
 		}
 	}
 
-	if (bidi_screen && button == '$') {
+	if (bidi_screen && mintty_version == 0 && button == '$') {
 		/* fix mlterm quirk for mouse wheel scroll down */
 		button = 'a';
 		button = '#';
 	}
-	button -= 32;
-	mouse_shift = button & 0x1C;
-	if (button & 32) {
-		mouse_button = movebutton;
-	} else if (button & 64) {
-		button &= 0x03;
-		if (button == 0x00) {
-			mouse_button = wheelup;
-		} else {
-			mouse_button = wheeldown;
-		}
-	} else {
-		button &= 0x03;
-		if (button == 0x00) {
-			mouse_button = leftbutton;
-		} else if (button == 0x01) {
-			mouse_button = middlebutton;
-		} else if (button == 0x02) {
-			mouse_button = rightbutton;
-		} else {
-			mouse_button = releasebutton;
-		}
-	}
+	get_mouse_button (button - 32);
   }
 
   fix_mouse_release_event ();
@@ -965,37 +822,42 @@ DIRECTxtermgetxy (code)
   trace_mouse ("DIRECTxtermgetxy");
 #endif
 
-#ifdef unix
+#if defined (unix) || defined (vms)
   if (use_mouse_button_event_tracking && mouse_button == releasebutton) {
 	mouse_button_event_mode (False);
   }
-
-  if (use_mouse_hilite_tracking && mouse_button == leftbutton) {
-	/* handle hilite tracking */
-	if (mouse_ypos < 0 || mouse_ypos > last_y 
-	    || (mouse_xpos == XMAX && disp_scrollbar) 
-	    || is_menu_open ())
-	{
-		/* abort hilite tracking */
-		mouse_hilite_tracking_off ();
-
-#ifdef dynamic_hilite_tracking
-		/* disable hilite tracking mode */
-		mouse_hilite_mode (False);
-#endif
-	} else {
-#ifdef dynamic_hilite_tracking
-		/* enable hilite tracking mode */
-		mouse_hilite_mode (True);
 #endif
+}
 
-		mouse_hilite_tracking_on (mouse_xpos + 1, 
-					  mouse_ypos + 1 + MENU, 
-					  1 + MENU, 
-					  last_y + 3);
-	}
+static
+void
+DIRECTsgrgetxy ()
+{
+  notice_previous_click ();
+
+  if (ansi_fini == 'm') {
+	ansi_param [0] |= 0x3;	/* release button */
   }
-#endif
+  get_mouse_button (ansi_param [0]);
+
+  fix_mouse_release_event ();
+
+  mouse_ypos = ansi_param [2] - 1 - MENU;
+  mouse_xpos = ansi_param [1] - 1;
+}
+
+static
+void
+DIRECTurxvtgetxy ()
+{
+  notice_previous_click ();
+
+  get_mouse_button (ansi_param [0] - 32);
+
+  fix_mouse_release_event ();
+
+  mouse_ypos = ansi_param [2] - 1 - MENU;
+  mouse_xpos = ansi_param [1] - 1;
 }
 
 static
@@ -1006,6 +868,11 @@ DIRECTcrttoolgetxy ()
   int xpos;
   int ypos;
 
+  /* fix crttool settings */
+  use_ascii_graphics = True;
+  menu_border_style = 'r';
+  use_mouse_release = False;
+
   notice_previous_click ();
 
   c = get_digits (& ypos); /* c should be ';' */
@@ -1026,6 +893,32 @@ DIRECTcrttoolgetxy ()
   mouse_xpos = xpos - 1;
 }
 
+static
+void
+DIRECTvtlocatorgetxy ()
+{
+  notice_previous_click ();
+
+  if (ansi_param [0] == 2) {
+	mouse_button = leftbutton;
+  } else if (ansi_param [0] == 4) {
+	mouse_button = middlebutton;
+  } else if (ansi_param [0] == 6) {
+	mouse_button = rightbutton;
+  } else if (ansi_param [0] == 9) {
+	mouse_button = wheelup;
+  } else if (ansi_param [0] == 11) {
+	mouse_button = wheeldown;
+  } else {
+	mouse_button = releasebutton;
+  }
+
+  fix_mouse_release_event ();
+
+  mouse_ypos = ansi_param [2] - 1 - MENU;
+  mouse_xpos = ansi_param [3] - 1;
+}
+
 /*
  * DIRECT... () reads in a direct cursor movement input sequence.
  * Then it either performs a direct function (move/copy/paste) or 
@@ -1069,6 +962,13 @@ DIRECTcrttool ()
   MOUSEfunction ();
 }
 
+void
+DIRECTvtlocator ()
+{
+  DIRECTvtlocatorgetxy ();
+  MOUSEfunction ();
+}
+
 
 #ifdef CURSES
 #ifdef KEY_MOUSE
@@ -1195,8 +1095,24 @@ CURSgetxy ()
  * The queue contents are always terminated by a '\0', so queue can also 
  * be taken as a character string.
  */
-static character queue [MAXCODELEN + 1], * endp = queue;
-static character ctrl_queue [MAXCODELEN + 1], * endcp = ctrl_queue;
+static character queue [MAXCODELEN + 1];
+static character * endp = queue;
+static character ctrl_queue [MAXCODELEN + 1];
+static character * endcp = ctrl_queue;
+static character raw_queue [MAXCODELEN + 1];
+static int rqi = 0;
+
+static
+int
+get1byte ()
+{
+  int c = read1byte ();
+  if (rqi < MAXCODELEN) {
+	raw_queue [rqi ++] = c;
+	raw_queue [rqi] = '\0';
+  }
+  return c;
+}
 
 #ifdef not_used
 static
@@ -1214,12 +1130,20 @@ q_notfull ()
   return endp - queue == MAXCODELEN ? 0 : 1;
 }
 
+static
+void
+raw_q_clear ()
+{
+  rqi = 0;
+}
+
 static
 void
 q_clear ()
 {
   endp = queue;
   endcp = ctrl_queue;
+  rqi = 0;
 }
 
 static
@@ -1292,144 +1216,6 @@ q_get ()
 }
 
 
-/*
-   Look up key sequence in keyboard mapping table.
-   Similar to findkeyin but with modified parameter details.
- * if matchmode == 0 (not used):
- * mapkb (str) >=  0: key sequence str found
- *			* mapped set to mapped string
- *		 == -1: key sequence prefix recognised
- *				(str is prefix of some table entry)
- *		 == -2: key sequence not found
- *				(str is not contained in table)
- * * found == index of exact match if found
- * if matchmode == 1:
- * return -1 if any prefix detected (even if other exact match found)
- * if matchmode == 2:
- * don't return -1, only report exact or no match
- */
-static
-int
-mapkb (str, kbmap, matchmode, found, mapped)
-  char * str;
-  keymaptabletype kbmap;
-  int matchmode;
-  char * * found;
-  char * * mapped;
-{
-#ifdef keymapping_stringtables
-
-  int lastmatch = 0;	/* last index with string matching prefix */
-  register int i;
-  int ret = -2;
-  int len = strlen (str);
-
-  * found = NIL_PTR;
-
-  if (* kbmap == '\0') {
-	return ret;
-  }
-
-  i = lastmatch;
-  do {
-	if (strncmp (str, kbmap, len) == 0) {
-		/* str is prefix of current entry */
-		lastmatch = i;
-		if (len == strlen (kbmap)) {
-			/* str is equal to current entry */
-			* found = kbmap;
-			* mapped = kbmap + len + 1;
-			if (matchmode == 1) {
-				/* return index unless other prefix 
-				   was or will be found */
-				if (ret == -2) {
-					ret = i;
-				}
-			} else {
-				return i;
-			}
-		} else {
-			/* str is partial prefix of current entry */
-			if (matchmode == 1) {
-				/* return -1 but continue to search 
-				   for exact match */
-				ret = -1;
-			} else if (matchmode != 2) {
-				return -1;
-			} else {
-			}
-		}
-	}
-	++ i;
-	kbmap += strlen (kbmap) + 1;
-	kbmap += strlen (kbmap) + 1;
-	if (* kbmap == '\0') {
-		i = 0;
-		return ret;
-	}
-  } while (i != lastmatch);
-
-  return ret;
-
-#else
-
-  int lastmatch = 0;	/* last index with string matching prefix */
-  register int i;
-  int ret = -2;
-
-  * found = NIL_PTR;
-
-  if (kbmap [0].fk == NIL_PTR) {
-	return ret;
-  }
-
-  i = lastmatch;
-  do {
-	if (strncmp (str, kbmap [i].fk, strlen (str)) == 0) {
-		/* str is prefix of current entry */
-		lastmatch = i;
-		if (strlen (str) == strlen (kbmap [i].fk)) {
-			/* str is equal to current entry */
-			* found = kbmap [i].fk;
-			* mapped = kbmap [i].fp;
-			if (! * mapped) {
-				/* fix missing mapping in keyboard mapping table */
-				* mapped = "";
-			}
-			if (matchmode == 1) {
-				/* return index unless other prefix 
-				   was or will be found */
-				if (ret == -2) {
-					ret = i;
-				}
-			} else {
-				return i;
-			}
-		} else {
-			/* str is partial prefix of current entry */
-			if (matchmode == 1) {
-				/* return -1 but continue to search 
-				   for exact match */
-				ret = -1;
-			} else if (matchmode != 2) {
-				return -1;
-			} else {
-			}
-		}
-	}
-	++ i;
-	if (kbmap [i].fk == NIL_PTR) {
-		i = 0;
-		return ret;
-	}
-  } while (i != lastmatch);
-
-  return ret;
-
-#endif
-}
-
-
 #ifndef msdos
 
 /*
@@ -1551,7 +1337,7 @@ findkey (str)
 	return res;
   }
 
-#ifndef vms
+#ifdef TERMINFO
   res = findkeyin (str, fkeymap_terminfo, 0, & dummy);
 # ifdef debug_findkey
   printf ("<^[%s> _terminfo [%d]\n", str + 1, res);
@@ -1577,7 +1363,7 @@ static
 void
 showfkey ()
 {
-  build_string (text_buffer, "Curses function key entered: %03X / %04o", key, key);
+  build_string (text_buffer, "Curses function key entered: %03X / %04o", (int) key, (int) key);
   status_msg (text_buffer);
 }
 
@@ -1648,32 +1434,32 @@ struct {
 	{KEY_DC, DELkey},
 	{KEY_HOME, HOMEkey},
 	{KEY_END, ENDkey},
-	{KEY_PPAGE, PU},
-	{KEY_NPAGE, PD},
-	{KEY_UP, MUP},
-	{KEY_DOWN, MDN},
-	{KEY_LEFT, MLF},
-	{KEY_RIGHT, MRT},
+	{KEY_PPAGE, PGUPkey},
+	{KEY_NPAGE, PGDNkey},
+	{KEY_UP, UPkey},
+	{KEY_DOWN, DNkey},
+	{KEY_LEFT, LFkey},
+	{KEY_RIGHT, RTkey},
 	{KEY_B2, HOP},
 	{KEY_A1, HOMEkey},
-	{KEY_A3, PU},
+	{KEY_A3, PGUPkey},
 	{KEY_C1, ENDkey},
-	{KEY_C3, PD},
+	{KEY_C3, PGDNkey},
 #endif
 #ifdef KEY_A2
-	{KEY_A2, MUP},
+	{KEY_A2, UPkey},
 #endif
 #ifdef KEY_C2
-	{KEY_C2, MDN},
+	{KEY_C2, DNkey},
 #endif
 #ifdef KEY_B1
-	{KEY_B1, MLF},
+	{KEY_B1, LFkey},
 #endif
 #ifdef KEY_B3
-	{KEY_B3, MRT},
+	{KEY_B3, RTkey},
 #endif
-	{0x1BF, BLINE},
-	{0x1C0, ELINE},
+	{0x1BF, BLINEORUP},
+	{0x1C0, ELINEORDN},
 
 #ifdef PAD0
 	{PAD0, PASTE},
@@ -1744,6 +1530,16 @@ _readchar_nokeymap ()
 }
 
 
+#define dont_debug_readbyte
+
+#define dont_debug_terminal_response
+#ifdef debug_terminal_response
+#define trace_terminal_response(what)	if (debug_tag) printf ("[%s (%d)] %s\n", debug_tag, msec, what);
+#else
+#define trace_terminal_response(what)
+#endif
+
+
 static int choose_char _((char *));
 
 static character rest_queue [MAXCODELEN + 1];
@@ -1755,28 +1551,30 @@ putback_rest (rest)
   char * rest;
 {
   char old_rest_queue [MAXCODELEN + 1];
+#ifdef debug_readbyte
+  printf ("putback_rest <%s>\n", rest);
+#endif
 
-	if (! streq (rest, " ")) {
-		if (have_rest_queue) {
-			/**
-			   overdraft characters to be mapped are 
-			   put back into rest_queue 
-			   to be considered for subsequent mapping
-				-> Hiragana nihongo
-				-> Hiragana nyi
-				-> test abcde
-			 */
-			strcpy (old_rest_queue, (char *) rest_queue);
-		} else {
-			old_rest_queue [0] = '\0';
-		}
-		strcpy ((char *) rest_queue, rest);
-		strcat ((char *) rest_queue, old_rest_queue);
-		have_rest_queue = True;
+  if (! streq (rest, " ")) {
+	if (have_rest_queue) {
+		/**
+		   overdraft characters to be mapped are 
+		   put back into rest_queue 
+		   to be considered for subsequent mapping
+			-> Hiragana nihongo
+			-> Hiragana nyi
+			-> test abcde
+		 */
+		strcpy (old_rest_queue, (char *) rest_queue);
+	} else {
+		old_rest_queue [0] = '\0';
 	}
+	strcpy ((char *) rest_queue, rest);
+	strcat ((char *) rest_queue, old_rest_queue);
+	have_rest_queue = True;
+  }
 }
 
-
 /* read 1 byte from queue; for curses, however, function 
    key indications must be passed on which are longer than 8 bits
  */
@@ -1794,29 +1592,29 @@ read1byte ()
 	if (* rest_queue == '\0') {
 		have_rest_queue = False;
 	}
+#ifdef debug_readbyte
+	printf ("read byte %02X from rest queue\n", ch);
+#endif
 	return ch;
-  } else
-  {
+  } else {
+	int ch;
 #ifdef CURSES
-	return curs_readchar ();
+	ch = curs_readchar ();
 #else
 	if (report_winchg) {
-		return __readchar_report_winchg ();
+		ch = __readchar_report_winchg ();
 	} else {
-		return __readchar ();
+		ch = __readchar ();
 	}
 #endif
+#ifdef debug_readbyte
+	printf ("read byte %02X from io\n", ch);
+#endif
+	return ch;
   }
 }
 
 
-#define dont_debug_terminal_response
-#ifdef debug_terminal_response
-#define trace_terminal_response(what)	if (debug_tag) printf ("[%s (%d)] %s\n", debug_tag, msec, what);
-#else
-#define trace_terminal_response(what)
-#endif
-
 static
 int
 bytereadyafter (msec, debug_tag)
@@ -1955,8 +1753,12 @@ _readchar ()
 	return q_get ();
   }
   queue_mode_mapped = False;
+  raw_q_clear ();
 
   keyshift = 0;
+#ifdef msdos
+  mouse_shift = 0;
+#endif
 
 
 #ifdef radical_stroke
@@ -1979,7 +1781,7 @@ _readchar ()
 	}
 
 	/* adjust menu position if on status line */
-	if (in_status_line) {
+	if (input_active) {
 		choice_line = YMAX - arrlen (radicalsmenu) - 2;
 		if (choice_line < 0) {
 			choice_line = 0;
@@ -2016,7 +1818,7 @@ _readchar ()
 				* found ++ = * mapped ++;
 			}
 			* found = '\0';
-			res = mapkb (radicalnumber, keyboard_map, 2, & found, & mapped);
+			res = map_key (radicalnumber, 2, & found, & mapped);
 			if (res >= 0) {
 				/* first goto in my life and then 
 				   such an exceptionally ugly one -
@@ -2084,7 +1886,7 @@ _readchar ()
 	if (! keyproc) {
 		keyproc = I;
 	}
-	keyshift = pc_xkey_map [ch - 0x100].fkeyshift;
+	keyshift |= pc_xkey_map [ch - 0x100].fkeyshift;
 	return FUNcmd;
   }
 
@@ -2098,13 +1900,12 @@ _readchar ()
 	if (! keyproc) {
 		keyproc = I;
 	}
-	keyshift = pc_xkey_map [ch].fkeyshift;
+	keyshift |= pc_xkey_map [ch].fkeyshift;
 # ifdef debug_whatever
+	/* VAXC needs the identity cast (voidfunc) */
 	if ((voidfunc) keyproc == (voidfunc) I) {
 		return ch;
 	}
-	/* (voidfunc) is an identity cast here. It seems to be required 
-	   for the sake of the apparently totally rotten microvax compiler */
 # endif
 	return FUNcmd;
   } else {
@@ -2116,16 +1917,28 @@ _readchar ()
   trace_keytrack ("_readchar<033", ch);
   if (ch == '\033') {
 	char second_esc_char = '\0';
+	int esc_wait = 2 * escape_delay;
 
 	q_put (ch);
 
-	while  ((res = findkey (ctrl_queue)) == -1 /* prefix of table entry */
+	if (! bytereadyafter (esc_wait, NIL_PTR)) {
+		return q_get ();
+	}
+
+	while ((res = findkey (ctrl_queue)) == -1 /* prefix of table entry */
 		&& q_notfull ()
+#ifdef simple_escape_delay_handling
 		&& bytereadyafter (escape_delay, NIL_PTR)
-	       )
+#endif
+	      )
 	{
 		char prev_ch = ch;
-		ch = read1byte ();
+
+		if (prev_ch != '\033' && ! input_active && ! bytereadyafter (escape_delay, NIL_PTR)) {
+			status_uni ("...awaiting slow key code sequence...");
+		}
+
+		ch = get1byte ();
 #ifdef debug_kbesc
 		printf ("<^[%s> + [%dms] %02X\n", ctrl_queue + 1, escape_delay, ch);
 #endif
@@ -2133,7 +1946,7 @@ _readchar ()
 		/* accept ESC prefix as Alt modifier */
 		if (detect_esc_alt && ch == '\033' && prev_ch == '\033') {
 			if (bytereadyafter (escape_delay, NIL_PTR)) {
-				ch = read1byte ();
+				ch = get1byte ();
 				keyshift |= alt_mask;
 #ifdef debug_kbesc
 				printf ("-> <^[%s> + ^[/%02X\n", ctrl_queue + 1, ch);
@@ -2149,7 +1962,7 @@ _readchar ()
 			if (ch == '?') {
 				keyshift |= report_mask;
 			}
-			ch = read1byte ();
+			ch = get1byte ();
 #ifdef debug_kbesc
 			printf ("-> <^[%s> + >?/%02X\n", ctrl_queue + 1, ch);
 #endif
@@ -2171,13 +1984,13 @@ check_ctrl_byte:
 		   applkeypad_mask			0x08
 		 */
 		if (ch == ';' && second_esc_char == '[') {
-			ch = read1byte ();
+			ch = get1byte ();
 			if (ch >= '2' && ch <= '9') {
 				/* ESC [1;2A → ESC [1A + shift etc */
 				/* special: map '9' to applkeypad_mask */
 				keyshift |= (ch & ~'0') - 1;
 			} else if (ch == '1') {
-				ch = read1byte ();
+				ch = get1byte ();
 				if (ch >= '0' && ch <= '6') {
 					/* xterm with alwaysUseMods: */
 					/* ESC [1;11A → ESC [1A + alt etc */
@@ -2191,12 +2004,12 @@ check_ctrl_byte:
 				q_put (';');
 				q_put (ch);
 			}
-			ch = read1byte ();
+			ch = get1byte ();
 #ifdef debug_kbesc
 			printf ("-> <^[%s> + 1-8/%02X\n", ctrl_queue + 1, ch);
 #endif
 		} else if (ch == ';' && second_esc_char == '1') {
-			ch = read1byte ();
+			ch = get1byte ();
 			if (ch >= '2' && ch <= '8') {
 				/* xterm/VT52 mode: ESC 1;2A → ESC 1A + shift etc */
 				keyshift |= (ch & ~'0') - 1;
@@ -2204,7 +2017,7 @@ check_ctrl_byte:
 				q_put (';');
 				q_put (ch);
 			}
-			ch = read1byte ();
+			ch = get1byte ();
 #ifdef debug_kbesc
 			printf ("-> <^[%s> + ;/%02X\n", ctrl_queue + 1, ch);
 #endif
@@ -2216,15 +2029,16 @@ check_ctrl_byte:
 				/* HP: ESC 2p → ESC p + shift etc */
 				/* special: map '9' to applkeypad_mask */
 				keyshift |= (ch & ~'0') - 1;
-			} else if (ch == '1') {
-				ch = read1byte ();
+			} else if (ch == '1' || (ch == '0' && mlterm_version)) {
+				ch = get1byte ();
 				if (ch >= '0' && ch <= '6') {
 					/* ESC O11P → ESC OP + alt etc */
 					/* HP: ESC 11p → ESC p + alt etc */
 					keyshift |= applkeypad_mask | ((ch & ~'0') + 1);
 				} else if (ch == ';') {
 					/* mlterm: ESC O1;2A → ESC OA + shift etc */
-					ch = read1byte ();
+					/* mlterm: ESC O0;3w → ESC O0w + alt etc */
+					ch = get1byte ();
 					keyshift |= ((ch & ~'0') - 1);
 				} else {
 					q_put ('1');
@@ -2233,7 +2047,7 @@ check_ctrl_byte:
 			} else {
 				q_put (ch);
 			}
-			ch = read1byte ();
+			ch = get1byte ();
 #ifdef debug_kbesc
 			printf ("-> <^[%s> + 0-9/%02X\n", ctrl_queue + 1, ch);
 #endif
@@ -2245,7 +2059,7 @@ check_ctrl_byte:
 				/* SCO: ESC [2M → ESC [M + shift etc */
 				/* special: map '9' to applkeypad_mask */
 				newkeyshift = (ch & ~'0') - 1;
-				newch = read1byte ();
+				newch = get1byte ();
 #ifdef debug_kbesc
 				printf ("-> <^[%s> + 2-8/%02X\n", ctrl_queue + 1, newch);
 #endif
@@ -2255,14 +2069,14 @@ check_ctrl_byte:
 					q_put (ch);
 				}
 			} else if (ch == '1') {
-				ch = read1byte ();
+				ch = get1byte ();
 #ifdef debug_kbesc
 				printf ("-> <^[%s> + 1/%02X\n", ctrl_queue + 1, ch);
 #endif
 				if (ch >= '0' && ch <= '6') {
 					/* SCO: ESC [11M → ESC [M + alt etc */
 					newkeyshift = applkeypad_mask | ((ch & ~'0') + 1);
-					newch = read1byte ();
+					newch = get1byte ();
 					if (newch >= 'A' && newch <= 'Z') {
 						keyshift |= newkeyshift;
 					} else {
@@ -2271,20 +2085,20 @@ check_ctrl_byte:
 					}
 				} else if (ch == ';') {
 					/* xterm SCO: ESC [1;5R → ESC [R + alt etc */
-					ch = read1byte ();
+					ch = get1byte ();
 					keyshift |= (ch & ~'0') - 1;
-					newch = read1byte ();
+					newch = get1byte ();
 				} else {
 					q_put ('1');
 					q_put (ch);
-					newch = read1byte ();
+					newch = get1byte ();
 				}
 			} else {
 #ifdef debug_kbesc
 				printf ("-> <^[%s>\n", ctrl_queue + 1);
 #endif
 				q_put (ch);
-				newch = read1byte ();
+				newch = get1byte ();
 			}
 			ch = newch;
 			if (ch == ';') {
@@ -2346,18 +2160,39 @@ check_ctrl_byte:
 	if (quit) {
 		keyproc = I;
 		return FUNcmd;
-	} else if (res < 0) { /* key pattern not found in fkeymap table */
+	} else if (res < 0) {	/* key pattern not found in fkeymap table */
 		if (second_esc_char == '[') {
 			/* generic scanning for unknown ANSI sequence */
+#ifdef debug_kbansi
+			character q1, q2;
+			q1 = q_get (); /* swallow ESC */
+			q2 = q_get (); /* swallow [ */
+			printf ("swallowing %X %X\n", q1, q2);
+#else
 			(void) q_get (); /* swallow ESC */
 			(void) q_get (); /* swallow [ */
-			putback_rest (ctrl_queue + 2);
+#endif
+			/* (Was putback_rest (ctrl_queue + 2);) */
+			/* Use raw_queue to ignore previous modifier 
+			   transformations like ESC [1;11A → ESC [1A + alt */
+			if ((keyshift & alt_mask) && raw_queue [1] == '[') {
+				/* consider double-ESC alt (rxvt) */
+				putback_rest (raw_queue + 2);
+			} else {
+				putback_rest (raw_queue + 1);
+			}
 
 			ansi_params = 0;
+			ansi_ini = '\0';
 			do {
-				int i;
+				int i = -1;
 				ansi_fini = get_digits (& i);
-				if (ansi_params < max_ansi_params) {
+				if (ansi_params == 0 && i == -1 &&
+				   (ansi_fini == '>' || ansi_fini == '?' || ansi_fini == '<')
+				   ) {
+					ansi_ini = ansi_fini;
+					ansi_fini = ';';
+				} else if (ansi_params < max_ansi_params) {
 					ansi_param [ansi_params] = i;
 					ansi_params ++;
 				}
@@ -2490,7 +2325,25 @@ check_ctrl_byte:
 					/* fall through to keyboard mapping */
 				}
 			} else {	/* not CSI 27 sequence */
-				keyproc = ANSIseq;
+				if (ansi_fini == '&') {	/* DEC locator report */
+					(void) read1byte ();	/* swallow 'w' */
+					if (ansi_params < 4) {
+						ansi_param [3] = 1;
+						if (ansi_params < 3) {
+							ansi_param [2] = 1;
+						}
+					}
+					DIRECTvtlocatorgetxy ();
+					keyproc = MOUSEfunction;
+				} else if (ansi_ini == '<') {
+					DIRECTsgrgetxy ();
+					keyproc = MOUSEfunction;
+				} else if (! ansi_ini && ansi_fini == 'M') {
+					DIRECTurxvtgetxy ();
+					keyproc = MOUSEfunction;
+				} else {
+					keyproc = ANSIseq;
+				}
 				return FUNcmd;
 			}
 		} else if (keyshift & alt_mask) {
@@ -2539,10 +2392,9 @@ check_ctrl_byte:
 
 #endif	/* #else msdos */
 
-  if (allow_keymap && dont_suppress_keymap
-		&& keyboard_map != (keymaptabletype) NIL_PTR) {
+  if (allow_keymap && dont_suppress_keymap && keyboard_mapping_active ()) {
 	disable_modify_keys (True);
-	while  ((res = mapkb (ctrl_queue, keyboard_map, 1, & found, & mapped)) == -1
+	while  ((res = map_key (ctrl_queue, 1, & found, & mapped)) == -1
 		 /* prefix of some table entry */
 		&& q_notfull ()
 		&& bytereadyafter (keymap_delay, NIL_PTR)
@@ -2583,7 +2435,7 @@ check_ctrl_byte:
 			putback_rest (ctrl_queue + strlen (prev_found));
 		}
 	} else if (res == -1) {
-		res = mapkb (ctrl_queue, keyboard_map, 2, & found, & mapped);
+		res = map_key (ctrl_queue, 2, & found, & mapped);
 #ifdef debug_kbmap
 		printf ("find again %s, found %s, ", ctrl_queue, found);
 #endif
@@ -2647,7 +2499,20 @@ handle_mapped:
 	} else {
 		ring_bell ();
 		flush ();	/* clear key mapping menu */
-		return _readchar ();
+#define return_after_esc
+#ifdef return_after_esc
+		return CHAR_UNKNOWN;	/* actually "CHAR_NONE" */
+#else
+		if (is_menu_open () /* doesn't seem to work */) {
+			return CHAR_UNKNOWN;	/* actually "CHAR_NONE" */
+		} else {
+			/* read another character */
+			/* prompt should have been redrawn;
+			   menu (file chooser) would stay blank...
+			 */
+			return _readchar ();
+		}
+#endif
 	}
   } else {	/* not keyboard mapped */
 	return q_get ();
@@ -2715,7 +2580,7 @@ choose_char (choices)
 	}
 
 	/* adjust menu position if on status line */
-	if (in_status_line) {
+	if (input_active) {
 		choice_line = YMAX - lastrow - 3;
 		if (choice_line < 0) {
 			choice_line = 0;
@@ -2734,6 +2599,10 @@ choose_char (choices)
 		choicelinepoi = thischoiceline;
 		if (li == lastrow) {
 			lastcol = (choice_count - 1) % 10;
+			if (lastcol < 0) {
+				/* should not occur as choice_count > 0 */
+				lastcol = 0;
+			}
 		} else {
 			lastcol = 9;
 		}
@@ -2770,7 +2639,7 @@ choose_char (choices)
 	if (lastrow >= YMAX - 2) {
 		lastrow = YMAX - 3;
 		reduced_menu = True;
-		if (! in_status_line) {
+		if (! input_active) {
 			status_msg ("Warning: full menu too large to display");
 		}
 	}
@@ -2778,8 +2647,8 @@ choose_char (choices)
 	selected = popup_menu (choicemenu, lastrow + 1, 
 				choice_col, choice_line, 
 				"Choose character", False, False, "1234567890");
-	if (in_status_line) {
-		redraw_prompt ();
+	if (input_active) {
+		/* redraw_prompt () now checked and called in input () */
 #ifdef cut_picklist_to_screen
 	} else if (reduced_menu) {
 		clear_status ();
