@@ -18,13 +18,13 @@
 voidfunc key_map [256] = {
 	/* characters to functions map */
   /* ^@ ^A ^B ^C ^D ^E ^F ^G */
-		MARK, MPW, DCC, PD, MRT, MUP, MNW, GOHOP,
+		MARK, MPW, DCC, PGDNkey, RTkey, UPkey, MNW, GOHOP,
   /* ^H ^I ^J ^K ^L ^M ^N ^O */
-		DPC, Sch, Sch, DLN, HOP, SNL, RS, LIB,
+		DPC, Sch, Sch, DLN, HOP, SNL, RES, APPNL,
   /* ^P ^Q ^R ^S ^T ^U ^V ^W */
-		PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
+		PASTEstay, HOP, PGUPkey, LFkey, DNW, CUT, CTRLINS, SU,
   /* ^X ^Y ^Z ^[ ^\ ^] ^^ ^_ */
-		MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
+		DNkey, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
   /* 20-2F 040-057 */	Sch, Sch, Sdoublequote, Sch, Sch, Sch, Sch, Ssinglequote,
 		Sch, Sch, Sch, Sch, Sch, Sdash, Sch, Sch,
   /* 30-3F 060-077 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
@@ -45,33 +45,33 @@ voidfunc key_map [256] = {
 
 voidfunc mined_key_map [32] = {
 	/* control characters to functions map */
-  /* ^@-^G */	MARK, MPW, DCC, PD, MRT, MUP, MNW, GOHOP,
-  /* ^H-^O */	DPC, Sch, Sch, DLN, HOP, SNL, RS, LIB,
-  /* ^P-^W */	PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
-  /* ^X-^_ */	MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
+  /* ^@-^G */	MARK, MPW, DCC, PGDNkey, RTkey, UPkey, MNW, GOHOP,
+  /* ^H-^O */	DPC, Sch, Sch, DLN, HOP, SNL, RES, APPNL,
+  /* ^P-^W */	PASTEstay, HOP, PGUPkey, LFkey, DNW, CUT, CTRLINS, SU,
+  /* ^X-^_ */	DNkey, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
 };
 
 voidfunc ws_key_map [32] = {
 	/* control characters to functions map */
-  /* ^@-^G */	I, MPW, JUS, PD, MRT, MUP, MNW, DCC,
-  /* ^H-^O */	MLF, Sch, Sch, ctrlK, RS, SNL, LIB, ctrlO,
-  /* ^P-^W */	CTRLINS, ctrlQ, PU, MLF, DNW, (voidfunc) BADch, TOGINS, SU,
-  /* ^X-^_ */	MDN, DLINE, SD, ESCAPE, CANCEL, (voidfunc) BADch, (voidfunc) BADch, (voidfunc) BADch,
+  /* ^@-^G */	I, MPW, JUS, PGDNkey, RTkey, UPkey, MNW, DCC,
+  /* ^H-^O */	LFkey, Sch, Sch, ctrlK, RES, SNL, APPNL, ctrlO,
+  /* ^P-^W */	CTRLINS, ctrlQ, PGUPkey, LFkey, DNW, (voidfunc) BADch, TOGINS, SU,
+  /* ^X-^_ */	DNkey, DLINE, SD, ESCAPE, CANCEL, (voidfunc) BADch, (voidfunc) BADch, (voidfunc) BADch,
 };
 
 voidfunc pico_key_map [32] = {
 	/* control characters to functions map */
-  /* ^@-^G */	MNW, GO_LINEBEGIN , MLF, FS, DCC, GO_LINEEND , MRT, HELP,
-  /* ^H-^O */	DPC, Sch, JUS, DLN, RDwin, SNL, MDN, WTU,
-  /* ^P-^W */	MUP, I, INSFILE, I, SPELL, PASTE, PD, SFW,
-  /* ^X-^_ */	QUED, PU, I, ESCAPE, CANCEL, MARK, toggleMARK, CTRLINS,
+  /* ^@-^G */	MNW, MOVLBEG , LFkey, FS, DCC, MOVLEND , RTkey, HELP,
+  /* ^H-^O */	DPC, Sch, JUS, DLN, RDwin, SNL, DNkey, WTU,
+  /* ^P-^W */	UPkey, I, INSFILE, I, SPELL, PASTE, PGDNkey, SFW,
+  /* ^X-^_ */	QUED, PGUPkey, I, ESCAPE, CANCEL, MARK, toggleMARK, CTRLINS,
 };
 
 voidfunc emacs_key_map [32] = {
 	/* control characters to functions map */
-  /* ^@-^G */	MARK, GO_LINEBEGIN , MLF, I, DCC, GO_LINEEND , MRT, CANCEL,
-  /* ^H-^O */	HELP, Sch, Sch, DLN, RD, SNL, MDN, LIB,
-  /* ^P-^W */	MUP, CTRLINS, SRV, SFW, I, (voidfunc) REPT, PD, CUT,
+  /* ^@-^G */	MARK, MOVLBEG , LFkey, I, DCC, MOVLEND , RTkey, CANCEL,
+  /* ^H-^O */	HELP, Sch, Sch, DLN, RD, SNL, DNkey, APPNL,
+  /* ^P-^W */	UPkey, CTRLINS, SRV, SFW, I, (voidfunc) REPT, PGDNkey, CUT,
   /* ^X-^_ */	EMAX, PASTE, SUSP, META, HOP, I, HOP, UNDO,
 };
 
@@ -79,7 +79,7 @@ voidfunc windows_key_map [32] = {
 	/* control characters to functions map */
   /* ^@-^G */	MARK, I, I, COPY, I, I, SFW, GOHOP,
   /* ^H-^O */	REPL, Sch, Sch, I, I, SNL, I, EDIT,
-  /* ^P-^W */	PRINT, QUED, I, WT, I, I, PASTE, I,
+  /* ^P-^W */	PRINT, QUED, I, WT, I, I, PASTE, CLOSEFILE,
   /* ^X-^_ */	CUT, I, I, ESCAPE, CANCEL, I, I, CTRLINS,
 };
 
@@ -152,16 +152,16 @@ struct pc_fkeyentry pc_xkey_map [512] = {
 /* #define K_F10                  0x144 */	{F10},
 					{I}, {I},
 /* #define K_Home                 0x147 */	{HOMEkey},
-/* #define K_Up                   0x148 */	{MUP},
-/* #define K_PageUp               0x149 */	{PU},
+/* #define K_Up                   0x148 */	{UPkey},
+/* #define K_PageUp               0x149 */	{PGUPkey},
 /* : #define K_Alt_KPMinus          0x14a */	{screenlesslines, alt_mask},
-/* #define K_Left                 0x14b */	{MLF},
+/* #define K_Left                 0x14b */	{LFkey},
 /* #define K_Center               0x14c */	{HOP},
-/* #define K_Right                0x14d */	{MRT},
+/* #define K_Right                0x14d */	{RTkey},
 /* : #define K_Alt_KPPlus           0x14e */	{screenmorelines, alt_mask},
 /* #define K_End                  0x14f */	{ENDkey},
-/* #define K_Down                 0x150 */	{MDN},
-/* #define K_PageDown             0x151 */	{PD},
+/* #define K_Down                 0x150 */	{DNkey},
+/* #define K_PageDown             0x151 */	{PGDNkey},
 /* #define K_Insert               0x152 */	{INSkey},
 /* #define K_Delete               0x153 */	{DELkey},
 /* #define K_Shift_F1             0x154 */	{F1, shift_mask},
@@ -195,10 +195,10 @@ struct pc_fkeyentry pc_xkey_map [512] = {
 /* #define K_Alt_F9               0x170 */	{F9, alt_mask},
 /* #define K_Alt_F10              0x171 */	{F10, alt_mask},
 /* #define K_Control_Print        0x172 */	{I},
-/* #define K_Control_Left         0x173 */	{MLF, ctrl_mask},
-/* #define K_Control_Right        0x174 */	{MRT, ctrl_mask},
+/* #define K_Control_Left         0x173 */	{LFkey, ctrl_mask},
+/* #define K_Control_Right        0x174 */	{RTkey, ctrl_mask},
 /* #define K_Control_End          0x175 */	{ENDkey, ctrl_mask},
-/* #define K_Control_PageDown     0x176 */	{PD, ctrl_mask},
+/* #define K_Control_PageDown     0x176 */	{PGDNkey, ctrl_mask},
 /* #define K_Control_Home         0x177 */	{HOMEkey, ctrl_mask},
 /* : #define K_Alt_1                0x178 */	{key_1, alt_mask},
 /* : #define K_Alt_2                0x179 */	{key_2, alt_mask},
@@ -212,7 +212,7 @@ struct pc_fkeyentry pc_xkey_map [512] = {
 /* : #define K_Alt_0                0x181 */	{key_0, alt_mask},
 /* #define K_Alt_Dash             0x182 */	{I},
 /* #define K_Alt_Equals           0x183 */	{I},
-/* #define K_Control_PageUp       0x184 */	{PU, ctrl_mask},
+/* #define K_Control_PageUp       0x184 */	{PGUPkey, ctrl_mask},
 /* #define K_F11                  0x185 */	{F11},
 /* #define K_F12                  0x186 */	{F12},
 /* #define K_Shift_F11            0x187 */	{F11, shift_mask},
@@ -221,27 +221,27 @@ struct pc_fkeyentry pc_xkey_map [512] = {
 /* #define K_Control_F12          0x18a */	{F12, ctrl_mask},
 /* #define K_Alt_F11              0x18b */	{F11, alt_mask},
 /* #define K_Alt_F12              0x18c */	{F12, alt_mask},
-/* #define K_Control_Up           0x18d */	{MUP, ctrl_mask},
+/* #define K_Control_Up           0x18d */	{UPkey, ctrl_mask},
 /* #define K_Control_KPDash       0x18e */	{screensmaller, ctrl_mask},
 /* #define K_Control_Center       0x18f */	{HOP, ctrl_mask},
 /* #define K_Control_KPPlus       0x190 */	{screenbigger, ctrl_mask},
-/* #define K_Control_Down         0x191 */	{MDN, ctrl_mask},
+/* #define K_Control_Down         0x191 */	{DNkey, ctrl_mask},
 /* #define K_Control_Insert       0x192 */	{INSkey, ctrl_mask},
 /* #define K_Control_Delete       0x193 */	{DELkey, ctrl_mask},
 					{I},
 /* : #define K_Control_KPSlash      0x195 */	{FIND},
 /* : #define K_Control_KPStar       0x196 */	{GOHOP},
 /* #define K_Alt_EHome            0x197 */	{smallHOMEkey, alt_mask},
-/* #define K_Alt_EUp              0x198 */	{MUP, alt_mask},
-/* #define K_Alt_EPageUp          0x199 */	{PU, alt_mask},
+/* #define K_Alt_EUp              0x198 */	{UPkey, alt_mask},
+/* #define K_Alt_EPageUp          0x199 */	{PGUPkey, alt_mask},
 					{I},
-/* #define K_Alt_ELeft            0x19b */	{MLF, alt_mask},
+/* #define K_Alt_ELeft            0x19b */	{LFkey, alt_mask},
 					{I},
-/* #define K_Alt_ERight           0x19d */	{MRT, alt_mask},
+/* #define K_Alt_ERight           0x19d */	{RTkey, alt_mask},
 					{I},
 /* #define K_Alt_EEnd             0x19f */	{smallENDkey, alt_mask},
-/* #define K_Alt_EDown            0x1a0 */	{MDN, alt_mask},
-/* #define K_Alt_EPageDown        0x1a1 */	{PD, alt_mask},
+/* #define K_Alt_EDown            0x1a0 */	{DNkey, alt_mask},
+/* #define K_Alt_EPageDown        0x1a1 */	{PGDNkey, alt_mask},
 /* #define K_Alt_EInsert          0x1a2 */	{smallINSkey, alt_mask},
 /* #define K_Alt_EDelete          0x1a3 */	{smallDELkey, alt_mask},
 /* #define K_Alt_KPSlash          0x1a4 */	{FIND, alt_mask},
@@ -260,33 +260,33 @@ struct pc_fkeyentry pc_xkey_map [512] = {
 	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I},
 /* #define K_EHome                0x247 */	{smallHOMEkey},
-/* #define K_EUp                  0x248 */	{MUP},
-/* #define K_EPageUp              0x249 */	{PU},
+/* #define K_EUp                  0x248 */	{UPkey},
+/* #define K_EPageUp              0x249 */	{PGUPkey},
 					{I},
-/* #define K_ELeft                0x24b */	{MLF},
+/* #define K_ELeft                0x24b */	{LFkey},
 					{I},
-/* #define K_ERight               0x24d */	{MRT},
+/* #define K_ERight               0x24d */	{RTkey},
 					{I},
 /* #define K_EEnd                 0x24f */	{smallENDkey},
-/* #define K_EDown                0x250 */	{MDN},
-/* #define K_EPageDown            0x251 */	{PD},
+/* #define K_EDown                0x250 */	{DNkey},
+/* #define K_EPageDown            0x251 */	{PGDNkey},
 /* #define K_EInsert              0x252 */	{smallINSkey},
 /* #define K_EDelete              0x253 */	{smallDELkey},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 					{I}, {I}, {I},
-/* #define K_Control_ELeft        0x273 */	{MLF, ctrl_mask},
-/* #define K_Control_ERight       0x274 */	{MRT, ctrl_mask},
+/* #define K_Control_ELeft        0x273 */	{LFkey, ctrl_mask},
+/* #define K_Control_ERight       0x274 */	{RTkey, ctrl_mask},
 /* #define K_Control_EEnd         0x275 */	{smallENDkey, ctrl_mask},
-/* #define K_Control_EPageDown    0x276 */	{PD, ctrl_mask},
+/* #define K_Control_EPageDown    0x276 */	{PGDNkey, ctrl_mask},
 /* #define K_Control_EHome        0x277 */	{smallHOMEkey, ctrl_mask},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 					{I}, {I}, {I}, {I},
-/* #define K_Control_EPageUp      0x284 */	{PU, ctrl_mask},
+/* #define K_Control_EPageUp      0x284 */	{PGUPkey, ctrl_mask},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-/* #define K_Control_EUp          0x28d */	{MUP, ctrl_mask},
+/* #define K_Control_EUp          0x28d */	{UPkey, ctrl_mask},
 					{I}, {I}, {I},
-/* #define K_Control_EDown        0x291 */	{MDN, ctrl_mask},
+/* #define K_Control_EDown        0x291 */	{DNkey, ctrl_mask},
 /* #define K_Control_EInsert      0x292 */	{smallINSkey, ctrl_mask},
 /* #define K_Control_EDelete      0x293 */	{smallDELkey, ctrl_mask},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
@@ -319,18 +319,31 @@ struct fkeyentry fkeymap [] = {
 	{"\033k",	toggleKEYMAP},
 	{"\033K",	setupKEYMAP},
 	{"\033I",	setupKEYMAP},
+	{"\033x",	AltX},
+
+	/* Alt-Digit sequences for diacritic combination prefixes */
+	{"\0330", key_0, alt_mask},
+	{"\0331", key_1, alt_mask},
+	{"\0332", key_2, alt_mask},
+	{"\0333", key_3, alt_mask},
+	{"\0334", key_4, alt_mask},
+	{"\0335", key_5, alt_mask},
+	{"\0336", key_6, alt_mask},
+	{"\0337", key_7, alt_mask},
+	{"\0338", key_8, alt_mask},
+	{"\0339", key_9, alt_mask},
 
 	/* VT100 and Sun */
-	{"\033[A",	MUP},	/* move cursor up	*/
-	{"\033[B",	MDN},	/* more cursor down	*/
-	{"\033[C",	MRT},	/* more cursor right	*/
-	{"\033[D",	MLF},	/* more cursor left	*/
+	{"\033[A",	UPkey},	/* move cursor up	*/
+	{"\033[B",	DNkey},	/* more cursor down	*/
+	{"\033[C",	RTkey},	/* more cursor right	*/
+	{"\033[D",	LFkey},	/* more cursor left	*/
 
 	/* Sun raw function key codes */
-	{"\033[215z",	MUP},	/* move cursor up	*/
-	{"\033[221z",	MDN},	/* more cursor down	*/
-	{"\033[219z",	MRT},	/* more cursor right	*/
-	{"\033[217z",	MLF},	/* more cursor left	*/
+	{"\033[215z",	UPkey},	/* move cursor up	*/
+	{"\033[221z",	DNkey},	/* more cursor down	*/
+	{"\033[219z",	RTkey},	/* more cursor right	*/
+	{"\033[217z",	LFkey},	/* more cursor left	*/
 
 	/* Linux console, Cygwin console */
 	{"\033[1~",	smallHOMEkey},	/* Pos1, VT220: FIND (@0)	*/
@@ -366,10 +379,10 @@ struct fkeyentry fkeymap [] = {
 	{"\033[1H",	HOMEkey},	/* Pos1	*/
 	{"\033[1F",	ENDkey},	/* End	*/
 	{"\033[1E",	HOP},	/* 5	*/
-	{"\033[1A",	MUP},	/* Up	*/
-	{"\033[1D",	MLF},	/* Left	*/
-	{"\033[1B",	MDN},	/* Down	*/
-	{"\033[1C",	MRT},	/* Right	*/
+	{"\033[1A",	UPkey},	/* Up	*/
+	{"\033[1D",	LFkey},	/* Left	*/
+	{"\033[1B",	DNkey},	/* Down	*/
+	{"\033[1C",	RTkey},	/* Right	*/
 
 	/* xterm generic patterns for modified function keys */
 	{"\033[1P",	F1},	/* F1	*/
@@ -404,26 +417,26 @@ struct fkeyentry fkeymap [] = {
 	{"\033[5H",	smallHOMEkey, ctrl_mask},	/* control-Pos1	*/
 	{"\033[5F",	smallENDkey, ctrl_mask},	/* control-Ende	*/
 
-	{"\0332A",	MUP, shift_mask},	/* shift-Up	*/
-	{"\0332B",	MDN, shift_mask},	/* shift-Down	*/
-	{"\0332C",	MRT, shift_mask},	/* shift-Right	*/
-	{"\0332D",	MLF, shift_mask},	/* shift-Left	*/
-	{"\0335A",	MUP, ctrl_mask},	/* control-Up	*/
-	{"\0335B",	MDN, ctrl_mask},	/* control-Down	*/
-	{"\0335C",	MRT, ctrl_mask},	/* control-Right	*/
-	{"\0335D",	MLF, ctrl_mask},	/* control-Left	*/
-	{"\0336A",	MUP, ctrlshift_mask},	/* control-shift-Up	*/
-	{"\0336B",	MDN, ctrlshift_mask},	/* control-shift-Down	*/
-	{"\0336C",	MRT, ctrlshift_mask},	/* control-shift-Right	*/
-	{"\0336D",	MLF, ctrlshift_mask},	/* control-shift-Left	*/
+	{"\0332A",	UPkey, shift_mask},	/* shift-Up	*/
+	{"\0332B",	DNkey, shift_mask},	/* shift-Down	*/
+	{"\0332C",	RTkey, shift_mask},	/* shift-Right	*/
+	{"\0332D",	LFkey, shift_mask},	/* shift-Left	*/
+	{"\0335A",	UPkey, ctrl_mask},	/* control-Up	*/
+	{"\0335B",	DNkey, ctrl_mask},	/* control-Down	*/
+	{"\0335C",	RTkey, ctrl_mask},	/* control-Right	*/
+	{"\0335D",	LFkey, ctrl_mask},	/* control-Left	*/
+	{"\0336A",	UPkey, ctrlshift_mask},	/* control-shift-Up	*/
+	{"\0336B",	DNkey, ctrlshift_mask},	/* control-shift-Down	*/
+	{"\0336C",	RTkey, ctrlshift_mask},	/* control-shift-Right	*/
+	{"\0336D",	LFkey, ctrlshift_mask},	/* control-shift-Left	*/
 
 	/* mined shortcuts */
 	{"\033\r",	Popmark},	/* alt-Enter	*/
 	{"\033\n",	Popmark},	/* alt-Enter	*/
 
 	/* vim mouse wheel mappings for xterm */
-	{"\033[64~",	PU},	/* shift-wheel-up */
-	{"\033[65~",	PD},	/* shift-wheel-down */
+	{"\033[64~",	PGUPkey},	/* shift-wheel-up */
+	{"\033[65~",	PGDNkey},	/* shift-wheel-down */
 	{"\033[62~",	SU},	/* wheel-up */
 	{"\033[63~",	SD},	/* wheel-down */
 
@@ -489,18 +502,18 @@ struct fkeyentry fkeymap [] = {
 	{"\033[63~",	F12, ctrlshift_mask}, /* control-shift-F12 */
 
 	/* further VT100 keys - some have been disabled by linux sequences */
-	{"\033[6~",	PD},	/* Next Screen, xterm PgDn	*/
-	{"\033[5~",	PU},	/* Prev Screen, xterm PgUp	*/
+	{"\033[6~",	PGDNkey},	/* Next Screen, xterm PgDn	*/
+	{"\033[5~",	PGUPkey},	/* Prev Screen, xterm PgUp	*/
 	{"\033[3~",	DELkey},	/* Remove, xterm Del	*/
 	{"\033[2~",	INSkey},	/* Insert Here, xterm Ins	*/
 	{"\033[@",	HOMEkey},	/* Pos1, PC-xterm	*/
 	{"\033[e",	ENDkey},	/* End, PC-xterm	*/
 
 	/* application cursor keys */
-	{"\033OA",	MUP},	/* up	*/
-	{"\033OB",	MDN},	/* dn	*/
-	{"\033OC",	MRT},	/* rt	*/
-	{"\033OD",	MLF},	/* lf	*/
+	{"\033OA",	UPkey},	/* up	*/
+	{"\033OB",	DNkey},	/* dn	*/
+	{"\033OC",	RTkey},	/* rt	*/
+	{"\033OD",	LFkey},	/* lf	*/
 	{"\033O@",	HOMEkey},	/* Pos1, PC-xterm	*/
 	{"\033Oe",	ENDkey},	/* End, PC-xterm	*/
 
@@ -515,39 +528,56 @@ struct fkeyentry fkeymap [] = {
 
 	{"\033Op",	INSkey},	/* Ins (kI), keypad 0	*/
 	{"\033Oq",	ENDkey},	/* End (K4), keypad 1	*/
-	{"\033Or",	MDN},	/* dn (kd), keypad 2	*/
-	{"\033Os",	PD},	/* PgDn (K5), keypad 3	*/
-	{"\033Ot",	MLF},	/* lf (kl), keypad 4	*/
-	{"\033Ou",	HOP},	/* mid (K2), keypad 5	*/
-	{"\033Ov",	MRT},	/* rt (kr), keypad 6	*/
+	{"\033Or",	DNkey},		/* dn (kd), keypad 2	*/
+	{"\033Os",	PGDNkey},	/* PgDn (K5), keypad 3	*/
+	{"\033Ot",	LFkey},		/* lf (kl), keypad 4	*/
+	{"\033Ou",	HOP},		/* mid (K2), keypad 5	*/
+	{"\033Ov",	RTkey},		/* rt (kr), keypad 6	*/
 	{"\033Ow",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
-	{"\033Ox",	MUP},	/* up (ku), keypad 8	*/
-	{"\033Oy",	PU},	/* PgUp (K3), keypad 9	*/
+	{"\033Ox",	UPkey},		/* up (ku), keypad 8	*/
+	{"\033Oy",	PGUPkey},	/* PgUp (K3), keypad 9	*/
 	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
 	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
 
 	/* mintty generic patterns for modified application mode keys */
 	{"\033[1p",	INSkey},	/* Ins (kI), keypad 0	*/
 	{"\033[1q",	ENDkey},	/* End (K4), keypad 1	*/
-	{"\033[1r",	MDN},	/* dn (kd), keypad 2	*/
-	{"\033[1s",	PD},	/* PgDn (K5), keypad 3	*/
-	{"\033[1t",	MLF},	/* lf (kl), keypad 4	*/
-	{"\033[1u",	HOP},	/* mid (K2), keypad 5	*/
-	{"\033[1v",	MRT},	/* rt (kr), keypad 6	*/
+	{"\033[1r",	DNkey},		/* dn (kd), keypad 2	*/
+	{"\033[1s",	PGDNkey},	/* PgDn (K5), keypad 3	*/
+	{"\033[1t",	LFkey},		/* lf (kl), keypad 4	*/
+	{"\033[1u",	HOP},		/* mid (K2), keypad 5	*/
+	{"\033[1v",	RTkey},		/* rt (kr), keypad 6	*/
 	{"\033[1w",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
-	{"\033[1x",	MUP},	/* up (ku), keypad 8	*/
-	{"\033[1y",	PU},	/* PgUp (K3), keypad 9	*/
+	{"\033[1x",	UPkey},		/* up (ku), keypad 8	*/
+	{"\033[1y",	PGUPkey},	/* PgUp (K3), keypad 9	*/
 	{"\033[1n",	DELkey},	/* Del (kD), keypad .	*/
 	{"\033[1l",	DELkey},	/* Del, keypad ,	*/
-	{"\033[1M",	SNL},	/* Enter (@8) */
-	{"\033[1o",	FIND},	/* keypad /	*/
-	{"\033[1j",	GOHOP},	/* keypad *	*/
+	{"\033[1M",	SNL},		/* Enter (@8) */
+	{"\033[1o",	FIND},		/* keypad /	*/
+	{"\033[1j",	GOHOP},		/* keypad *	*/
 	{"\033[1m",	KP_minus},	/* keypad -	*/
 	{"\033[1k",	KP_plus},	/* keypad +	*/
 
 	/* mintty generic patterns for modified TAB */
 	{"\033[1Z",	STAB},	/* TAB	*/
 
+	/* mlterm generic patterns for modified application mode keys */
+	{"\033O0p",	INSkey},	/* Ins (kI), keypad 0	*/
+	{"\033O0q",	ENDkey},	/* End (K4), keypad 1	*/
+	{"\033O0r",	DNkey},		/* dn (kd), keypad 2	*/
+	{"\033O0s",	PGDNkey},	/* PgDn (K5), keypad 3	*/
+	{"\033O0t",	LFkey},		/* lf (kl), keypad 4	*/
+	{"\033O0u",	HOP},		/* mid (K2), keypad 5	*/
+	{"\033O0v",	RTkey},		/* rt (kr), keypad 6	*/
+	{"\033O0w",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
+	{"\033O0x",	UPkey},		/* up (ku), keypad 8	*/
+	{"\033O0y",	PGUPkey},	/* PgUp (K3), keypad 9	*/
+	{"\033O0n",	DELkey},	/* Del (kD), keypad .	*/
+	{"\033O0o",	FIND},		/* keypad /	*/
+	{"\033O0j",	GOHOP},		/* keypad *	*/
+	{"\033O0m",	KP_minus},	/* keypad -	*/
+	{"\033O0k",	KP_plus},	/* keypad +	*/
+
 	/* Sun */
 	{"\033[208z",	F6, shift_mask},	/* R1 -> grave */
 	{"\033[209z",	F6, ctrl_mask},		/* R2 -> circumflex */
@@ -556,10 +586,10 @@ struct fkeyentry fkeymap [] = {
 	{"\033[212z",	F5, shift_mask},	/* R5 -> tilde */
 	{"\033[213z",	F5, ctrl_mask},		/* R6 -> angstrom */
 	{"\033[214z",	SU},	/* R7	*/
-	{"\033[216z",	PU},	/* R9	*/
+	{"\033[216z",	PGUPkey},	/* R9	*/
 	{"\033[218z",	HOP},	/* R11	*/
 	{"\033[220z",	SD},	/* R13	*/
-	{"\033[222z",	PD},	/* R15	*/
+	{"\033[222z",	PGDNkey},	/* R15	*/
 	{"\033[254z",	GOMA},	/* - */
 	{"\033[253z",	MARK},	/* + */
 	{"\033[250z",	COPY},	/* Enter */
@@ -569,8 +599,8 @@ struct fkeyentry fkeymap [] = {
 	{"\033[4z",	ENDkey},	/* End, non-appl */
 	{"\033[2z",	INSkey},	/* Ins on VT100 block */
 	{"\033[3z",	DELkey},	/* Del (xterm/Sun) */
-	{"\033[5z",	PU},	/* PgUp on VT100 block */
-	{"\033[6z",	PD},	/* PgDn on VT100 block */
+	{"\033[5z",	PGUPkey},	/* PgUp on VT100 block */
+	{"\033[6z",	PGDNkey},	/* PgDn on VT100 block */
 	{"\033[192z",	MARK},	/* Stop (F11 with xterm "Sun Function-Keys") */
 	{"\033[193z",	AGAIN},	/* Again (F12 with xterm "Sun Function-Keys") */
 	{"\033[194z",	FS},	/* Props */
@@ -601,9 +631,9 @@ struct fkeyentry fkeymap [] = {
 	{"\033[139q",	INSkey},	/* Insert, Ins */
 	{"\033[P",	DELkey},	/* Del */
 /*	{"\033[H",	smallHOMEkey},	*//* Home */
-	{"\033[150q",	PU},	/* PgUp */
+	{"\033[150q",	PGUPkey},	/* PgUp */
 	{"\033[146q",	smallENDkey},	/* End */
-	{"\033[154q",	PD},	/* PgDn */
+	{"\033[154q",	PGDNkey},	/* PgDn */
 	{"\033[000q",	HOP},	/* 5 */
 	{"\033[001q",	F1},	/* F1 */
 	{"\033[002q",	F2},	/* F2 */
@@ -649,11 +679,14 @@ struct fkeyentry fkeymap [] = {
 
 	/* xterm mouse */
 	{"\033[M",	DIRECTxterm},	/* direct cursor address */
-	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
-	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */
+	{"\033[t",	TRACKxterm},	/* xterm mouse highlight tracking */
+	{"\033[T",	TRACKxtermT},	/* xterm mouse highlight tracking */
 	{"\033[O",	FOCUSout},	/* window loses focus */
 	{"\033[I",	FOCUSin},	/* window gains focus */
 
+	/* DEC locator */
+	{"\033[&w",	DIRECTvtlocator},	/* DECterm locator */
+
 	/* mintty ambiguous width change report */
 	{"\033[1W",	AMBIGnarrow},	/* ambig. width chars are narrow */
 	{"\033[2W",	AMBIGwide},	/* ambig. width chars are wide */
@@ -661,15 +694,18 @@ struct fkeyentry fkeymap [] = {
 	/* mintty application escape key mode (enabled with ^[[?7727h) */
 	{"\033O[",	ESCAPE},	/* ESC */
 
+	/* crttool */
+	{"\033m",	DIRECTcrttool},	/* direct cursor address */
+
 	{NIL_PTR}
 };
 
 struct fkeyentry fkeymap_vt52 [] = {
 	/* VT52 and HP */
-	{"\033A",	MUP},	/* up	*/
-	{"\033B",	MDN},	/* dn	*/
-	{"\033C",	MRT},	/* rt	*/
-	{"\033D",	MLF},	/* lf	*/
+	{"\033A",	UPkey},	/* up	*/
+	{"\033B",	DNkey},	/* dn	*/
+	{"\033C",	RTkey},	/* rt	*/
+	{"\033D",	LFkey},	/* lf	*/
 	{"\033P",	F1},	/* PF1 */
 	{"\033Q",	F2},	/* PF2 */
 	{"\033R",	F3},	/* PF3 */
@@ -678,14 +714,14 @@ struct fkeyentry fkeymap_vt52 [] = {
 	/* keypad application mode */
 	{"\033?p",	INSkey},	/* keypad 0 */
 	{"\033?q",	ENDkey},	/* keypad 1 */
-	{"\033?r",	MDN},	/* keypad 2 */
-	{"\033?s",	PD},	/* keypad 3 */
-	{"\033?t",	MLF},	/* keypad 4 */
+	{"\033?r",	DNkey},	/* keypad 2 */
+	{"\033?s",	PGDNkey},	/* keypad 3 */
+	{"\033?t",	LFkey},	/* keypad 4 */
 	{"\033?u",	HOP},	/* keypad 5 */
-	{"\033?v",	MRT},	/* keypad 6 */
+	{"\033?v",	RTkey},	/* keypad 6 */
 	{"\033?w",	HOMEkey},	/* keypad 7 */
-	{"\033?x",	MUP},	/* keypad 8 */
-	{"\033?y",	PU},	/* keypad 9 */
+	{"\033?x",	UPkey},	/* keypad 8 */
+	{"\033?y",	PGUPkey},	/* keypad 9 */
 	{"\033?l",	DELkey},	/* keypad , */
 	{"\033?m",	SU},	/* keypad - */
 	{"\033?n",	DELkey},	/* keypad . */
@@ -706,10 +742,10 @@ struct fkeyentry fkeymap_vt52 [] = {
 	{"\0331H",	HOMEkey},	/* modified Pos1	*/
 	{"\0331F",	ENDkey},	/* modified End	*/
 	{"\0331E",	HOP},		/* modified keypad 5	*/
-	{"\0331A",	MUP},	/* modified up	*/
-	{"\0331B",	MDN},	/* modified dn	*/
-	{"\0331C",	MRT},	/* modified rt	*/
-	{"\0331D",	MLF},	/* modified lf	*/
+	{"\0331A",	UPkey},	/* modified up	*/
+	{"\0331B",	DNkey},	/* modified dn	*/
+	{"\0331C",	RTkey},	/* modified rt	*/
+	{"\0331D",	LFkey},	/* modified lf	*/
 	{"\0331P",	F1},	/* modified PF1 (could be) */
 	{"\0331Q",	F2},	/* modified PF2 (could be) */
 	{"\0331R",	F3},	/* modified PF3 (could be) */
@@ -720,26 +756,26 @@ struct fkeyentry fkeymap_vt52 [] = {
 
 struct fkeyentry fkeymap_hp [] = {
 	/* VT52 and HP */
-	{"\033A",	MUP},	/* up	*/
-	{"\033B",	MDN},	/* dn	*/
-	{"\033C",	MRT},	/* rt	*/
-	{"\033D",	MLF},	/* lf	*/
+	{"\033A",	UPkey},	/* up	*/
+	{"\033B",	DNkey},	/* dn	*/
+	{"\033C",	RTkey},	/* rt	*/
+	{"\033D",	LFkey},	/* lf	*/
 	{"\033P",	DELkey}, /* VT52 PF1 Del */
 	{"\033Q",	INSkey}, /* VT52 PF2 Ins (termcap: insert character) */
 	{"\033R",	HOP},	/* VT52 PF3 */
-	{"\033S",	PD},	/* VT52 PF4
+	{"\033S",	PGDNkey}, /* VT52 PF4
 				   termcap xterm-hp: PgDn
 				   termcap hp*: scroll-forward/backward key
-				*/
+				  */
 	/* HP */
-	{"\033T",	PU},	/* termcap xterm-hp: PgUp
+	{"\033T",	PGUPkey}, /* termcap xterm-hp: PgUp
 				   termcap hp*: scroll-backward/forward key
-				*/
+				  */
 	{"\033h",	HOMEkey},	/* Pos1	*/
 	{"\033F",	ENDkey},	/* End	*/
-	{"\033V",	PU},
-	{"\033U",	PD},
-	{"\033L",	LIB},	/* insert line	*/
+	{"\033V",	PGUPkey},
+	{"\033U",	PGDNkey},
+	{"\033L",	APPNL},	/* insert line	*/
 	{"\033J",	DELchar},	/* erase (clear to (end of) screen)	*/
 	{"\033K",	DLN},	/* clear to end of line	*/
 	{"\033M",	DLINE},	/* delete line	*/
@@ -765,13 +801,13 @@ struct fkeyentry fkeymap_hp [] = {
 
 	/* xterm HP mode generic patterns for modified keypad keys */
 	{"\033[1h",	HOMEkey},	/* Pos1	*/
-	{"\033[1T",	PU},		/* PgUp */
-	{"\033[1S",	PD},		/* PgDn */
+	{"\033[1T",	PGUPkey},	/* PgUp */
+	{"\033[1S",	PGDNkey},	/* PgDn */
 	{"\033[1Q",	INSkey},	/* Ins */
 	{"\033[1P",	DELkey},	/* Del */
 
 	/* crttool */
-	{"\033m",	DIRECTcrttool},	/* direct cursor address */
+/*	{"\033m",	DIRECTcrttool},	direct cursor address */
 
 	/* ?? */
 	{"\033&",	SNL},	/* keypad Enter */
@@ -783,12 +819,12 @@ struct fkeyentry fkeymap_hp [] = {
 	/* HP workstation keyboard without NumLock */
 	/* conflicts with VT100+ etc, see above */
 	{"\033OU",	MARK},	/* Pos1	*/
-	{"\033OV",	MLF},	/* lf	*/
-	{"\033OW",	MUP},	/* up	*/
-	{"\033OX",	MRT},	/* rt	*/
-	{"\033OY",	MDN},	/* dn	*/
-	{"\033OZ",	PU},	/* PgUp	*/
-	{"\033O?",	PD},	/* PgDn; End, Ins, Del?	*/
+	{"\033OV",	LFkey},
+	{"\033OW",	UPkey},
+	{"\033OX",	RTkey},
+	{"\033OY",	DNkey},
+	{"\033OZ",	PGUPkey},	/* PgUp	*/
+	{"\033O?",	PGDNkey},	/* PgDn; End, Ins, Del?	*/
 #endif
 
 struct fkeyentry fkeymap_rxvt [] = {
@@ -813,14 +849,14 @@ struct fkeyentry fkeymap_rxvt [] = {
 
 	/* rxvt-unicode application keypad */
 	{"\033OU",	HOMEkey},	/* Pos1	*/
-	{"\033OV",	MLF},	/* lf	*/
-	{"\033OW",	MUP},	/* up	*/
-	{"\033OX",	MRT},	/* rt	*/
-	{"\033OY",	MDN},	/* dn	*/
-	{"\033OZ",	PU},	/* PgUp	*/
+	{"\033OV",	LFkey},	/* lf	*/
+	{"\033OW",	UPkey},	/* up	*/
+	{"\033OX",	RTkey},	/* rt	*/
+	{"\033OY",	DNkey},	/* dn	*/
+	{"\033OZ",	PGUPkey},	/* PgUp	*/
 	{"\033O]",	HOP},	/* 5	*/
 	{"\033O\\",	ENDkey},	/* End	*/
-	{"\033O[",	PD},	/* PgDn	*/
+	{"\033O[",	PGDNkey},	/* PgDn	*/
 	{"\033O^",	INSkey},	/* Ins	*/
 	{"\033O_",	DELkey},	/* Del	*/
 
@@ -851,8 +887,8 @@ struct fkeyentry fkeymap_linux [] = {
 struct fkeyentry fkeymap_xterm [] = {
 	/* xterm */
 	{"\033[M",	DIRECTxterm},	/* direct cursor address */
-	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
-	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */
+	{"\033[t",	TRACKxterm},	/* xterm mouse highlight tracking */
+	{"\033[T",	TRACKxtermT},	/* xterm mouse highlight tracking */
 	{"\033[O",	FOCUSout},	/* window loses focus */
 	{"\033[I",	FOCUSin},	/* window gains focus */
 
@@ -869,14 +905,14 @@ struct fkeyentry fkeymap_vt100 [] = {
 	/* override wrong terminfo-derived settings */
 	{"\033Op",	INSkey},	/* Ins (kI), keypad 0	*/
 	{"\033Oq",	ENDkey},	/* End (K4), keypad 1	*/
-	{"\033Or",	MDN},	/* dn (kd), keypad 2	*/
-	{"\033Os",	PD},	/* PgDn (K5), keypad 3	*/
-	{"\033Ot",	MLF},	/* lf (kl), keypad 4	*/
-	{"\033Ou",	HOP},	/* mid (K2), keypad 5	*/
-	{"\033Ov",	MRT},	/* rt (kr), keypad 6	*/
+	{"\033Or",	DNkey},		/* dn (kd), keypad 2	*/
+	{"\033Os",	PGDNkey},	/* PgDn (K5), keypad 3	*/
+	{"\033Ot",	LFkey},		/* lf (kl), keypad 4	*/
+	{"\033Ou",	HOP},		/* mid (K2), keypad 5	*/
+	{"\033Ov",	RTkey},		/* rt (kr), keypad 6	*/
 	{"\033Ow",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
-	{"\033Ox",	MUP},	/* up (ku), keypad 8	*/
-	{"\033Oy",	PU},	/* PgUp (K3), keypad 9	*/
+	{"\033Ox",	UPkey},		/* up (ku), keypad 8	*/
+	{"\033Oy",	PGUPkey},	/* PgUp (K3), keypad 9	*/
 	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
 	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
 
@@ -924,15 +960,15 @@ struct fkeyentry fkeymap_scoansi [] = {
 	{"\033[t",	F10, ctrl_mask},	/* control-F10, PuTTY SCO */
 	{"\033[u",	F11, ctrl_mask},	/* control-F11, PuTTY SCO */
 	{"\033[v",	F12, ctrl_mask},	/* control-F12, PuTTY SCO */
-	{"\033OA",	MUP, ctrl_mask},	/* control-up, PuTTY SCO */
-	{"\033OB",	MDN, ctrl_mask},	/* control-dn, PuTTY SCO */
-	{"\033OC",	MRT, ctrl_mask},	/* control-rt, PuTTY SCO */
-	{"\033OD",	MLF, ctrl_mask},	/* control-lf, PuTTY SCO */
+	{"\033OA",	UPkey, ctrl_mask},	/* control-up, PuTTY SCO */
+	{"\033OB",	DNkey, ctrl_mask},	/* control-dn, PuTTY SCO */
+	{"\033OC",	RTkey, ctrl_mask},	/* control-rt, PuTTY SCO */
+	{"\033OD",	LFkey, ctrl_mask},	/* control-lf, PuTTY SCO */
 	{"\033[L",	INSkey},	/* (termcap: insert character) */
 	{"\033[H",	HOMEkey},	/* 	*/
 	{"\033[F",	ENDkey},	/* (xterm-sco)	*/
-	{"\033[G",	PD},	/* 	*/
-	{"\033[I",	PU},	/* 	*/
+	{"\033[G",	PGDNkey},	/* 	*/
+	{"\033[I",	PGUPkey},	/* 	*/
 
 	{NIL_PTR}
 };
@@ -940,15 +976,15 @@ struct fkeyentry fkeymap_scoansi [] = {
 struct fkeyentry fkeymap_interix [] = {
 	/* Interix console ("Windows Services for UNIX") */
 	{"\033[H",	HOMEkey},
-	{"\033[S",	PU},
-	{"\033[T",	PD},
+	{"\033[S",	PGUPkey},
+	{"\033[T",	PGDNkey},
 	{"\033[U",	ENDkey},
-	{"\033[F-",	MUP, shift_mask},
-	{"\033[F+",	MDN, shift_mask},
-	{"\033[F$",	MRT, shift_mask},
-	{"\033[F^",	MLF, shift_mask},
+	{"\033[F-",	UPkey, shift_mask},
+	{"\033[F+",	DNkey, shift_mask},
+	{"\033[F$",	RTkey, shift_mask},
+	{"\033[F^",	LFkey, shift_mask},
 	{"\033[L",	INSkey},
-/*	{"",	DELkey, ctrl_mask},	*/
+/*	{"\037",	DELkey, ctrl_mask},	ctrl-DEL sends ^_ */
 
 	{"\033F1",	F1},
 	{"\033F2",	F2},
@@ -1021,8 +1057,8 @@ struct fkeyentry fkeymap_siemens [] = {
 	{"\033:",	MNW},		/* -~> */
 	{"\0336",	HOP},		/* key next to END, above cursor block */
 	{"\033[@",	INSkey},	/* Insert */
-	{"\033[S",	PU},
-	{"\033[T",	PD},
+	{"\033[S",	PGUPkey},
+	{"\033[T",	PGDNkey},
 	{"\033^",	HOP},
 
 	{"\033@",	F1},	/* F1 */
