@@ -30,13 +30,57 @@
 #endif
 
 
+/* system options */
+#ifdef _AIX
+#define _XOPEN_SOURCE_EXTENDED 1
+#define _POSIX_SOURCE
+#include <sys/access.h>
+#endif
+
+#ifdef __QNX__
+/* prototypes getpwuid, ttyname */
+#define _POSIX_SOURCE
+#endif
+
+#ifdef __minix
+#define stat _stat
+#define fstat _fstat
+#define errprintf(msg, par)	{char errbuf [maxMSGlen]; sprintf (errbuf, msg, par); write (2, errbuf, strlen (errbuf));}
+#else
+#define errprintf(msg, par)	fprintf (stderr, msg, par)
+#endif
+
+
 /* common includes */
+#include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#if defined (__TURBOC__) || defined (VAXC)
+# ifdef __TURBOC__
+# define off_t long
+# endif
+#else
+#define __USE_XOPEN_EXTENDED
 #include <unistd.h>
+#endif
 
 
-/* MSDOS includes and declarations */
+/* PC includes and declarations */
+#ifdef __MINGW32__
+#warning [41m mined does not compile with the MinGW compiler - use the MSYS compiler instead [0m
+#warning [41m see COMPILE.DOC for details [0m
+#endif
+
+#ifdef __MINGW32__
+/*#define sleep _sleep - this is __MINGW_ATTRIB_DEPRECATED*/
+#define sleep(n) usleep ((unsigned long) n * 1000000)
+#endif
+
+#ifdef __MINGW32__
+#define mkdir__ mkdir
+#define mkdir(f, m)	mkdir__(f)
+#endif
+
 #ifdef msdos
 extern void delay ();
 extern void exit ();
@@ -48,35 +92,24 @@ extern void exit ();
 extern char * getenv ();
 extern char * getcwd ();
 
-# ifdef __TURBOC__
+#ifdef __TURBOC__
 extern void setdisk ();
 extern void sleep ();
-# endif
-
-# ifdef __TURBOC__
 #define time_t	unsigned long
-/* #define conio with combined positioning/output method...? */
-#define conio
-# else
-#define conio
-# endif
+#endif
 
 #endif /* msdos */
 
-#ifdef conio
-#define dosmouse
-#endif
-
 
 /* VMS includes and declarations */
 #ifdef vms
-#  include <unixio.h>
-#  include <file.h>	/* for the O_... attributes to open () */
-#  include <unixlib.h>
-#  undef putchar
-#  undef FALSE
-#  undef TRUE
-#  define CURSES
+# include <unixio.h>
+# include <file.h>	/* for the O_... attributes to open () */
+# include <unixlib.h>
+# undef putchar
+# undef FALSE
+# undef TRUE
+# define use_usleep
 #endif
 
 
@@ -115,33 +148,31 @@ extern char * getenv ();
 #include <sys/types.h>
 #endif
 
+#ifdef _CLASSIC_ID_TYPES
+/* HP, AIX, IRIX */
+#define pid_t int
+#endif
+
+#ifdef __sgi
+/* IRIX: no vfork */
+#define FORK
+#endif
+
 #ifdef DEFVFORK
 /* already declared in <unistd.h> */
-#ifdef __STDC__
 extern pid_t vfork ();
-#else
-#ifdef _CLASSIC_ID_TYPES
-/* is this a HP special? */
-extern int vfork ();
-#else
-extern pid_t vfork ();
-#endif
-#endif
 #endif
 
 #ifdef __CYGWIN__
-extern int vfork ();
+extern pid_t vfork ();
 #endif
 
-#ifdef __GCC__
 extern pid_t wait ();
-#else
-extern pid_t wait ();
-#endif
 
 extern int select ();
+#ifndef __MINGW32__
 extern unsigned int sleep ();
-
+#endif
 #ifdef use_usleep
 extern unsigned int usleep ();
 #endif
@@ -167,20 +198,7 @@ extern unsigned int usleep ();
 #endif
 
 
-/* CJK tables mode */
-#ifndef msdos
-# ifndef vms
-# define use_cjk_tables
-# endif
-#define use_keymap_tables
-#else
-# ifndef __TURBOC__
-# define use_keymap_tables
-# endif
-#endif
-
-
-/* this is again unused :) since it broke compilation on SunOS 5.10 */
+/* this is again unused since it broke compilation on SunOS 5.10 */
 #ifdef __GNUC__
 #define __unused__ __attribute__((unused))
 #else
@@ -188,6 +206,22 @@ extern unsigned int usleep ();
 #endif
 
 
+/* missing definitions */
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
+#ifndef F_OK
+#define F_OK 0
+#endif
+#ifndef X_OK
+#define X_OK 1
+#endif
+#ifndef W_OK
+#define W_OK 2
+#endif
+
+
 /***************************************************************************\
 	declaration of types, basic macros, and constants
 \***************************************************************************/
@@ -213,21 +247,23 @@ extern short MENU;
 	screen size 2560x1600 (WQXGA+)
 		font size 4x6:	267 x 639
    Restrain to reasonable size assumption (esp. @ reasonable font size), 
-   so prompt line buffer stays @ reasonable size < MAX_CHARS,
-   so let maxXMAX < MAX_CHARS / 3
+   so prompt line buffer stays @ reasonable size < maxLINElen,
+   so let maxXMAX < maxLINElen / 3
  */
-#define maxXMAX		319
-#define maxPROMPTlen	(maxXMAX + 1)	/* max screen line length, also used for prompt buffer */
+#define maxXMAX		638
+#define maxPROMPTlen	(maxXMAX + 1)	/* limit prompt to max screen width */
 #define maxFILENAMElen	maxPROMPTlen
 #define maxCMDlen	2 * maxPROMPTlen + 9
 
-#define MAX_CHARS	1024		/* max chars on one line of text */
-		/* LINE_START must be rounded up to the lowest SHIFT_SIZE */
-#define LINE_START	(((-MAX_CHARS - 1) / SHIFT_SIZE) * SHIFT_SIZE \
-				   - SHIFT_SIZE)
-#define LINE_END	(MAX_CHARS * 8)	/* Must be larger than highest 
-					x-coordinate for line; MAX_CHARS + 1
-					is not sufficient due to TABs */
+#define maxLINElen	1024		/* max chars on one line of text */
+#define maxMSGlen	maxLINElen
+
+/* pseudo x coordinates to indicate positioning to start/end of line */
+	/* LINE_START must be rounded up to the lowest SHIFT_SIZE */
+#define LINE_START	(((-maxLINElen - 1) / SHIFT_SIZE) * SHIFT_SIZE - SHIFT_SIZE)
+	/* LINE_END must be larger than highest x-coordinate for line; 
+	   considering wide characters ands TABs */
+#define LINE_END	(maxLINElen * 8)
 
 
 /* Return values of functions */
@@ -244,10 +280,6 @@ extern short MENU;
 
 #define REPORT_CHANGED_LINES	1	/* Report change of lines on # lines */
 
-#ifndef O_BINARY
-#define O_BINARY 0
-#endif
-
 /*
  * mouse button selection
  */
@@ -260,7 +292,7 @@ typedef enum {
 } mousebutton_type;
 
 #define shift_button	4
-#define meta_button	8
+#define alt_button	8
 #define control_button	16
 
 /*
@@ -293,6 +325,13 @@ typedef enum {
   OPEN
 } FLAG;
 
+typedef enum {LEFTDOUBLE, RIGHTDOUBLE, LEFTSINGLE, RIGHTSINGLE} quoteposition_type;
+
+#ifdef VAXC
+/* prevent %CC-E-INVOPERAND */
+#define FLAG unsigned char
+#endif
+
 /*
 	line end definitions
  */
@@ -322,10 +361,14 @@ struct Line {
   FLAG dirty;		/* flag that display attributes have changed */
 };
 /* display flags for struct Line */
-#define syntax_none	0
-#define syntax_HTML	1
-#define syntax_JSP	2
-#define syntax_comment	4
+#define syntax_none	0x00
+#define syntax_HTML	0x01
+#define syntax_JSP	0x02
+#define syntax_PHP	0x04
+#define syntax_scripting	(syntax_JSP | syntax_PHP)
+#define syntax_comment	0x08
+#define syntax_attrib	0x10
+#define syntax_value	0x20
 
 typedef struct Line LINE;
 
@@ -382,6 +425,7 @@ extern int lines_per_page;	/* assumption for file_status */
 
 extern int x, y;		/* x, y coordinates on screen */
 
+extern FLAG use_file_tabs;
 #define viewonly	(viewonly_mode || viewonly_locked || viewonly_err)
 extern FLAG viewonly_mode;	/* Set when view only mode is selected */
 extern FLAG viewonly_locked;	/* Enforced view only status */
@@ -393,11 +437,14 @@ extern FLAG winchg;		/* Set when the window size changes */
 extern FLAG interrupted;	/* Set when a signal interrupts */
 extern FLAG waitingforinput;	/* Set while waiting for the next command key */
 extern FLAG isscreenmode;	/* Set when screen mode is on */
-extern char text_buffer [MAX_CHARS];	/* Buffer for modifying text */
+extern char text_buffer [maxLINElen];	/* Buffer for modifying text */
 
+extern char * minedprog;	/* store argv [0] for help invocation */
 extern char file_name [];	/* name of file being edited */
 extern FLAG writable;		/* Set if file cannot be written */
 extern FLAG file_is_dir;	/* Tried to open a directory as file? */
+extern FLAG file_is_dev;	/* Tried to open a char/block device file? */
+extern FLAG file_is_fifo;	/* Read from a named pipe? */
 extern FLAG reading_pipe;	/* Set if file should be read from stdin */
 extern FLAG writing_pipe;	/* Set if file should be written to stdout */
 
@@ -406,12 +453,13 @@ extern PROT fprot0;		/* default prot. mode for new files */
 extern PROT fprot1;		/* prot. mode for new file being edited */
 extern PROT bufprot;		/* prot. mode for paste buffer file */
 
-#ifdef unix
+#if defined (unix) || defined (ANSI)
 extern char * title_string_pattern;
 #endif
 
 extern char yank_file [];	/* Temp. file for buffer */
 extern char yankie_file [];	/* Temp. file for inter-window buffer */
+extern char spool_file [];	/* Temp. file for printing */
 extern char html_file [];	/* Temp. file for HTML embedding buffer */
 extern char panic_file [];	/* file for panic-write-back */
 extern FLAG yank_status;	/* Status of yank_file */
@@ -425,8 +473,27 @@ extern int buffer_open_flag;	/* counter flag for the collective buffer */
 extern int JUSlevel;		/* keep justified while typing? */
 extern int JUSmode;		/* 1: paragraphs end at empty line */
 extern FLAG autoindent;		/* auto indent on input of Enter? */
-extern FLAG mark_HTML;		/* display HTML marked/dimmed ? */
-extern FLAG insert_mode;	/* insert or overwrite */
+extern FLAG lowcapstrop;	/* capitalize letter symbols on input? */
+extern FLAG dispunstrop;	/* display stropped letters in bold? */
+extern FLAG strop_selected;	/* was stropping explictly selected? */
+extern char strop_style;	/* bold/underline? */
+extern FLAG mark_HTML;		/* display HTML marked ? */
+extern FLAG mark_JSP;		/* Display JSP marked ? */
+extern FLAG viewing_help;
+
+/* emulation and keyboard assignment flags */
+extern char emulation;		/* 'w' for WordStar, 'e' for emacs */
+extern char keypad_mode;	/* 'm'ined, 'S'hift-select, 'w'indows */
+extern FLAG shift_selection;	/* selection highlighting on shift-cursor */
+#define apply_shift_selection (shift_selection && (shift_selection == True || (keyshift & shift_mask)))
+extern FLAG mined_keypad;	/* Apply mined keypad assignments */
+extern FLAG home_selection;	/* numeric keypad Home/End forced to Mark/Copy */
+extern FLAG small_home_selection;	/* small keypad Home/End: Mark/Copy */
+extern FLAG emacs_buffer;	/* enable emacs buffer fct for ^K/^T */
+extern FLAG paste_stay_left;	/* cursor stays before pasted region */
+extern FLAG tab_left;		/* should moving up/down on TAB go left? */
+extern FLAG tab_right;		/* should moving up/down on TAB go right? */
+
 extern FLAG append_flag;	/* Set when buffer should be appended to */
 extern FLAG pastebuf_utf8;	/* Paste buffer always treated as UTF-8? */
 extern FLAG rectangular_paste_flag;	/* Copy/Paste rectangular area? */
@@ -435,20 +502,17 @@ extern FLAG visselect_anymouse;		/* Visual selection on any mouse move? */
 extern FLAG visselect_keeponsearch;	/* Keep visual selection on search ? */
 extern FLAG visselect_setonfind;	/* Select search result? */
 extern FLAG visselect_autocopy;		/* Auto-copy selection? */
+extern FLAG visselect_copydeselect;	/* Deselect on copy? */
 extern FLAG visselect_autodelete;	/* Delete selection on insert? */
 
+extern FLAG insert_mode;	/* insert or overwrite */
+
 extern char backup_mode;	/* none, simple, numbered (e/v), automatic */
 extern char * backup_directory;	/* directory name for backup files */
 extern char * recover_directory;	/* directory name for recovery files */
 extern character erase_char;	/* effective (configured) char for erase left */
-extern char emulation;		/* 'w' for WordStar, 'e' for emacs */
-extern FLAG emacs_buffer;	/* enable emacs buffer fct for ^K/^T */
-extern FLAG paste_stay_left;	/* cursor stays before pasted region */
-extern FLAG tab_left;		/* should moving up/down on TAB go left? */
-extern FLAG tab_right;		/* should moving up/down on TAB go right? */
+extern FLAG prefer_comspec;	/* for ESC ! command */
 extern character quit_char;	/* ^\/^G character to cancel command */
-extern FLAG Turkish;		/* use Turkish case toggle specials ? */
-extern FLAG Lithuanian;		/* use Lithuanian case toggle specials ? */
 extern FLAG smart_quotes;	/* replace " with typographic quote ? */
 
 extern FLAG suppress_unknown_cjk;	/* on CJK terminal if no Unicode mapping */
@@ -463,12 +527,14 @@ extern FLAG utf_cjk_wide_padding;	/* always display CJK on UTF double-width ? */
 
 extern FLAG dark_term;		/* dark colour terminal ? */
 extern FLAG bw_term;		/* black/white terminal ? */
+extern FLAG suppress_colour;	/* don't use ANSI color settings */
 
 extern FLAG configure_xterm_keyboard;	/* deleteIsDEL, metaSendsEscape */
 
 /* from mined1.c */
 extern FLAG auto_detect;	/* Auto detect character encoding from file ? */
 extern char * detect_encodings;	/* List of encodings to detect */
+extern char language_tag;
 
 extern FLAG cjk_text;		/* text in CJK encoding ? */
 extern FLAG utf8_text;		/* text in UTF-8 representation ? */
@@ -492,8 +558,8 @@ extern FLAG loading;		/* Loading a file? Init True for error handling */
 extern FLAG only_detect_text_encoding;
 extern FLAG wordnonblank;	/* handle all non-blank sequences as words */
 extern FLAG proportional;	/* Enable support for proportional fonts? */
-extern int hide_password_mode;	/* Hide / Hide in dot files / Show */
-extern FLAG hide_password;	/* Hide passwords in display */
+extern FLAG show_vt100_graph;	/* Display letters as VT100 block graphics ? */
+extern FLAG hide_passwords;	/* Hide passwords in display */
 extern FLAG translate_output;	/* Transform output diacritics to strings */
 extern int translen;		/* length of " */
 extern char * transout;		/* Output transformation table */
@@ -535,7 +601,8 @@ extern char * menu_cont_marker;	/* Char to mark menu border continuation */
 extern char * menu_cont_fatmarker; /* Char to mark menu border continuation */
 
 extern FLAG stat_visible;	/* Set if status line is visible */
-extern FLAG top_line_scrolled;	/* Was menu line scrolled away? */
+extern FLAG top_line_dirty;	/* Was menu line cleared? */
+extern FLAG text_screen_dirty;	/* Was text display position affected? */
 
 extern FLAG page_scroll;	/* use scroll for page up/down */
 extern FLAG page_stay;		/* stay at edge of page after page up/down */
@@ -565,8 +632,6 @@ extern void move_address_w_o_RD _((char *, int y));
 extern int find_x _((LINE * line, char * address));
 extern int get_cur_pos _((void));
 extern int get_cur_col _((void));
-extern int get_text_col _((LINE *, char *));
-/* from edit.c */
 extern int find_y _((LINE * line));
 extern int find_y_w_o_RD _((LINE * line));
 
@@ -577,10 +642,11 @@ extern void set_modified _((void));
 extern void file_status _((char * message, long bytecount, long charcount, 
 			char * filename, int lines, FLAG textstat, 
 			FLAG writefl, FLAG changed, FLAG viewing));
+extern void recount_chars _((void));
 
 /* from edit.c */
 extern char syntax_state _((char prev_mask, char old_mask, char * spoi, char * sbeg));
-extern void update_text_state _((LINE *));
+extern void update_syntax_state _((LINE *));
 
 /* from pastebuf.c */
 extern void delete_yank_files _((void));
@@ -593,10 +659,12 @@ extern void delete_text_buf _((LINE * start_line, char * start_textp, LINE * end
 /* selection highlighting */
 /* from pastebuf.c */
 extern void start_highlight_selection _((void));
-extern void continue_highlight_selection _((void));
+extern void continue_highlight_selection _((int pos_x));
+extern void update_selection_marks _((int pos_x));
 extern void clear_highlight_selection _((void));
 extern FLAG has_active_selection _((void));
-extern void update_selection_marks _((int pos_x));
+extern void trigger_highlight_selection _((void));
+extern void adjust_rectangular_mode _((FLAG alt_mouse));
 
 /* editing functions */
 /* from edit.c */
@@ -645,6 +713,7 @@ extern void rd_bottom_line _((void));
 extern void redraw_prompt _((void));
 extern void clear_lastline _((void));
 /* file chooser */
+extern int wildcard _((char * pat, char * s));
 extern int get_filename _((char * message, char * filename, FLAG directory));
 extern FLAG sort_by_extension;
 extern FLAG sort_dirs_first;
@@ -653,7 +722,8 @@ extern FLAG sort_dirs_first;
 /*
  * Convert cnt to nearest tab position
  */
-extern int tabsize;		/* Width of tab positions, 4 or 8 */
+extern int tabsize;		/* Width of tab positions, 2 or 4 or 8 */
+extern FLAG tabsize_selected;	/* Tab width selected by explicit option? */
 extern FLAG expand_tabs;	/* Expand TABs to Spaces? */
 #define tab(cnt)		(((cnt) + tabsize) & ~(tabsize - 1))
 #define tab8(cnt)		(((cnt) + 8) & ~07)
@@ -666,7 +736,10 @@ extern FLAG expand_tabs;	/* Expand TABs to Spaces? */
 
 
 /* from edit.c */
+extern int ishex _((char));
+extern unsigned int hexval _((char));
 extern int get_idf _((char * idf_buf, char * text, char * start_line));
+extern FLAG iscombined_unichar _((unsigned long ucs, char * charpos, char * linebegin));
 
 /* from search.c */
 extern void search_for _((char *, FLAG direction, FLAG case_ignore));
@@ -674,14 +747,24 @@ extern void search_expr _((char *, FLAG direction, FLAG case_ignore));
 extern void search_corresponding _((char *, FLAG, char *));
 
 /* from mined1.c */
-extern void viewonlyerr _((void));
-extern void restrictederr _((void));
+extern void splash_logo _((void));
+extern void switchAltScreen _((void));
+extern void KEYREC _((void));
 
 extern void emul_emacs _((void));
 extern void emul_WordStar _((void));
 extern void emul_pico _((void));
 extern void emul_Windows _((void));
 extern void emul_mined _((void));
+extern void set_keypad_mined _((void));
+extern void set_keypad_shift_selection _((void));
+extern void set_keypad_windows _((void));
+
+extern void handle_locale_quotes _((char * lang, FLAG alt));
+extern void configure_preferences _((FLAG applycommon));
+
+extern void viewonlyerr _((void));
+extern void restrictederr _((void));
 
 extern void convlineend_cur_LF _((void));
 extern void convlineend_cur_CRLF _((void));
@@ -705,22 +788,23 @@ extern int systemcall _((char * msg, char * cmd, int delay));
 extern int progcallpp _((char * msg, int delay, char * binprepath [], 
 	char * dir, char * prog, char * p1, char * p2, char * p3, char * p4));
 #endif
-#ifdef msdos
+#if defined (msdos) || defined (__MINGW32__)
 extern int is_Windows _((void));
 #endif
 extern void panic _((char *, char *));
 extern void panicio _((char * message, char * err));
 extern char * scan_int _((char *, int *));
 extern char * dec_out _((long));
-extern char * serror _((void));
-extern char * serrorof _((int));
-extern int geterrno _((void));
 extern int delete_file _((char *));
 extern FLAG copyfile _((char * from_file, char * to_file));
 extern char * unnull _((char *));
+extern void strip_trailingslash _((char *));
 extern char * num_out _((long number, long radix));
 extern void copy_string _((char * to, char * from));
 /* from legacy.c */
+extern char * serror _((void));
+extern char * serrorof _((int));
+extern int geterrno _((void));
 extern int portable_getpgrp _((int mypid));
 
 /* memory allocation */
@@ -730,10 +814,6 @@ extern void free_space _((void *));
 extern LINE * alloc_header _((void));
 extern void free_header _((LINE *));
 
-/* command dispatch mapper */
-/* from mined1.c */
-extern voidfunc command _((unsigned long));
-
 
 /* prompt line */
 /* from prompt.c */
@@ -744,21 +824,6 @@ extern long get_number _((char * message, char firstdigit, int * result));
 extern int prompt_num_char _((unsigned long * result, unsigned long maxvalue));
 
 
-/* from edit.c */
-extern char * script _((unsigned long ucs));
-extern unsigned long decomposition_base _((unsigned long ucs));
-extern char * category _((unsigned long ucs));
-extern FLAG isLetter _((unsigned long ucs));
-extern FLAG is_wideunichar _((unsigned long ucs));
-extern FLAG iscombining_unichar _((unsigned long ucs));
-extern FLAG isspacingcombining_unichar _((unsigned long ucs));
-extern FLAG iscombined_unichar _((unsigned long ucs, char * charpos, char * linebegin));
-extern unsigned long case_convert _((unsigned long unichar, int dir));
-/* from charcode.c */
-extern int iscontrol _((unsigned long));
-extern int iswhitespace _((unsigned long));
-extern character controlchar _((character));
-
 /* scrollbar handling */
 /* from output.c */
 extern FLAG display_scrollbar _((FLAG update));
@@ -766,9 +831,6 @@ extern void scrollbar_scroll_up _((int from));
 extern void scrollbar_scroll_down _((int from));
 extern int check_scrollbar_pos _((void));
 
-/* output functions */
-extern void clear_filebuf _((void));
-extern int flush_filebuf _((int fd));
 /* from mined1.c */
 extern void RDwin _((void));
 extern void RDwinchg _((void));
@@ -798,9 +860,19 @@ extern char * dupstr _((char * s));
 extern char * nextutf8 _((char *s));
 extern int stringcompare_ _((char * s1, char * s2));
 extern int stringprefix_ _((char * s1, char * s2));
-/* from mined1.c */
+/* from minedaux.c */
+extern int UTF8_len _((char));
+extern int CJK_len _((character *));
 extern int char_count _((char *));
 extern int col_count _((char *));
+extern void utf8_info _((char *, int *, unsigned long *));
+extern int isjoined _((unsigned long, char *, char *));
+extern int iscombined _((unsigned long, char *, char *));
+#define multichar(c)	((character) c >= 0x80 && (cjk_text == False || text_encoding_tag != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
+extern int iscombining _((unsigned long ucs));
+extern int iswide _((unsigned long ucs));
+extern int uniscrwidth _((unsigned long, char *, char *));
+extern int cjkscrwidth _((unsigned long, char *, char *));
 extern char * charbegin _((char *, char *));
 extern void advance_utf8_scr _((char * *, int *, char *));
 extern void advance_utf8 _((char * *));
@@ -812,54 +884,17 @@ extern unsigned long charvalue _((character *));
 extern unsigned long unicodevalue _((character *));
 extern unsigned long unicode _((unsigned long));
 extern unsigned long precedingchar _((char *, char *));
-extern int uniscrwidth _((unsigned long, char *, char *));
-extern int cjkscrwidth _((unsigned long, char *, char *));
-extern int UTF8_len _((char));
-extern int CJK_len _((character *));
-extern void utf8_info _((char *, int *, unsigned long *));
-extern int isjoined _((unsigned long, char *, char *));
-extern int iscombined _((unsigned long, char *, char *));
-#define multichar(c)	((character) c >= 0x80 && (cjk_text == False || text_encoding_tag != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
-extern int iscombining _((unsigned long ucs));
-extern int iswide _((unsigned long ucs));
+/* from mined1.c */
 extern int isglyph_uni _((unsigned long));
 
 /* from isrtol.c */
 extern int is_right_to_left _((unsigned long ucs));
 
-/* from charcode.c */
-extern int utfencode _((unsigned long, character *));
-extern int cjkencode _((unsigned long, character *));
-extern int cjkencode_char _((FLAG term, unsigned long, character *));
-extern char * encode_char _((unsigned long));
-extern FLAG valid_cjk _((unsigned long cjkchar, /* opt */ character * cjkbytes));
-extern FLAG valid_cjkchar _((FLAG term, unsigned long cjkchar, /* opt */ character * cjkbytes));
-extern unsigned long isolated_alef _((unsigned long));
-extern unsigned long ligature_lam_alef _((unsigned long));
-extern FLAG no_char _((unsigned long c));
-extern FLAG no_unichar _((unsigned long u));
-
 /* character mapping handling */
 #define CHAR_UNKNOWN (unsigned long) -2	/* unknown mnemonic */
 #define CHAR_INVALID (unsigned long) -1	/* not mappable to encoding */
 
 #define FUNcmd (unsigned long) -7	/* indicates function key */
-/* from charcode.c */
-extern unsigned long mappedtermchar _((unsigned long));
-extern unsigned long lookup_mappedtermchar _((unsigned long));
-extern FLAG remap_chars _((void));
-extern unsigned long encodedchar _((unsigned long));
-extern unsigned long encodedchar2 _((unsigned long, unsigned long));
-extern unsigned long lookup_encodedchar _((unsigned long));
-extern unsigned long max_char_value _((void));
-extern FLAG encoding_has_combining _((void));
-extern char * get_text_encoding _((void));
-extern char * get_term_encoding _((void));
-extern FLAG set_text_encoding _((char * charmap, char tag, char * debug_tag));
-extern FLAG set_term_encoding _((char * charmap, char tag));
-extern char text_encoding_tag;
-extern char * text_encoding_flag;
-extern char term_encoding_tag;
 
 /* information display preferences */
 extern FLAG always_disp_fstat;	/* Permanent file status display on status line? */
@@ -893,32 +928,20 @@ extern struct raw_hanentry hantable [];
 extern unsigned int hantable_len;
 
 
-/* keyboard mapping tables */
-#ifndef NOSTRINGTABLES
-#define keymapping_stringtables
-#endif
-
-#ifdef keymapping_stringtables
-#define keymaptabletype char *
-#else
-#define keymaptabletype struct keymap *
-struct keymap {
-	char * fk;
-	char * fp;
-};
-#endif
-
-extern struct keymapping {
-	keymaptabletype table;
-	char * shortcut;
-} keymappingtable [];
-extern unsigned int keymappingtable_len;
-
-
 /* various keyboard options */
-extern FLAG mined_keypad;	/* Apply mined keypad assignments */
 extern FLAG detect_esc_alt;	/* Enable detection of Alt key by ESC prefix? */
 
+/* keyboard mapping setup functions (keymaps.c) */
+extern void toggleKEYMAP _((void));
+extern void setupKEYMAP _((void));
+extern FLAG setKEYMAP _((char *));
+extern char * keyboard_mapping;
+extern char * last_keyboard_mapping;
+
+/* keyboard mapping (keymaps.c) */
+extern int map_key _((char * str, int matchmode, char * * found, char * * mapped));
+extern int keyboard_mapping_active _((void));
+
 /* keyboard mapping menu handling */
 extern char selection_space;
 extern FLAG enforce_keymap;	/* enable keyboard mapping even on non-suitable terminal */
@@ -949,7 +972,7 @@ extern int popup_menu _((menuitemtype *, int menulen, int column, int line,
 			FLAG is_flag_menu, FLAG disp_only, char * select_keys));
 extern void clean_menus _((void));
 extern int is_menu_open _((void));
-extern void displaymenuline _((void));
+extern void displaymenuline _((FLAG show_filelist));
 extern void redrawmenu _((void));
 extern void displayflags _((void));
 extern void toggle_append _((void));
@@ -959,19 +982,23 @@ extern void handleEncodingmenu _((void));
 
 /* from edit.c */
 extern int quote_type;
-extern void quote_type_down _((void));
-extern void quote_type_up _((void));
+extern int prev_quote_type;
+extern int default_quote_type;
+extern FLAG spacing_quotes;	/* French quote marks spacing? */
 extern void set_quote_type _((int));
 extern void set_quote_style _((char *));
 extern void reset_smart_replacement _((void));
 /* from mousemen.c */
 extern int count_quote_types _((void));
-extern char * quote_mark _((int, int));
+extern FLAG spacing_quote_type _((int qt));
+extern char * quote_mark _((int, quoteposition_type));
+extern void set_quote_menu _((void));
 extern int lookup_quotes _((char *));
 extern menuitemtype * lookup_Keymap_menuitem _((char * hopitem));
 
 
 /* main editing functions */
+extern voidfunc command _((unsigned long));
 extern void invoke_key_function _((unsigned long));
 
 extern void QUICKMENU _((void)), FILEMENU _((void)), EDITMENU _((void));
@@ -981,13 +1008,12 @@ extern void MOUSEfunction _((void)), FOCUSout _((void)), FOCUSin _((void));
 extern void AMBIGnarrow _((void)), AMBIGwide _((void));
 extern void ANSIseq _((void));
 extern void OSC _((void));
-extern void MUP _((void)), MDN _((void)), MLF _((void)), MRT _((void));
-extern void MNW _((void)), MPW _((void));
-extern void BSEN _((void)), ESEN _((void));
-extern void SD _((void)), SU _((void)), PD _((void)), PU _((void));
+extern void MOVUP _((void)), MOVDN _((void)), MOVLF _((void)), MOVRT _((void));
+extern void SD _((void)), SU _((void)), MOVPD _((void)), MOVPU _((void));
 extern void BFILE _((void)), EFILE _((void));
-extern void BLINE _((void)), ELINE _((void)), GO_LINEBEGIN _((void)), GO_LINEEND _((void));
+extern void BLINEORUP _((void)), ELINEORDN _((void)), MOVLBEG _((void)), MOVLEND _((void));
 extern void HIGH _((void)), LOW _((void));
+extern void MNW _((void)), MPW _((void)), BSEN _((void)), ESEN _((void));
 extern void MPPARA _((void)), MNPARA _((void));
 extern void DPC _((void)), DCC _((void)), DLN _((void)), DNW _((void)), DPW _((void));
 extern void DCC0 _((void)), DPC0 _((void));
@@ -1001,20 +1027,23 @@ extern void RD _((void)), RDwin _((void)), RDwin_nomenu _((void)), I _((void));
 extern void EXED _((void)), VIEW _((void));
 extern void HOP _((void)), CANCEL _((void)), ESCAPE _((void));
 extern void EMAX _((void)), META _((void)), UNDO _((void));
-extern void CHDI _((void)), PRINT _((void)), SH _((void)), SUSP _((void));
+extern void CHDI _((void));
+extern void PRINT _((void)), CMD _((void)), SH _((void)), SUSP _((void));
 extern void LNCI _((void)), LNSW _((void));
 extern void LOWCAP _((void)), LOWER _((void)), UPPER _((void));
 extern void CAPWORD _((void)), LOWCAPWORD _((void));
 extern void HTML _((void)), MARKER _((void)), GOMARKER _((void));
 extern void EDITmode _((void)), VIEWmode _((void)), toggle_VIEWmode _((void));
-extern void NXTFILE _((void)), PRVFILE _((void)), NTHFILE _((void)), EXFILE _((void)), EXMINED _((void));
+extern void NXTFILE _((void)), PRVFILE _((void)), SELECTFILE _((void)), CLOSEFILE _((void));
+extern void EXFILE _((void)), EXMINED _((void));
 extern void ADJLM _((void)), ADJFLM _((void)), ADJNLM _((void)), ADJRM _((void)), ADJPAGELEN _((void));
 extern void toggle_tabsize _((void)), toggle_tab_expansion _((void));
 extern void changeuni _((void)), changehex _((void)), changeoct _((void)), changedec _((void));
-extern void screensmaller _((void)), screenbigger _((void)), screenlesslines _((void)), screenmorelines _((void));
+extern void screensmaller _((void)), screenlesslines _((void)), screenlesscols _((void));
+extern void screenbigger _((void)), screenmorelines _((void)), screenmorecols _((void));
 extern void fontsmaller _((void)), fontbigger _((void)), fontdefault _((void));
 
-extern void SFW _((void)), SRV _((void)), RS _((void)), RSreverse _((void));
+extern void SFW _((void)), SRV _((void)), RES _((void)), RESreverse _((void));
 extern void LR _((void)), GR _((void)), REPL _((void));
 extern void SIDFW _((void)), SIDRV _((void));
 extern void Stag _((void));
@@ -1027,7 +1056,7 @@ extern void change_encoding _((char * new_text_encoding));
 extern void S0 _((character)), Scharacter _((unsigned long));
 extern FLAG insert_unichar _((unsigned long));
 extern void SNL _((void)), STAB _((void)), SSPACE _((void));
-extern void LIB _((void)), Underline _((void));
+extern void APPNL _((void)), Underline _((void));
 extern void Sdoublequote _((void)), Ssinglequote _((void)), Sdash _((void)), Sacute _((void)), Sgrave _((void));
 extern void SCURCHARFW _((void)), SCURCHARRV _((void));
 
@@ -1063,6 +1092,8 @@ extern void FHELP _((voidfunc));
 extern void check_file_modified _((void));
 
 /* function keys */
+extern void LFkey _((void)), RTkey _((void)), UPkey _((void)), DNkey _((void));
+extern void PGUPkey _((void)), PGDNkey _((void));
 extern void HOMEkey _((void)), ENDkey _((void)), smallHOMEkey _((void)), smallENDkey _((void));
 extern void INSkey _((void)), DELkey _((void)), smallINSkey _((void)), smallDELkey _((void));
 extern void KP_plus _((void)), KP_minus _((void));
