@@ -4,6 +4,7 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "charprop.h"
 #include "io.h"
 #include "termprop.h"
 
@@ -424,7 +425,7 @@ void
 clearscreen ()
 {
   clear_screen ();
-  top_line_scrolled = True;
+  top_line_dirty = True;
   set_scrollbar_dirty (0);
   set_scrollbar_dirty (SCREENMAX - 1);
 }
@@ -441,38 +442,44 @@ clearscreen ()
 		disp_colour ();
 		disp_normal ();
 	disp_syntax ();
+		dispXML_value ();
+		dispXML_attrib ();
 		dispHTML_jsp ();
 		dispHTML_comment ();
 		dispHTML_code ();
-		dispHTML_off ();
 	reverse_on/off (for selection highlighting)
  */
 
 #define dont_debug_attr
 
 /* screen attributes are listed in order of precedence (increasing) */
-#define attr_html	0x001
-#define attr_comment	0x002
-#define attr_jsp	0x004
+#define attr_html	0x0001
+#define attr_jsp	0x0002
+#define attr_comment	0x0004
+#define attr_attrib	0x0008
+#define attr_value	0x0010
 
-#define attr_script	0x008
+#define attr_script	0x0020
 
-#define attr_uni	0x010
-#define attr_special	0x020
-#define attr_mark	0x040
-#define attr_combining	0x080
-#define attr_control	0x100
+#define attr_uni	0x0040
+#define attr_special	0x0080
+#define attr_mark	0x0100
+#define attr_combining	0x0200
+#define attr_control	0x0400
 
-#define attr_selected	0x200
+#define attr_keyword	0x0800
 
-#define attr_all	0xFFF
+#define attr_selected	0x1000
+
+#define attr_all	0x1FFF
 
 static unsigned short attr = 0;
 static int attr_colour = 0;
 
 static
 void
-do_set_attr (unsigned short set)
+do_set_attr (set)
+  unsigned short set;
 {
 #ifdef debug_attr
   printf ("do_set_attr %03X\n", set);
@@ -494,8 +501,11 @@ do_set_attr (unsigned short set)
 	case attr_script:
 		disp_colour (attr_colour);
 		return;
-	case attr_html:
-		dispHTML_code ();
+	case attr_value:
+		dispXML_value ();
+		return;
+	case attr_attrib:
+		dispXML_attrib ();
 		return;
 	case attr_jsp:
 		dispHTML_jsp ();
@@ -503,6 +513,16 @@ do_set_attr (unsigned short set)
 	case attr_comment:
 		dispHTML_comment ();
 		return;
+	case attr_html:
+		dispHTML_code ();
+		return;
+	case attr_keyword:
+		if (strop_style == '_') {
+			underline_on ();
+		} else {
+			bold_on ();
+		}
+		return;
 	case attr_control:
 		if (attr & attr_selected) {
 			disp_normal ();
@@ -534,10 +554,16 @@ refresh_attrs ()
 	do_set_attr (attr_special);
   } else if (attr & attr_combining) {
 	do_set_attr (attr_combining);
+  } else if (attr & attr_keyword) {
+	do_set_attr (attr_keyword);
   } else if (attr & attr_control && ! (attr & attr_selected)) {
 	do_set_attr (attr_control);
   } else {
-	if (attr & attr_jsp) {
+	if (attr & attr_value) {
+		do_set_attr (attr_value);
+	} else if (attr & attr_attrib) {
+		do_set_attr (attr_attrib);
+	} else if (attr & attr_jsp) {
 		do_set_attr (attr_jsp);
 	} else if (attr & attr_comment) {
 		do_set_attr (attr_comment);
@@ -556,7 +582,8 @@ refresh_attrs ()
 
 static
 void
-clear_attr (unsigned short clear)
+clear_attr (clear)
+  unsigned short clear;
 {
 #ifdef debug_attr
   printf ("clear_attr %03X (%03X->%03X)\n", clear, attr, attr & ~ clear);
@@ -571,7 +598,8 @@ clear_attr (unsigned short clear)
 
 static
 void
-set_attr (unsigned short set)
+set_attr (set)
+  unsigned short set;
 {
 #ifdef debug_attr
   printf ("set_attr %03X (%03X->%03X)\n", set, attr, attr | set);
@@ -623,19 +651,21 @@ static struct {
 	{0x2713, 'V', False},	/* ✓ */
 	{0x2717, 'X', False},	/* ✗ */
 	{0x21AF, 'Z', False},	/* ↯ */
+	{0x2118, 'P', False},	/* ℘ */
 	{0x02BB, '\'', False}	/* glottal stop ʻokina */
 };
 
 static
 unsigned long
-disp_subst (unichar)
+disp_subst (unichar, check_glyph)
   unsigned long unichar;
+  FLAG check_glyph;
 {
-  if (! glyphs_checked) {
+  if (check_glyph && ! glyphs_checked) {
 	glyphs_available = isglyph_uni (0);
 	glyphs_checked = True;
   }
-  if (glyphs_available) {
+  if (glyphs_available || ! check_glyph) {
 	int i;
 	for (i = 0; i < arrlen (glyphs_subst); i ++) {
 		if (glyphs_subst [i].unichar == unichar) {
@@ -1051,45 +1081,6 @@ static int utfcount = 0;
 static unsigned long unichar;
 static unsigned short prev_cjkchar = 0L;
 
-struct interval {
-    unsigned long first;
-    unsigned long last;
-};
-
-/*
-static struct interval list_Quotation_Mark [] =
-static struct interval list_Dash [] =
-*/
-#include "charprop.t"
-
-static
-int
-lookup (ucs, table, length)
-  unsigned long ucs;
-  struct interval * table;
-  int length;
-{
-  int min = 0;
-  int mid;
-  int max = length - 1;
-
-  if (ucs < table [0].first || ucs > table [max].last) {
-	return 0;
-  }
-  while (max >= min) {
-	mid = (min + max) / 2;
-	if (ucs > table [mid].last) {
-		min = mid + 1;
-	} else if (ucs < table [mid].first) {
-		max = mid - 1;
-	} else {
-		return 1;
-	}
-  }
-
-  return 0;
-}
-
 #define dont_debug_put_cjk
 #ifdef debug_put_cjk
 #define trace_put_cjk(tag, c)	printf ("put_cjkchar %04X %s\n", c, tag)
@@ -1425,9 +1416,10 @@ put_unichar (unichar)
 		}
 	}
   } else {	/* not utf8_screen */
-	unsigned long termchar;
 	int iswide_unichar;
+	/* only used if cjk_term || mapped_term: */
 	static char * dummyline = "";
+	unsigned long termchar = 0;	/* (avoid -Wmaybe-uninitialized) */
 
 	if (cjk_term || mapped_term) {
 		termchar = mappedtermchar (unichar);
@@ -1473,22 +1465,10 @@ put_unichar (unichar)
 		/*indicate_narrow ('\'');*/
 		indicate_narrow ('`');
 	} else {
-		if (unichar == 0x20AC) {
-			/* EURO SIGN */
-			indicate_narrow ('E');
-		} else if (unichar == 0x2713) {
-			/* ✓ */
-			indicate_narrow ('V');
-		} else if (unichar == 0x2717) {
-			/* ✗ */
-			indicate_narrow ('X');
-		} else if (unichar == 0x21AF) {
-			/* ↯ */
-			indicate_narrow ('Z');
-		} else if (unichar == 0x02BB) {
-			/* glottal stop ʻokina */
-			indicate_narrow ('\'');
-		} else if (lookup (unichar, list_Quotation_Mark, arrlen (list_Quotation_Mark))) {
+		unsigned long substchar = disp_subst (unichar, False);
+		if (substchar != 0) {
+			indicate_narrow (substchar);
+		} else if (isquotationmark (unichar)) {
 			if (unichar == 0xAB || unichar == 0x2039) {
 				indicate_narrow ('<');
 			} else if (unichar == 0xBB || unichar == 0x203A) {
@@ -1502,7 +1482,7 @@ put_unichar (unichar)
 			} else {
 				indicate_narrow ('"');
 			}
-		} else if (lookup (unichar, list_Dash, arrlen (list_Dash))) {
+		} else if (isdash (unichar)) {
 			if (unichar == 0x301C || unichar == 0x3030) {
 				indicate_narrow ('~');
 			} else if (unichar == 0x2E17 || unichar == 0x30A0) {
@@ -1729,7 +1709,17 @@ disp_syntax (prev_syntax_mask, syntax_mask)
 	char new_syntax = syntax_mask & ~ prev_syntax_mask;
 	char old_syntax = prev_syntax_mask & ~ syntax_mask;
 
-	if (old_syntax & syntax_JSP) {
+	if (viewing_help) {
+		if (old_syntax & syntax_comment) {
+			clear_attr (attr_jsp);
+		}
+		if (new_syntax & syntax_comment) {
+			set_attr (attr_jsp);
+		}
+		return;
+	}
+
+	if (old_syntax & syntax_scripting) {
 		clear_attr (attr_jsp);
 	}
 	if (old_syntax & syntax_comment) {
@@ -1738,6 +1728,12 @@ disp_syntax (prev_syntax_mask, syntax_mask)
 	if (old_syntax & syntax_HTML) {
 		clear_attr (attr_html);
 	}
+	if (old_syntax & syntax_attrib) {
+		clear_attr (attr_attrib);
+	}
+	if (old_syntax & syntax_value) {
+		clear_attr (attr_value);
+	}
 
 	if (new_syntax & syntax_HTML) {
 		set_attr (attr_html);
@@ -1745,9 +1741,15 @@ disp_syntax (prev_syntax_mask, syntax_mask)
 	if (new_syntax & syntax_comment) {
 		set_attr (attr_comment);
 	}
-	if (new_syntax & syntax_JSP) {
+	if (new_syntax & syntax_scripting) {
 		set_attr (attr_jsp);
 	}
+	if (new_syntax & syntax_attrib) {
+		set_attr (attr_attrib);
+	}
+	if (new_syntax & syntax_value) {
+		set_attr (attr_value);
+	}
   }
 }
 
@@ -1814,7 +1816,7 @@ void
 restore_attr ()
 {
   if (mark_HTML) {
-	disp_syntax (0, syntax_mask);
+	disp_syntax (syntax_none, syntax_mask);
   }
   if (current_script_colour > 0) {
 	disp_script (current_script_colour);
@@ -1845,6 +1847,21 @@ print_line (ly, line)
   line->dirty = False;
 }
 
+static
+char *
+after_password (tp)
+  char * tp;
+{
+  char * afterpw = strcontains (tp, "assword");
+  if (afterpw) {
+	afterpw += 7;
+	while (* afterpw == ':' || * afterpw == '=' || white_space (* afterpw)) {
+		afterpw ++;
+	}
+  }
+  return afterpw;
+}
+
 /*
  * Put_line prints the given line on the standard output.
  * If offset is not zero, printing will start at that x-coordinate.
@@ -1871,8 +1888,9 @@ put_line (ly, line, offset, clear_line, positioning)
   FLAG separate_combining = True;	/* at line beginning, after TAB */
   FLAG save_combining_mode = combining_mode;
 
-  char syntax_mask_old;
-  char syntax_mask_new;
+  /* only used if mark_HTML: */
+  char syntax_mask_old = syntax_none;
+  char syntax_mask_new = syntax_none;	/* avoid -Wmaybe-uninitialized */
 
   FLAG sel_highlighting = False;
 
@@ -1880,28 +1898,35 @@ put_line (ly, line, offset, clear_line, positioning)
 
   int line_script_colour = no_script;
 
+  int qn = 0;	/* quote count for heuristic string context */
+  FLAG do_strop = UNSURE;
+  unsigned long small_unichar = 0;	/* if (do_strop); avoid -Wmaybe-uninitialized */
+  unsigned long prev_unichar = 0;
+
+#define dont_debug_syntax_state
+#ifdef debug_syntax_state
+  printf ("put_line @%d (%02X/%02X) @%s", offset, line->prev->syntax_mask, line->syntax_mask, line->text);
+#endif
+
   clear_attr (attr_all);
 
-  if (hide_password) {
-	password = strcontains (textp, "assword");
-	if (password) {
-		password += 7;
-		while (* password == ':' || * password == '=' || white_space (* password)) {
-			password ++;
-		}
-	}
+  if (hide_passwords) {
+	password = after_password (textp);
   }
 
   if (mark_HTML) {
 	syntax_mask = line->prev->syntax_mask;
 	syntax_mask_old = syntax_none;
 	syntax_mask_new = syntax_mask;
+#ifdef highlight_complete_line_for_uni_lineend
+	/* renders line unreadable in dark terminals */
 	switch (line->return_type) {
 		case lineend_LS:
 		case lineend_PS:
 			line_script_colour = 6;
 			break;
 	}
+#endif
   }
   current_script_colour = line_script_colour;
 
@@ -1915,13 +1940,38 @@ put_line (ly, line, offset, clear_line, positioning)
 
 /* Skip all chars as indicated by the offset_start and the shift_count field */
   while (count < offset_start) {
+	/* ... and perform all hidden state updates */
 	if (mark_HTML) {
 		syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
-		if (* textp == '<' || * textp == '>') {
-			syntax_mask_old = syntax_mask;
-			syntax_mask = syntax_mask_new;
-		}
+		syntax_mask_old = syntax_mask;
+		syntax_mask = syntax_mask_new;
 	}
+	if (dispunstrop) {
+		unichar = unicodevalue (textp);
+		if ((unichar == '"' || unichar == '\'')
+		    && (qn % 2 == 0 || prev_unichar != '\\')) {
+			qn ++;
+			do_strop = UNSURE;
+		} else if (qn % 2 == 0) {
+			if (isLetter (unichar)) {
+				small_unichar = case_convert (unichar, -1);
+				if (small_unichar == unichar) {
+					/* small letter */
+					do_strop = False;
+				} else if (do_strop) {	/* True or UNSURE */
+					/* capital not preceded by small */
+					do_strop = True;
+				}
+			} else {
+				do_strop = UNSURE;
+			}
+		}
+		prev_unichar = unichar;
+	}
+	if (password && textp >= password && white_space (* textp)) {
+		password = after_password (textp);
+	}
+
 	advance_char_scr (& textp, & count, line->text);
   }
 
@@ -1944,10 +1994,8 @@ put_line (ly, line, offset, clear_line, positioning)
 			if (! is_tab (* textp)) {
 				if (mark_HTML) {
 					syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
-					if (* textp == '<' || * textp == '>') {
-						syntax_mask_old = syntax_mask;
-						syntax_mask = syntax_mask_new;
-					}
+					syntax_mask_old = syntax_mask;
+					syntax_mask = syntax_mask_new;
 				}
 				if (iswide (unicodevalue (textp))) {
 					/* skipped character is double-width, 
@@ -1983,10 +2031,15 @@ put_line (ly, line, offset, clear_line, positioning)
 	separate_combining = False;
   }
 
+  if (show_vt100_graph) {
+	altcset_on ();
+  }
+
   while (True) {
 	unsigned long unichar;
 	int follow;
-	unsigned long cjkchar;
+	/* only used if cjk_text: (avoid -Wmaybe-uninitialized) */
+	unsigned long cjkchar = 0;
 
 	if (separate_isolated_combinings) {
 		if (separate_combining) {
@@ -2048,15 +2101,35 @@ put_line (ly, line, offset, clear_line, positioning)
 
 	if (mark_HTML) {
 		syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
-		if (* textp == '<' || * textp == '>') {
-			syntax_mask_old = syntax_mask;
-			syntax_mask = syntax_mask_new;
-		}
-		if (* textp == '<') {
+		syntax_mask_old = syntax_mask;
+		syntax_mask = syntax_mask_new;
+		if (syntax_mask != syntax_mask_old) {
 			disp_syntax (syntax_mask_old, syntax_mask);
 		}
 	}
 
+	if (dispunstrop) {
+		if ((unichar == '"' || unichar == '\'')
+		    && (qn % 2 == 0 || prev_unichar != '\\')) {
+			qn ++;
+			do_strop = UNSURE;
+		} else if (qn % 2 == 0) {
+			if (isLetter (unichar)) {
+				small_unichar = case_convert (unichar, -1);
+				if (small_unichar == unichar) {
+					/* small letter */
+					do_strop = False;
+				} else if (do_strop) {	/* True or UNSURE */
+					/* capital not preceded by small */
+					do_strop = True;
+				}
+			} else {
+				do_strop = UNSURE;
+			}
+		}
+		prev_unichar = unichar;
+	}
+
 	if (is_tab (* textp)) {		/* Expand tabs to spaces */
 		int tab_count = tab (count);
 		int tab_mid = count + (tab_count - count) / 2 + 1;
@@ -2120,14 +2193,7 @@ put_line (ly, line, offset, clear_line, positioning)
 		clear_attr (attr_control);
 		advance_char (& textp);
 		if (white_space (* textp)) {
-			/*password = NIL_PTR;*/
-			password = strcontains (textp, "assword");
-			if (password) {
-				password += 7;
-				while (* password == ':' || white_space (* password)) {
-					password ++;
-				}
-			}
+			password = after_password (textp);
 		}
 		count += charwidth;
 	} else if (unichar == 0xA0 /* nbsp */) {
@@ -2139,7 +2205,7 @@ put_line (ly, line, offset, clear_line, positioning)
 
 		advance_char (& textp);
 		count += charwidth;
-	} else if (iscontrol (unichar)) {
+	} else if (iscontrol (* textp)) {
 		/* this covers only 1 byte controls; 
 		   GB18030 "C1" controls are handled elsewhere
 		 */
@@ -2165,13 +2231,11 @@ put_line (ly, line, offset, clear_line, positioning)
 		FLAG disp_separate_combining = False;
 		FLAG disp_separate_joining = False;
 		FLAG disp_joined = False;
-		char syntax_byte_printed;
-
-		syntax_byte_printed = * textp;
 
 		if (utf8_text) {
 			int last_script_colour;
 			int script_colour0;
+			unsigned long substchar;
 
 			follow --;
 
@@ -2269,14 +2333,23 @@ put_line (ly, line, offset, clear_line, positioning)
 					follow --;
 				}
 #endif
-			} else if ((unichar = disp_subst (unichar))) {
-				indicate_uni (unichar, True, attr_uni);
+			} else if ((substchar = disp_subst (unichar, True)) != 0) {
+				indicate_uni (substchar, True, attr_uni);
 				/* skip UTF-8 char */
 				textp ++;
 				while (follow > 0 && (* textp & 0xC0) == 0x80) {
 					textp ++;
 					follow --;
 				}
+			} else if (do_strop == True) {
+				set_attr (attr_keyword);
+				put_unichar (small_unichar);
+				clear_attr (attr_keyword);
+				textp ++;
+				while (follow > 0 && (* textp & 0xC0) == 0x80) {
+					textp ++;
+					follow --;
+				}
 			} else {
 				put_char (* textp ++);
 				while (follow > 0 && (* textp & 0xC0) == 0x80) {
@@ -2290,7 +2363,18 @@ put_line (ly, line, offset, clear_line, positioning)
 			}
 
 		} else if (cjk_text) {
-			if (multichar (* textp)) {
+			if (do_strop == True) {
+				int cjklen = CJK_len (textp);
+				set_attr (attr_keyword);
+				put_unichar (small_unichar);
+				clear_attr (attr_keyword);
+
+				while (cjklen > 0 && * textp != '\n' && * textp != '\0') {
+					textp ++;
+					cjklen --;
+				}
+				count += charwidth;
+			} else if (multichar (* textp)) {
 				int charlen;
 				character cjkbytes [5];
 				int cjklen = CJK_len (textp);
@@ -2396,7 +2480,11 @@ put_line (ly, line, offset, clear_line, positioning)
 				}
 			}
 
-			if (mapped_text && unichar == 0xA0 /* nbsp */) {
+			if (do_strop == True) {
+				set_attr (attr_keyword);
+				put_unichar (small_unichar);
+				clear_attr (attr_keyword);
+			} else if (mapped_text && unichar == 0xA0 /* nbsp */) {
 				indicate_special ('');
 			} else {
 				if (disp_joined) {
@@ -2424,12 +2512,16 @@ put_line (ly, line, offset, clear_line, positioning)
 			count += charwidth;
 		}
 
-		if (mark_HTML && syntax_byte_printed == '>') {
+		if (mark_HTML && syntax_mask != syntax_mask_old) {
 			disp_syntax (syntax_mask_old, syntax_mask);
 		}
 	}
   }
 
+  if (show_vt100_graph) {
+	altcset_off ();
+  }
+
   if (separate_isolated_combinings) {
 	combining_mode = save_combining_mode;
   }
@@ -2441,8 +2533,10 @@ put_line (ly, line, offset, clear_line, positioning)
   }
 
   if (mark_HTML) {
-	/* now reset HTML display attribute */
-	disp_syntax (syntax_mask, 0);
+	/* cancel syntax future state */
+	(void) syntax_state (syntax_mask_new, syntax_mask_old, "", line->text);
+	/* reset HTML display attribute */
+	disp_syntax (syntax_mask, syntax_none);
   }
 
   if (positioning) {
@@ -2638,6 +2732,8 @@ set_cursor_xy ()
 |*			Text screen display				*|
 \*======================================================================*/
 
+#define dont_debug_display
+
 /*
  * Display line at screen line y_pos if it lies between y_min and y_max.
  * If it is no text line (end of file), clear screen line.
@@ -2674,7 +2770,10 @@ display_line_at (y_pos, line, y_min, y_max, first, dirty_only)
 		return;
 	}
 
-/*printf ("display line @ %d\n", y_pos); display_delay = 50;*/
+#ifdef debug_display
+	printf ("display line @ %d\n", y_pos); display_delay = 50;
+#endif
+
 	if (first == False) {
 		if (display_delay >= 0) {
 			flush ();
@@ -2716,7 +2815,10 @@ do_display (y_coord, line, count, new_pos, dirty_only)
 {
   int y_max = y_coord + count;
   int y_off;
-/*printf ("display %d+%d @ %d (dirty_only %d)\n", y_coord, count, new_pos, dirty_only);*/
+
+#ifdef debug_display
+  printf ("display %d+%d @ %d (dirty_only %d)\n", y_coord, count, new_pos, dirty_only);
+#endif
 
 /* Find new startline if count is negative */
   if (count < 0) {
@@ -2770,6 +2872,7 @@ display (y_coord, line, count, new_pos)
   int new_pos;
 {
   do_display (y_coord, line, count, new_pos, False);
+  text_screen_dirty = False;
 }
 
 
