@@ -6,6 +6,7 @@
 #include "mined.h"
 #include "io.h"		/* for flush, set_cursor, highlight_selection */
 #include "textfile.h"
+#include "charprop.h"
 
 #include <errno.h>
 
@@ -21,6 +22,9 @@
 #ifdef __pcgcc__
 #undef linkyank
 #endif
+#ifdef __MINGW32__
+#undef linkyank
+#endif
 #ifdef __CYGWIN__
 #define linkyank
 #endif
@@ -44,6 +48,10 @@ static char * pasted_end_textp = NIL_PTR;
 int buffer_open_flag = 0;	/* Counter flag for the collective buffer */
 
 static FLAG highlight_selection = False;
+static FLAG alt_rectangular_mode = False;
+#define rectangular_paste_mode (rectangular_paste_flag != alt_rectangular_mode)
+
+static int last_sel_x = 0;	/* Last selection mouse column */
 
 
 /*======================================================================*\
@@ -55,16 +63,30 @@ static LINE * mark_line = NIL_LINE;		/* For marking position. */
 static char * mark_text = NIL_PTR;
 
 /* explicit markers */
-#define maxmarkers 10
-static LINE * mark_n_line [maxmarkers] = {
-	NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, 
-	NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE};
-static char * mark_n_text [maxmarkers] = {
-	NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, 
-	NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR};
+static struct {
+	LINE * line;
+	char * text;
+} marker_n [] = {
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+	{NIL_LINE, NIL_PTR},
+};
+#define maxmarkers arrlen (marker_n)
 
 /* implicit marker stack */
-#define markstacklen 10
 static struct {
 	LINE * line;
 	char * text;
@@ -72,7 +94,7 @@ static struct {
 	int lineno;
 	int pos;
 	FLAG detectCR;
-} mark_stack [markstacklen] = {
+} mark_stack [] = {
 	{NIL_LINE, NIL_PTR, NIL_PTR, -1, -1, False},
 	{NIL_LINE, NIL_PTR, NIL_PTR, -1, -1, False},
 	{NIL_LINE, NIL_PTR, NIL_PTR, -1, -1, False},
@@ -84,6 +106,7 @@ static struct {
 	{NIL_LINE, NIL_PTR, NIL_PTR, -1, -1, False},
 	{NIL_LINE, NIL_PTR, NIL_PTR, -1, -1, False},
 };
+#define markstacklen arrlen (mark_stack)
 static int mark_stack_poi = 0;
 static int mark_stack_top = 0;
 static int mark_stack_bottom = 0;
@@ -113,6 +136,7 @@ copied_file_name ()
   for (i = 0; i < markstacklen; i ++) {
 	filei = mark_stack [i].file;
 	if (filei != NIL_PTR && streq (filei, file_name)) {
+		/* reuse the copy; note: it's not being freed ! */
 		return filei;
 	}
   }
@@ -129,17 +153,23 @@ copied_file_name ()
    Setmark sets the current position into the current slot of the mark stack.
  */
 static
-void
+FLAG
 Setmark ()
 {
+  char * fn = copied_file_name ();
+  if (fn) {
 	mark_stack [mark_stack_poi].line = cur_line;
 	mark_stack [mark_stack_poi].text = cur_text;
-	mark_stack [mark_stack_poi].file = copied_file_name ();
+	mark_stack [mark_stack_poi].file = fn;
 	mark_stack [mark_stack_poi].lineno = line_number;
 	mark_stack [mark_stack_poi].pos = get_cur_pos ();
 	mark_stack [mark_stack_poi].detectCR = lineends_detectCR;
 
 	printf_mark_stack ("Filled");
+	return True;
+  } else {
+	return False;
+  }
 }
 
 /*
@@ -148,12 +178,12 @@ Setmark ()
 void
 Pushmark ()
 {
-	/* truncate stack to current position */
-	mark_stack_top = mark_stack_poi;
-	mark_stack_count_top = mark_stack_count_poi;
+  /* truncate stack to current position */
+  mark_stack_top = mark_stack_poi;
+  mark_stack_count_top = mark_stack_count_poi;
 
-	/* fill data */
-	Setmark ();
+  /* fill data */
+  if (Setmark ()) {
 
 	/* increase top of stack */
 	mark_stack_top = (mark_stack_top + 1) % markstacklen;
@@ -169,6 +199,11 @@ Pushmark ()
 	mark_stack_count_poi = mark_stack_count_top;
 
 	printf_mark_stack ("Pushed");
+  } else {
+	ring_bell ();
+
+	printf_mark_stack ("Not Pushed - no mem");
+  }
 }
 
 /*
@@ -187,79 +222,94 @@ Upmark ()
 void
 Popmark ()
 {
-	FLAG switch_files;
+  FLAG switch_files;
 
-	if (hop_flag > 0) {
-		/* climb up stack towards top */
-		printf_mark_stack ("HOP Pop");
-		if (mark_stack_count_poi == mark_stack_count_top) {
-			printf_debug_mark_stack ("HOP Pop no more\n");
-			error ("No more stacked positions");
-			return;
-		}
-		/* adjust current position */
-		Setmark ();
-
-		mark_stack_poi = (mark_stack_poi + 1) % markstacklen;
-		mark_stack_count_poi ++;
-		printf_mark_stack ("Up:");
-	} else {
-		/* climb down stack towards bottom */
-		if (mark_stack_count_poi == 0) {
-			printf_debug_mark_stack ("Pop no more\n");
-			error ("No more stacked positions");
-			return;
-		}
-		/* adjust current position */
-		Setmark ();
-
-		mark_stack_poi = (mark_stack_poi - 1) % markstacklen;
-		mark_stack_count_poi --;
-		printf_mark_stack ("Dn:");
-	}
-
-	if (mark_stack [mark_stack_poi].file == NIL_PTR) {
-		printf_debug_mark_stack ("not valid\n");
-		error ("Stacked position not valid");
+  if (hop_flag > 0) {
+	/* climb up stack towards top */
+	printf_mark_stack ("HOP Pop");
+	if (mark_stack_count_poi == mark_stack_count_top) {
+		printf_debug_mark_stack ("HOP Pop no more\n");
+		error ("No more stacked positions");
 		return;
 	}
-	switch_files = ! streq (mark_stack [mark_stack_poi].file, file_name);
-	if (switch_files ||
-		checkmark (mark_stack [mark_stack_poi].line, 
-			mark_stack [mark_stack_poi].text) == NOT_VALID)
-	{
-		int mark_lineno;
-		LINE * open_line;
+	/* adjust current position */
+	if (Setmark ()) {
+		mark_stack_poi = (mark_stack_poi + 1) % markstacklen;
+		mark_stack_count_poi ++;
 
-		if (switch_files) {
-			lineends_detectCR = mark_stack [mark_stack_poi].detectCR;
-			if (save_text_load_file (mark_stack [mark_stack_poi].file) == ERRORS) {
-				return;
-			}
-		}
-
-		mark_lineno = mark_stack [mark_stack_poi].lineno - 1;
-		open_line = proceed (header->next, mark_lineno);
-		if (open_line == tail) {
-			EFILE ();
-			error ("Stacked position not present anymore");
-		} else {
-			int mark_col = mark_stack [mark_stack_poi].pos;
-			int cur_pos = 0;
-			char * cpoi;
-
-			move_to (0, find_y (open_line));
-			cpoi = cur_line->text;
-			while (* cpoi != '\n' && cur_pos < mark_col) {
-				advance_char (& cpoi);
-				cur_pos ++;
-			}
-			move_address (cpoi, y);
-		}
+		printf_mark_stack ("Up:");
 	} else {
-		move_address (mark_stack [mark_stack_poi].text, 
-				find_y (mark_stack [mark_stack_poi].line));
+		ring_bell ();
+
+		printf_mark_stack ("Not Up: - no mem");
 	}
+  } else {
+	/* climb down stack towards bottom */
+	if (mark_stack_count_poi == 0) {
+		printf_debug_mark_stack ("Pop no more\n");
+		error ("No more stacked positions");
+		return;
+	}
+	/* adjust current position */
+	if (Setmark ()) {
+		if (mark_stack_poi == 0) {
+			mark_stack_poi = markstacklen;
+		}
+		mark_stack_poi --;
+		mark_stack_count_poi --;
+
+		printf_mark_stack ("Dn:");
+	} else {
+		ring_bell ();
+
+		printf_mark_stack ("Not Dn: - no mem");
+	}
+  }
+
+  if (mark_stack [mark_stack_poi].file == NIL_PTR) {
+	printf_debug_mark_stack ("not valid\n");
+	error ("Stacked position not valid");
+	return;
+  }
+  switch_files = ! streq (mark_stack [mark_stack_poi].file, file_name);
+  if (switch_files ||
+	checkmark (mark_stack [mark_stack_poi].line, 
+		mark_stack [mark_stack_poi].text) == NOT_VALID)
+  {
+	int mark_lineno;
+	LINE * open_line;
+
+	if (switch_files) {
+		lineends_detectCR = mark_stack [mark_stack_poi].detectCR;
+		if (save_text_load_file (mark_stack [mark_stack_poi].file) == ERRORS) {
+			return;
+		}
+	}
+
+	mark_lineno = mark_stack [mark_stack_poi].lineno - 1;
+	open_line = proceed (header->next, mark_lineno);
+	if (open_line == tail) {
+		EFILE ();
+		error ("Stacked position not present anymore");
+	} else {
+		int mark_col = mark_stack [mark_stack_poi].pos;
+		int cur_pos = 0;
+		char * cpoi;
+
+		move_to (0, find_y (open_line));
+		cpoi = cur_line->text;
+		while (* cpoi != '\n' && cur_pos < mark_col) {
+			advance_char (& cpoi);
+			cur_pos ++;
+		}
+		move_address (cpoi, y);
+	}
+  } else {
+	clear_highlight_selection ();
+
+	move_address (mark_stack [mark_stack_poi].text, 
+			find_y (mark_stack [mark_stack_poi].line));
+  }
 }
 
 
@@ -442,6 +492,33 @@ revert_yank_buf ()
 |*			Yank text handling				*|
 \*======================================================================*/
 
+#define dont_debug_rectangular_paste_area
+#define dont_debug_sel
+#define dont_debug_sel_text
+#define dont_debug_rectangular_paste_insert
+
+/*
+   Return text column position within line.
+ */
+static
+int
+get_text_col (line, textp, with_shift)
+  LINE * line;
+  char * textp;
+  FLAG with_shift;
+{
+  char * tp = line->text;
+  int col = 0;
+  if (with_shift) {
+	col = line->shift_count * SHIFT_SIZE;
+  }
+
+  while (tp != textp && * tp != '\n' && * tp != '\0') {
+	advance_char_scr (& tp, & col, line->text);
+  }
+  return col;
+}
+
 static
 char *
 text_at (line, colpoi, targcol)
@@ -465,6 +542,12 @@ text_at (line, colpoi, targcol)
   return prev_cpoi;
 }
 
+#ifdef debug_rectangular_paste_area
+#define trace_cols(tag)	printf ("(%s) %s->%d, %d\n", tag, line->text, start_col, end_col);
+#else
+#define trace_cols(tag)	
+#endif
+
 /*
    yank_text () puts copies text between start position and end position 
    into the buffer.
@@ -480,7 +563,7 @@ static
 void
 yank_text (fd, buf_status, 
 	start_line, start_textp, end_line, end_textp, 
-	do_remove, appending, do_rectangular_paste)
+	do_remove, appending, do_rectangular_paste, start_end_reversed)
   int fd;
   FLAG * buf_status;
   LINE * start_line;
@@ -490,13 +573,10 @@ yank_text (fd, buf_status,
   FLAG do_remove;	/* == DELETE if text should be deleted */
   FLAG appending;	/* == True if text should is appended to yank buffer */
   FLAG do_rectangular_paste;
+  FLAG start_end_reversed;
 {
   LINE * line = start_line;
   char * textp = start_textp;
-  char * line_startp;	/* start of line text for rectangular area (to delete) */
-  int start_col;	/* left border of rectangular area */
-  int end_col;		/* right border of rectangular area */
-  int col;		/* measure rectangular area */
   FLAG do_continue;
   FLAG at_eol = False;	/* end of rectangular area in line */
   long chars_written = 0L;	/* chars written to buffer this time */
@@ -504,6 +584,11 @@ yank_text (fd, buf_status,
   int lines_written = 0;	/* lines written to buffer this time */
   int return_len;
 
+  /* only used if do_rectangular_paste: (avoid -Wmaybe-uninitialized) */
+  char * line_startp = NIL_PTR;	/* start of line text for area to delete */
+  int start_col = 0;	/* left border of rectangular area */
+  int end_col = 0;	/* right border of rectangular area */
+  int col = 0;		/* measure rectangular area */
 
 /* Check file to hold buffer */
   if (fd == ERRORS) {
@@ -522,18 +607,34 @@ yank_text (fd, buf_status,
 
 /* Prepare */
   if (do_rectangular_paste) {
-	/* check block boundaries */
-	start_col = get_text_col (start_line, start_textp);
-	end_col = get_text_col (end_line, end_textp);
-	if (start_col == end_col) {
-		error ("Rectangular area empty");
-		return;
+	/* check left and right block boundaries */
+	start_col = get_text_col (start_line, start_textp, False);
+	end_col = get_text_col (end_line, end_textp, False);
+	trace_cols ("...");
+
+	/* adapt column to actual mouse position */
+	if (start_end_reversed == REVERSE) {
+		if (last_sel_x > start_col) {
+			start_col = last_sel_x;
+		}
+	} else {
+		if (last_sel_x > end_col) {
+			end_col = last_sel_x;
+		}
 	}
+	/* fix right-to-left selection */
 	if (start_col > end_col) {
 		int dum = end_col;
 		end_col = start_col;
 		start_col = dum;
 	}
+	trace_cols ("r-to-l fix");
+	/* reject empty copy */
+	if (start_col == end_col) {
+		error ("Rectangular area empty");
+		return;
+	}
+
 	line_startp = textp = text_at (line, & col, start_col);
 	do_continue = True;
   } else {
@@ -712,6 +813,11 @@ yank_text (fd, buf_status,
 		}
 		display (y, cur_line, bottom_y - top_y, y);
 	}
+	/* Fix position:
+	   avoid leaving the cursor aside the delection area;
+	   go to upper left corner of deleted block */
+	move_to (LINE_START, find_y (start_line));	/* unshift line */
+	move_to (start_col, y);
   }
 
 /* Flush the I/O buffer and close file */
@@ -748,7 +854,7 @@ yank_text (fd, buf_status,
   chars_saved += chars_written;
   lines_saved += lines_written;
 
-  build_string (text_buffer, "%s %s: lines:chars(bytes) %d:%ld(%ld) - Paste with %s/Insert", 
+  build_string (text_buffer, "%s %s: lines %d chars %ld (bytes %ld) - Paste with %s/Insert", 
 	(do_remove == DELETE) ?
 		appending ? "Cut/appended" : "Cut/moved"
 		: appending ? "Appended" : "Copied",
@@ -783,7 +889,7 @@ delete_text_buf (start_line, start_textp, end_line, end_textp)
 	set_buffer_open (False);
 	yank_text (yankfile (WRITE, True), & yank_status, 
 			start_line, start_textp, end_line, end_textp, 
-			DELETE, True, False);
+			DELETE, True, False, FORWARD);
   } else {
 	(void) delete_text (start_line, start_textp, end_line, end_textp);
   }
@@ -801,23 +907,23 @@ static
 int
 get_pasteline (fd, buffer, len)
   int fd;
-  register char buffer [MAX_CHARS];
+  char buffer [maxLINElen];
   int * len;
 {
   int ret = get_line (fd, buffer, len, False);
 
   if (ret == NO_INPUT || ret == ERRORS) {
-	return ERRORS;
+	return ret;
   }
 
   if (utf8_text || ! pastebuf_utf8) {
 	return ret;
   } else {
-	char nativebuf [2 * MAX_CHARS];
+	char nativebuf [2 * maxLINElen];
 	char * poi = buffer;
 	char * npoi = nativebuf;
 	unsigned long prev_uc = 0;
-	char * prev_npoi;
+	char * prev_npoi = npoi;
 
 	while (* poi) {
 		int ulen = UTF8_len (* poi);
@@ -861,7 +967,7 @@ get_pasteline (fd, buffer, len)
 	* npoi = '\0';
 
 	* len = strlen (nativebuf);
-	if (* len >= MAX_CHARS) {
+	if (* len >= maxLINElen) {
 		error ("Line too long in current encoding");
 		return ERRORS;
 	} else {
@@ -872,9 +978,7 @@ get_pasteline (fd, buffer, len)
 }
 
 
-#define dont_debug_rectangular_paste_mode
-
-#ifdef debug_rectangular_paste_mode
+#ifdef debug_rectangular_paste_insert
 static
 void
 print_str (s)
@@ -932,7 +1036,7 @@ trace_line (tag, line)
  * paste_file () inserts the contents of an opened file (as given by
    filedescriptor fd) at the current location.
    Call insert_file () only via paste_file ().
-   If the flag rectangular_paste_flag is True, a rectangular area will 
+   If the rectangular_paste_mode is True, a rectangular area will 
    be inserted, proceeding line by line.
    After the insertion, if stay_old_pos is True, the cursor remains at the
    start of the inserted text, if stay_old_pos is False, it is placed to
@@ -945,24 +1049,30 @@ insert_file (fd, stay_old_pos, from_text_file)
   FLAG stay_old_pos;
   FLAG from_text_file;	/* consider UTF-16 ? */
 {
-  char line_buffer [MAX_CHARS];	/* buffer for next line to insert */
+  char line_buffer [maxLINElen];	/* buffer for next line to insert */
   LINE * line = cur_line;
-  int line_count = total_lines;	/* determine # lines inserted */
+  int line_count = total_lines;		/* determine # lines inserted */
   LINE * page = cur_line;
   int ret;
   int len;
   lineend_type return_type;
-  FLAG do_rectangular_paste = rectangular_paste_flag;
+  /* copy rectangular option; this working flag will be reset at end-of-file */
+  FLAG do_rectangular_paste = rectangular_paste_mode;
   int paste_col = get_cur_col ();	/* for rectangular paste */
 
   reset_get_line (from_text_file);
   trace_rectangular_paste (("-------------------------------------------\n"));
 
+  pasted_end_line = NIL_LINE;
+
 /* Get the first line of text (might be ended with a '\n') */
   ret = get_pasteline (fd, line_buffer, & len);
-  if (ret == ERRORS) {
+  if (ret == NO_INPUT) {
 	/* empty file */
 	return;
+  } else if (ret == ERRORS) {
+	/* error message? */
+	return;
   }
 
 /* Adjust line end type if present */
@@ -988,7 +1098,6 @@ insert_file (fd, stay_old_pos, from_text_file)
   trace_line ("first<", line);
   trace_rectangular_paste (("insert <%s>\n", line_buffer));
   if (insert_text (line, cur_text, line_buffer) == ERRORS) {
-	pasted_end_line = NIL_LINE;
 	return;
   }
   trace_line ("first>", line);
@@ -1001,10 +1110,11 @@ insert_file (fd, stay_old_pos, from_text_file)
 
 
 /* Repeat getting lines (and inserting lines) until EOF is reached */
-  while (line != NIL_LINE
+  while (ret != ERRORS /*line != NIL_LINE*/
 	 && (ret = get_pasteline (fd, line_buffer, & len)) != ERRORS
 	 /* NO_LINE (line without newline) is handled separately
 	    unless in rectangular paste mode */
+	 && ret != NO_INPUT
 	 && (ret != NO_LINE || do_rectangular_paste)
 	)
   {
@@ -1029,7 +1139,13 @@ insert_file (fd, stay_old_pos, from_text_file)
 		char * nl;
 
 		if (line->next == tail) {
-			line = line_insert_after (line, "\n", 1, line->return_type);
+			LINE * new_line = line_insert_after (line, "\n", 1, line->return_type);
+			if (new_line == NIL_LINE) {
+				ret = ERRORS;
+				break;
+			} else {
+				line = new_line;
+			}
 			line = line->prev;
 			trace_rectangular_paste (("EOF: append line\n"));
 			trace_line ("eof++", line);
@@ -1053,7 +1169,7 @@ insert_file (fd, stay_old_pos, from_text_file)
 
 		if (inscol < paste_col) {
 			/* fill short line up to paste column */
-			char fill_buffer [2 * MAX_CHARS];
+			char fill_buffer [2 * maxLINElen];
 			char * fill_poi = fill_buffer;
 
 			char * reftext = line->prev->text;
@@ -1117,7 +1233,7 @@ insert_file (fd, stay_old_pos, from_text_file)
 			/* append fill buffer */
 			trace_rectangular_paste (("fill <%s>\n", fill_buffer));
 			if (insert_text (line, inspoi, fill_buffer) == ERRORS) {
-				line = NIL_LINE;
+				ret = ERRORS;
 				break;
 			}
 			trace_line ("fill>", line);
@@ -1132,13 +1248,14 @@ insert_file (fd, stay_old_pos, from_text_file)
 		len = inspoi - line->text;
 		trace_rectangular_paste (("insert_text <%s>\n", line_buffer));
 		if (insert_text (line, inspoi, line_buffer) == ERRORS) {
-			line = NIL_LINE;
+			ret = ERRORS;
 			break;
 		}
 		trace_line ("insert", line);
 		pasted_end_line = line;
 		pasted_end_textp = line->text + len + length_of (line_buffer);
 	} else {
+		LINE * new_line;
 		if (ret == SPLIT_LINE) {
 			return_type = lineend_NONE;
 		} else if (ret == NUL_LINE) {
@@ -1147,7 +1264,13 @@ insert_file (fd, stay_old_pos, from_text_file)
 			return_type = extract_lineend_type (line_buffer, len);
 		}
 		trace_rectangular_paste (("line_insert <%s>\n", line_buffer));
-		line = line_insert_after (line, line_buffer, len, return_type);
+		new_line = line_insert_after (line, line_buffer, len, return_type);
+		if (new_line == NIL_LINE) {
+			ret = ERRORS;
+			break;
+		} else {
+			line = new_line;
+		}
 		trace_line ("insert", line);
 	}
   }
@@ -1158,15 +1281,17 @@ insert_file (fd, stay_old_pos, from_text_file)
   line_count = total_lines - line_count;
 
 /* finish pasting, check last line */
-  if (line == NIL_LINE) {
+  if (ret == ERRORS /* || line == NIL_LINE*/) {
 	pasted_end_line = NIL_LINE;
 	/* show memory allocation error msg */
 	sleep (2);
   } else if (ret == NO_LINE && ! do_rectangular_paste) {
 	/* Last line read not ended by a '\n' */
 	if (line->next == tail) {	/* after do_rectangular_paste */
-		line = line_insert_after (line, "\n", 1, line->return_type);
-		line = line->prev;
+		LINE * new_line = line_insert_after (line, "\n", 1, line->return_type);
+		if (new_line == NIL_LINE) {
+			/* cannot append empty line; ignore */
+		}
 		trace_rectangular_paste (("NO_LINE: append line\n"));
 		trace_line ("eof++", line);
 	}
@@ -1182,7 +1307,7 @@ insert_file (fd, stay_old_pos, from_text_file)
 		pasted_end_line = line;
 		pasted_end_textp = line->text + length_of (line_buffer);
 	}
-  } else if (line_count > 0 && ! rectangular_paste_flag) {
+  } else if (line_count > 0 && ! rectangular_paste_mode) {
 	pasted_end_line = line->next;
 	pasted_end_textp = line->next->text;
 	trace_line ("pasted_end", pasted_end_line);
@@ -1191,9 +1316,13 @@ insert_file (fd, stay_old_pos, from_text_file)
 /* report get_line errors */
   show_get_l_errors ();
 
+/* rescue post-error situation */
+  if (ret == ERRORS) {
+	pasted_end_textp = line->text;
+  }
 
 /* Update the screen */
-  if (line_count == 0 && ! rectangular_paste_flag) {
+  if (line_count == 0 && ! rectangular_paste_mode) {
 	/* Only one line changed */
 	print_line (y, line);
 
@@ -1268,16 +1397,29 @@ paste_file (fd, stay_old_pos, from_text_file)
  */
 static
 void
-paste_buffer (old_pos)
+paste_buffer (old_pos, use_clipboard)
   FLAG old_pos;
+  FLAG use_clipboard;
 {
   register int fd;		/* File descriptor for buffer */
+  FLAG save_lineends_CRLFtoLF = lineends_CRLFtoLF;
 
   if (dont_modify ()) {
 	return;
   }
 
-  if (hop_flag > 0) {
+#ifdef __CYGWIN__
+  if (use_clipboard) {
+	if ((fd = open ("/dev/clipboard", O_RDONLY | O_BINARY, 0)) < 0) {
+		error ("Cannot access clipboard");
+		return;
+	}
+	if (cur_line->return_type == lineend_LF) {
+		lineends_CRLFtoLF = True;	/* temporary, reset below */
+	}
+  } else
+#endif
+  if (use_clipboard || hop_flag > 0) {
 	if ((fd = open (yankie_file, O_RDONLY | O_BINARY, 0)) < 0) {
 		error ("No inter window buffer present");
 		return;
@@ -1298,6 +1440,8 @@ paste_buffer (old_pos)
   }
   /* Insert the buffer */
   paste_file (fd, old_pos, False);
+
+  lineends_CRLFtoLF = save_lineends_CRLFtoLF;
 }
 
 
@@ -1324,7 +1468,7 @@ set_yank_file_name (buf_name, which, no)
 #ifdef msdos
   build_string (buf_name, "%s-%s.%d", yankie_file, which, no);
 #else
-  build_string (buf_name, "%s.%s.%d_%d", yankie_file, which, getpid (), no);
+  build_string (buf_name, "%s_%s.%d_%d", yankie_file, which, (int) getpid (), no);
 #endif
 }
 
@@ -1367,7 +1511,7 @@ scratchfile (mode, append, buf_name, which, buf_status)
 	/* Generate file name. */
 	/*set_yank_file_name (buf_name, which, yank_buf_no);*/
 	/* Check file existence */
-	if (access (buf_name, 0 /* F_OK */) == 0
+	if (access (buf_name, F_OK) == 0
 	    || (
 		fd = open (buf_name, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, bufprot)
 		) < 0)
@@ -1429,20 +1573,27 @@ htmlfile (mode, append)
 void
 PASTEEXT ()
 {
-  hop_flag = 1;
-  PASTE ();
+  if (hop_flag) {	/* paste from clipboard */
+	hop_flag = 0;
+	paste_buffer (paste_stay_left, True);
+  } else {		/* paste from other mined session */
+	hop_flag = 1;
+	paste_buffer (paste_stay_left, False);
+  }
 }
 
 void
 PASTE ()
 {
-  paste_buffer (paste_stay_left);
+  FLAG use_clipboard = (keyshift & shift_mask) || (hop_flag && (keyshift & alt_mask));
+  paste_buffer (paste_stay_left, use_clipboard);
 }
 
 void
 PASTEstay ()
 {
-  paste_buffer (True);
+  FLAG use_clipboard = (keyshift & shift_mask) || (hop_flag && (keyshift & alt_mask));
+  paste_buffer (True, use_clipboard);
 }
 
 void
@@ -1463,7 +1614,7 @@ YANKRING ()
 		clear_status ();
 
 		revert_yank_buf ();
-		paste_buffer (False);
+		paste_buffer (False, False);
 	}
   } else if (cur_line == pasted_start_line && cur_text == pasted_start_textp 
 	     && check == SAME)
@@ -1477,7 +1628,7 @@ YANKRING ()
 		clear_status ();
 
 		revert_yank_buf ();
-		paste_buffer (True);
+		paste_buffer (True, False);
 	}
   } else {
 	error ("No previous paste to exchange");
@@ -1523,7 +1674,7 @@ INSFILE ()
   }
 
 /* Get the file name */
-  if (get_filename ("Get and insert file:", name, False) != FINE) {
+  if (get_filename ("Insert file:", name, False) != FINE) {
 	return;
   }
   clear_status ();
@@ -1649,7 +1800,9 @@ setMARK (set_only)
   mark_line = cur_line;
   mark_text = cur_text;
   start_highlight_selection ();
-  status_uni ("Mark set - type F1 k for help on copy/paste - HOP Mark toggles rectangular");
+  if (! set_only) {
+	status_uni ("Mark set - type F1 k for help on copy/paste - HOP Mark toggles rectangular");
+  }
 }
 
 void
@@ -1660,7 +1813,7 @@ MARK ()
 	GOMA ();
   } else if (highlight_selection && mark_line == cur_line && mark_text == cur_text) {
 	clear_highlight_selection ();
-	status_uni ("Mark set - selection hidden");
+	status_uni ("Mark set -selection hidden- HOP Mark toggles rectangular");
   } else {
 	setMARK (False);
   }
@@ -1723,8 +1876,8 @@ MARKn (n)
 		error ("Marker # out of range");
 		return;
 	}
-	mark_n_line [n] = cur_line;
-	mark_n_text [n] = cur_text;
+	marker_n [n].line = cur_line;
+	marker_n [n].text = cur_text;
 	status_msg ("Marker set");
   }
 }
@@ -1740,8 +1893,8 @@ mark_n (n)
 		error ("Marker # out of range");
 		return;
 	} else {
-		mark_n_line [n] = cur_line;
-		mark_n_text [n] = cur_text;
+		marker_n [n].line = cur_line;
+		marker_n [n].text = cur_text;
 	}
 }
 
@@ -1759,10 +1912,10 @@ GOMAn (n)
 	return;
   }
 
-  if (checkmark (mark_n_line [n], mark_n_text [n]) == NOT_VALID) {
+  if (checkmark (marker_n [n].line, marker_n [n].text) == NOT_VALID) {
 	error ("Marker not set");
   } else {
-	move_address (mark_n_text [n], find_y (mark_n_line [n]));
+	move_address (marker_n [n].text, find_y (marker_n [n].line));
   }
 }
 
@@ -1775,6 +1928,13 @@ static
 void
 yankie ()
 {
+#ifdef __CYGWIN__
+  /* also copy to Windows clipboard */
+  if (copyfile (yank_file, "/dev/clipboard") != True) {
+	/* ignore error */
+  }
+#endif
+
 #ifdef linkyank
   (void) delete_file (yankie_file);
   if (link (yank_file, yankie_file) == 0) {
@@ -1783,6 +1943,12 @@ yankie ()
   if (geterrno () != EPERM
 #ifdef EOPNOTSUPP
    && geterrno () != EOPNOTSUPP	/* Haiku */
+#endif
+#ifdef ENOTSUP
+   && geterrno () != ENOTSUP	/* just in case */
+#endif
+#ifdef ENOSYS
+   && geterrno () != ENOSYS	/* just in case */
 #endif
   ) {
     /*	no error handling here as a message would inappropriately 
@@ -1836,14 +2002,14 @@ yank_block (remove, append)
 		set_buffer_open (append);
 		yank_text (yankfile (WRITE, append), & yank_status, 
 				mark_line, mark_text, cur_line, cur_text, 
-				remove, append, rectangular_paste_flag);
+				remove, append, rectangular_paste_mode, FORWARD);
 		yankie ();
 		break;
 	case BIGGER :
 		set_buffer_open (append);
 		yank_text (yankfile (WRITE, append), & yank_status, 
 				cur_line, cur_text, mark_line, mark_text, 
-				remove, append, rectangular_paste_flag);
+				remove, append, rectangular_paste_mode, REVERSE);
 		yankie ();
 		break;
 	case SAME :
@@ -1853,6 +2019,7 @@ yank_block (remove, append)
 		error ("Internal mark error");
 		return;
   }
+  alt_rectangular_mode = False;
 }
 
 void
@@ -1866,12 +2033,12 @@ yank_HTML (remove)
 	case SMALLER :
 		yank_text (htmlfile (WRITE, False), & html_status, 
 				mark_line, mark_text, cur_line, cur_text, 
-				remove, False, False);
+				remove, False, False, FORWARD);
 		break;
 	case BIGGER :
 		yank_text (htmlfile (WRITE, False), & html_status, 
 				cur_line, cur_text, mark_line, mark_text, 
-				remove, False, False);
+				remove, False, False, REVERSE);
 		break;
 	case SAME :
 		error ("HTML tag selection failed");
@@ -1889,7 +2056,9 @@ yank_HTML (remove)
 void
 COPY ()
 {
-  clear_highlight_selection ();
+  if (visselect_copydeselect) {
+	clear_highlight_selection ();
+  }
 
   if (append_flag) {
 	yank_block (NO_DELETE, True);
@@ -1926,8 +2095,7 @@ CUT ()
 |*			Selection highlighting				*|
 \*======================================================================*/
 
-#define dont_debug_sel
-#ifdef debug_sel
+#ifdef debug_sel_text
 static
 void
 trace_line_text (line, text)
@@ -1981,7 +2149,7 @@ select_line (line, sel_begin, sel_end)
 	line->dirty = True;
   }
 #ifdef debug_sel
-  printf ("%08X %08X %c ", sel_begin, sel_end, line->dirty ? '!' : '-');
+  printf ("%p %p %c ", sel_begin, sel_end, line->dirty ? '!' : '-');
   trace_line_text (line, 0);
 #endif
 }
@@ -2005,7 +2173,10 @@ do_update_selection_marks (select, pos_x)
      print_line should clear and optionally evaluate the dirty flag
    */
   if (highlight_selection) {
-	int start_col, end_col;	/* in case of rectangular selection */
+	/* only used if rectangular_paste_mode: (avoid -Wmaybe-uninitialized) */
+	int start_col = 0;
+	int end_col = 0;
+
 	LINE * clearup;
 	LINE * cleardown;
 	FLAG check = NOT_VALID;
@@ -2017,16 +2188,39 @@ do_update_selection_marks (select, pos_x)
 		highlight_selection = False;
 	}
 
-	if (check != NOT_VALID && rectangular_paste_flag) {
+	if (check != NOT_VALID && rectangular_paste_mode) {
 		/* check block boundaries */
-		start_col = get_text_col (mark_line, mark_text);
-		/*end_col = get_text_col (cur_line, cur_text);*/
-		end_col = pos_x;
+		start_col = get_text_col (mark_line, mark_text, False);
+		/*end_col = pos_x;*/
+		end_col = get_text_col (cur_line, cur_text, False);
+#ifdef debug_sel
+		printf ("start: %d/%d .. end @%d(%d): %d/%d -> %d..%d",
+			get_text_col (mark_line, mark_text, False),
+			get_text_col (mark_line, mark_text, True),
+			pos_x, last_sel_x,
+			get_text_col (cur_line, cur_text, False),
+			get_text_col (cur_line, cur_text, True),
+			start_col, end_col);
+#endif
+		/* adapt column to actual mouse position */
+		if (last_sel_x > end_col) {
+			end_col = last_sel_x;
+#ifdef debug_sel
+			printf (" ->> %d..%d", start_col, end_col);
+#endif
+		}
+		/* fix right-to-left selection */
 		if (start_col > end_col) {
 			int dum = end_col;
 			end_col = start_col;
 			start_col = dum;
+#ifdef debug_sel
+			printf (" -/> %d..%d", start_col, end_col);
+#endif
 		}
+#ifdef debug_sel
+		printf ("\n");
+#endif
 	}
 
 	if (check == SMALLER) {
@@ -2041,7 +2235,7 @@ do_update_selection_marks (select, pos_x)
 		while (line != header) {
 			char * sel_begin;
 			int dumcol;
-			if (rectangular_paste_flag) {
+			if (rectangular_paste_mode) {
 				sel_begin = text_at (line, & dumcol, start_col);
 				sel_end = text_at (line, & dumcol, end_col);
 			} else if (line == mark_line) {
@@ -2078,7 +2272,7 @@ do_update_selection_marks (select, pos_x)
 		while (line != tail) {
 			char * sel_end;
 			int dumcol;
-			if (rectangular_paste_flag) {
+			if (rectangular_paste_mode) {
 				sel_begin = text_at (line, & dumcol, start_col);
 				sel_end = text_at (line, & dumcol, end_col);
 			} else if (line == mark_line) {
@@ -2124,13 +2318,6 @@ do_update_selection_marks (select, pos_x)
   }
 }
 
-void
-update_selection_marks (pos_x)
-  int pos_x;
-{
-  do_update_selection_marks (True, pos_x);
-}
-
 
 void
 start_highlight_selection ()
@@ -2139,18 +2326,35 @@ start_highlight_selection ()
 }
 
 void
-continue_highlight_selection ()
+clear_highlight_selection ()
 {
-  highlight_selection = True;
-  update_selection_marks (x);
+  do_update_selection_marks (False, 0);
+}
+
+
+void
+update_selection_marks (until_x)
+  int until_x;
+{
+#define fix_shift_update
+#ifdef fix_shift_update
+  until_x -= cur_line->shift_count * SHIFT_SIZE;
+#endif
+  if (until_x != LINE_END) {
+	last_sel_x = until_x;
+  }
+  do_update_selection_marks (True, until_x);
 }
 
 void
-clear_highlight_selection ()
+continue_highlight_selection (until_x)
+  int until_x;
 {
-  do_update_selection_marks (False, x);
+  highlight_selection = True;
+  update_selection_marks (until_x);
 }
 
+
 FLAG
 has_active_selection ()
 {
@@ -2163,6 +2367,21 @@ has_active_selection ()
   return False;
 }
 
+void
+trigger_highlight_selection ()
+{
+  if (keyshift & shift_mask) {
+	if (has_active_selection ()) {
+		continue_highlight_selection (last_sel_x);
+	} else {
+		setMARK (True);
+	}
+	keyshift &= ~ shift_mask;
+  } else {
+	clear_highlight_selection ();
+  }
+}
+
 void
 toggle_rectangular_paste_mode ()
 {
@@ -2176,13 +2395,25 @@ toggle_rectangular_paste_mode ()
 /*	display_flush ();	*/
 }
 
+void
+adjust_rectangular_mode (alt_mouse)
+  FLAG alt_mouse;
+{
+  if (alt_mouse != alt_rectangular_mode) {
+	alt_rectangular_mode = alt_mouse;
+/*	update_selection_marks (x);	call followed by move_to anyway */
+/*	display_flush ();	*/
+  }
+}
+
+
 void
 SELECTION ()
 {
   if (hop_flag > 0) {
 	clear_highlight_selection ();
   } else {
-	continue_highlight_selection ();
+	continue_highlight_selection (last_sel_x);
 	if (! highlight_selection) {
 		error ("Mark not set");
 	}
