@@ -4,12 +4,25 @@
 \*======================================================================*/
 
 #include "mined.h"
+
+#ifndef VAXC
+# ifndef __USE_BSD	/* avoid warning [enabled by default] on Raspberry */
+# define __USE_BSD	/* enable file types for d_type (DT_DIR, DT_REG) */
+# endif
+#include <dirent.h>
+#include <sys/stat.h>
+#endif
+
+#include "charprop.h"
 #include "io.h"
 #include "termprop.h"
 
-#ifndef VAXC
-#include <dirent.h>
-#include <sys/stat.h>
+#if ! defined (__TURBOC__) && ! defined (VAXC)
+#define use_hint_timer
+#endif
+
+#ifdef use_hint_timer
+#include <sys/time.h>
 #endif
 
 
@@ -22,7 +35,7 @@ FLAG input_active = False;
 \*======================================================================*/
 
 /* bottom status contents, kept static for redraw */
-static char * bottom_inbuf;
+static char * bottom_inbuf = NIL_PTR;
 static FLAG bottom_attrib;
 static char bottom_status_buf [2 * maxPROMPTlen + maxFILENAMElen];
 static int bottom_begin_fmt;
@@ -274,14 +287,38 @@ print_unichar (unichar, width, char_begin, line_begin)
 	char_on_status_line = False;
 }
 
+static
+int
+prompt_col_count (string, middle)
+  char * string;
+  int middle;
+{
+  int count = 0;
+  char * start = string;
+
+  while (* string != '\0') {
+	if ((character) (* string) < (character) ' ') {
+		string ++;
+		if (count < middle) {
+			count += 2;
+		}
+	} else {
+		advance_char_scr (& string, & count, start);
+	}
+  }
+  return count;
+}
+
 /**
    Print a string to status line, limited by screen width if limit > 0.
    int limit: either 0 or XBREAK_bottom
-   FLAG attributes = True enables formatting controls in text string:
+   FLAG attributes = VALID enables formatting controls in text string:
 			unicode indicator (cyan background)
 			highlight (red background)
 			reverse
 			plain (normal)
+			(unused) drop initial formatting (for filename)
+	attributes = BIGGER enables right alignment if needed
  */
 static
 void
@@ -291,22 +328,66 @@ printlim_string (text, limit, attributes)
   FLAG attributes;
 {
   int utfcount;
-  unsigned long unichar;
-  unsigned long cjkchar;
-  int charwidth;
   char * start = text;
+  int charwidth;
+  unsigned long unichar;
+  /* if ! transformed && cjk_text: (avoid -Wmaybe-uninitialized) */
+  unsigned long cjkchar = 0;
 
   char_on_status_line = True;
+
+  if (attributes == BIGGER) {
+	int scols = prompt_col_count (text, bottom_begin_fmt);
+	int colsok = XMAX - 2;
+
+	if (scols > colsok) {
+		int dropped = 0;
+
+		/* print indicator */
+		emph_on ();
+		menuheader_on ();
+		putstring ("...");
+		disp_normal ();	
+		diagdisp_on ();
+		lpos += 3;
+		/* adjust available space */
+		colsok -= 3;
+
+		while (scols - dropped > colsok && * text) {
+			if ((character) (* text) < (character) ' ') {
+				text ++;
+				if (dropped < bottom_begin_fmt) {
+					dropped += 2;
+				}
+			} else {
+				advance_char_scr (& text, & dropped, start);
+			}
+		}
+	}
+
+	attributes = VALID;
+  }
+
   while (* text != '\0') {
+	FLAG transformed = False;
 	if (text == mnemo) {
 		diagdisp_off ();
 	}
 
-	if (utf8_text) {
+	if (* text == '-' && text >= start + bottom_begin_fmt && text > start
+			&& (character) * (text - 1) <= (character) ' '
+			&& (character) * (text + 1) <= (character) ' '
+			&& bottom_inbuf == NIL_PTR
+			&& utf8_screen) {
+		transformed = True;
+		utfcount = 1;
+		unichar = 0x2013;	/* – */
+		charwidth = uniscrwidth (unichar, text, start);
+	} else if (utf8_text) {
 		utf8_info (text, & utfcount, & unichar);
 		charwidth = uniscrwidth (unichar, text, start);
 	} else if (iscontrol ((character) * text)) {
-		if (attributes && text >= start + bottom_begin_fmt && * text != '\n') {
+		if (attributes == VALID && text >= start + bottom_begin_fmt && * text != '\n') {
 			charwidth = 0;
 		} else {
 			charwidth = 2;
@@ -333,7 +414,11 @@ printlim_string (text, limit, attributes)
 		print_mark ();	/* also incrementing lpos */
 		text ++;
 	} else if (iscontrol ((character) * text)) {
-		if (attributes && text >= start + bottom_begin_fmt) {
+		if (attributes == BIGGER && * text == '') {
+			text ++;
+			attributes = VALID;
+		} else if (attributes == BIGGER ||
+			   (attributes == VALID && text >= start + bottom_begin_fmt)) {
 			switch (* text) {
 			case '':	unidisp_on ();
 					break;
@@ -352,8 +437,12 @@ printlim_string (text, limit, attributes)
 			putchar ('^');
 			put_unichar (controlchar (* text ++));
 		}
+	} else if (transformed) {
+		/* print and increment lpos */
+		print_unichar (unichar, charwidth, text, start);
+		text += utfcount;
 	} else if (utf8_text) {
-		if (unichar == 0xA0 && attributes && text >= start + bottom_begin_fmt) {
+		if (unichar == 0xA0 && attributes == VALID && text >= start + bottom_begin_fmt) {
 			/* skip NO-BREAK SPACE separator */
 		} else {
 			/* print and increment lpos */
@@ -397,7 +486,7 @@ prompt (term)
   char * term;
 {
   character c;
-  while ((c = readcharacter ()) != '\033' && ! strchr (term, c) && ! quit) {
+  while ((c = readcharacter_unicode ()) != '\033' && ! strchr (term, c) && ! quit) {
 	ring_bell ();
 	flush ();
   }
@@ -752,6 +841,31 @@ input_character (inbuf, ptr, c)
 	}
 }
 
+static
+char *
+input_hexdig (inbuf, ptr, c)
+  char * inbuf;
+  char * ptr;
+  unsigned long c;
+{
+	if (c < 10) {
+		return input_character (inbuf, ptr, c + '0');
+	} else {
+		return input_character (inbuf, ptr, c - 10 + 'A');
+	}
+}
+
+static
+char *
+input_hex (inbuf, ptr, c)
+  char * inbuf;
+  char * ptr;
+  unsigned long c;
+{
+	ptr = input_hexdig (inbuf, ptr, (c >> 4) & 0xF);
+	return input_hexdig (inbuf, ptr, c & 0xF);
+}
+
 void
 redraw_prompt ()
 {
@@ -761,7 +875,7 @@ redraw_prompt ()
 		bottom_lines --;
 	}
 	if (MENU && ! can_add_line) {
-		displaymenuline ();
+		displaymenuline (True);
 	}
 	rd_bottom_line ();
 	if (can_clear_eol) {
@@ -913,7 +1027,7 @@ input (inbuf, clearfl, term_input)
   char * term_input;
 {
   char * ptr;
-  unsigned long c;
+  unsigned long c = 0;	/* (avoid -Wmaybe-uninitialized) */
   unsigned long mnemoc;
   char * term_chars;
 
@@ -936,11 +1050,36 @@ input (inbuf, clearfl, term_input)
 		flush ();
 		c = readcharacter ();
 	} else {
+		char * prev_bottom_inbuf = bottom_inbuf;
+		FLAG prev_bottom_attrib = bottom_attrib;
+		int prev_bottom_begin_fmt = bottom_begin_fmt;
+		FLAG prev_bottom_utf8_text = bottom_utf8_text;
+		char prev_bottom_status_buf [sizeof (bottom_status_buf)];
+		strncpy (prev_bottom_status_buf, bottom_status_buf, sizeof (bottom_status_buf));
+
 		flush ();
 		c = readcharacter_mapped ();
+
+		if (prev_bottom_inbuf != bottom_inbuf) {
+			/* after nested status line (overlayed by pick list),
+			   restore bottom inbuf information,
+			   and redraw prompt line
+			 */
+			bottom_inbuf = prev_bottom_inbuf;
+			bottom_attrib = prev_bottom_attrib;
+			bottom_begin_fmt = prev_bottom_begin_fmt;
+			bottom_utf8_text = prev_bottom_utf8_text;
+			strncpy (bottom_status_buf, prev_bottom_status_buf, sizeof (bottom_status_buf));
+
+			redraw_prompt ();
+		}
 	}
 	if (no_char (c)) {
-		ring_bell ();
+		if (c == CHAR_INVALID) {
+			ring_bell ();
+		} else {	/* CHAR_UNKNOWN, aborted pick list */
+			/* redraw_prompt () already called */
+		}
 	} else {
 		break;
 	}
@@ -1000,6 +1139,53 @@ input (inbuf, clearfl, term_input)
 	} else {
 		ring_bell ();
 	}
+     } else if (command (c) == AltX) {
+	char * cp = ptr;
+	char * pp = cp;
+	char * pp1;
+	int n = 0;
+	unsigned long c = 0;
+	precede_char (& cp, inbuf);
+	pp1 = cp;
+	while (n < 6 && cp != pp && ishex (* cp)) {
+		n ++;
+		pp = cp;
+		precede_char (& cp, inbuf);
+	}
+	if (pp != ptr && n >= 2) {
+		/* hex value found */
+		char * hp = pp;
+		while (hp != ptr) {
+			c = ((c << 4) + hexval (* hp));
+			hp ++;
+		}
+		if (c > (unsigned long) 0x10FFFF) {
+			n = 1;
+		}
+	}
+	if (n >= 2) {
+		/* hex -> uni */
+		while (n > 0) {
+			ptr = erase1 (inbuf, ptr, False);
+			n --;
+		}
+		ptr = input_character (inbuf, ptr, c);
+	} else if (mnemo == NIL_PTR ? ptr > inbuf : (char *) ptr > mnemo) {
+		/* uni -> hex */
+		unsigned long cv = unicodevalue (pp1);
+		if (! no_char (cv)) {
+			ptr = erase1 (inbuf, ptr, False);
+
+			if (cv > 0xFFFF) {
+				if (cv > 0x10FFFF) {
+					ptr = input_hex (inbuf, ptr, cv >> 24);
+				}
+				ptr = input_hex (inbuf, ptr, cv >> 16);
+			}
+			ptr = input_hex (inbuf, ptr, cv >> 8);
+			ptr = input_hex (inbuf, ptr, cv);
+		}
+	}
      } else if (command (c) == CTRLINS) {
 	c = readcharacter_unicode ();
 	if (command (c) == COMPOSE && keyshift >= '@' && keyshift <= '_') {
@@ -1079,7 +1265,6 @@ clear_status ()		 is (void) bottom_line (False, NIL_PTR, NIL_PTR, NIL_PTR, False
 get_string (str1, str2, fl, term_chars)
 			 is bottom_line (True, (str1), NIL_PTR, (str2), fl, term_chars)
  */
-
 static
 void
 status2_uni (msg1, msg2)
@@ -1087,12 +1272,18 @@ status2_uni (msg1, msg2)
   char * msg2;
 {
   FLAG save_utf8_text = utf8_text;
+  FLAG save_cjk_text = cjk_text;
+  FLAG save_mapped_text = mapped_text;
   utf8_text = True;
+  cjk_text = False;
+  mapped_text = False;
   /* don't need to save and set utf16_file = False; */
 
   bottom_line (VALID, msg1, msg2, NIL_PTR, False, "");
 
   utf8_text = save_utf8_text;
+  cjk_text = save_cjk_text;
+  mapped_text = save_mapped_text;
 }
 
 void
@@ -1108,7 +1299,7 @@ status2_prompt (term, msg1, msg2)
   char * msg1;
   char * msg2;
 {
-  (void) bottom_line (VALID, msg1, msg2, NIL_PTR, False, term);
+  (void) bottom_line (BIGGER, msg1, msg2, NIL_PTR, False, term);
   return prompt (term);
 }
 
@@ -1153,10 +1344,10 @@ rd_bottom_line ()
 	lpos = 0;
 	if (bottom_utf8_text && ! utf8_text) {
 		utf8_text = True;
-		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib == VALID);
+		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib);
 		utf8_text = False;
 	} else {
-		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib == VALID);
+		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib);
 	}
   } else {
 	lpos = 0;
@@ -1174,15 +1365,17 @@ rd_bottom_line ()
    Clear or display status line, read prompt if desired.
    s1, s2: strings to display
    inbuf: buffer for prompt to read
-   attrib: True: turn on reverse video on both strings
-           False: don't (clear_status)
-           VALID: interpret control chars in string for attributes
+   attrib:
+	True: turn on reverse video on both strings
+	False: don't (clear_status)
+	VALID: interpret control chars in string for attributes
 			(at beginning) suppress embedding spaces
 		plus printlim_string controls:
 			unicode indicator (cyan background)
 			highlight (red background)
 			reverse
 			plain (normal)
+	BIGGER: like VALID, and align right (cutting left if needed)
  */
 int
 bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
@@ -1194,6 +1387,14 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   char * term_input;
 {
   int ret = FINE;
+#ifdef redraw_info_stack_workaround_doesnt_work
+  char * prev_bottom_inbuf = bottom_inbuf;
+  FLAG prev_bottom_attrib = bottom_attrib;
+  int prev_bottom_begin_fmt = bottom_begin_fmt;
+  FLAG prev_bottom_utf8_text = bottom_utf8_text;
+  char prev_bottom_status_buf [sizeof (bottom_status_buf)];
+  strncpy (prev_bottom_status_buf, bottom_status_buf, sizeof (bottom_status_buf));
+#endif
 
   if (debug_mined && (s1 || s2)) {
 	debuglog ("prompt", s1, s2);
@@ -1221,6 +1422,7 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   bottom_attrib = attrib;
   bottom_utf8_text = utf8_text;
   bottom_begin_fmt = 0;
+
   if (attrib == VALID && s1 && * s1 == '') {
 	build_string (bottom_status_buf, "%s%s", unnull (s1), unnull (s2));
 	if (s2 != NIL_PTR) {
@@ -1248,7 +1450,7 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   }
   if (inbuf == NIL_PTR) {
 	lpos = 0;
-	printlim_string (bottom_status_buf, XBREAK_bottom, attrib == VALID);
+	printlim_string (bottom_status_buf, XBREAK_bottom, attrib);
   } else {
 	lpos = 0;
 	print_string (bottom_status_buf);
@@ -1269,7 +1471,7 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
 		diagdisp_on ();
 		if (bottom_inbuf == NIL_PTR) {
 			lpos = 0;
-			printlim_string (bottom_status_buf, XBREAK_bottom, attrib == VALID);
+			printlim_string (bottom_status_buf, XBREAK_bottom, attrib);
 		} else {
 			lpos = 0;
 			print_string (bottom_status_buf);
@@ -1295,6 +1497,20 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   if (ret != FINE) {
 	clear_status ();
   }
+
+#ifdef redraw_info_stack_workaround_doesnt_work
+  if (prev_bottom_inbuf != NIL_PTR) {
+	/* restore bottom line info for redrawing after nested status lines,
+	   or rather status line overlayed by pick list
+	 */
+	bottom_inbuf = prev_bottom_inbuf;
+	bottom_attrib = prev_bottom_attrib;
+	bottom_begin_fmt = prev_bottom_begin_fmt;
+	bottom_utf8_text = prev_bottom_utf8_text;
+	strncpy (bottom_status_buf, prev_bottom_status_buf, sizeof (bottom_status_buf));
+  }
+#endif
+
   return ret;
 }
 
@@ -1327,7 +1543,7 @@ get_number (message, firstdigit, result)
   if (firstdigit > '\0') {
 	ch = firstdigit;
   } else {
-	ch = readcharacter_mapped ();
+	ch = readcharacter ();
   }
 
   if (ch == quit_char || ch == '\033') {
@@ -1361,7 +1577,7 @@ get_number (message, firstdigit, result)
 		value *= 10;
 		value += ch - '0';
 	}
-	ch = readcharacter_mapped ();
+	ch = readcharacter ();
 	if (ch == quit_char || ch == '\033') {
 		quit = True;
 	}
@@ -1384,7 +1600,7 @@ prompt_num_char (result, maxvalue)
   unsigned long * result;
   unsigned long maxvalue;
 {
-  unsigned long ch;
+  unsigned long ch = 0;	/* (avoid -Wmaybe-uninitialized) */
   int value = 0;
   int i;
   FLAG ok = False;
@@ -1513,12 +1729,11 @@ get_filename (message, file, directory_chooser)
 
   strcpy (file, "");
 
-#ifndef msdos
   if (file0 [0] == '~' && (file0 [1] == '/' || file0 [1] == '\0')) {
 	strappend (file, gethomedir (), maxFILENAMElen);
 	filep ++;
+	strip_trailingslash (file);
   }
-#endif
 
   strappend (file, filep, maxFILENAMElen);
   return ret;
@@ -1559,19 +1774,26 @@ fncompare (s1, s2)
   if (sort_by_extension) {
 	char * suf1 = getsuffix (s1);
 	char * suf2 = getsuffix (s2);
-	int cmp = stringcompare_ (suf1, suf2);
+	int cmp = strcmp (suf1, suf2);
 	if (cmp != 0) {
 		return cmp;
 	} else {
-		return stringcompare_ (s1, s2);
+		return strcmp (s1, s2);
 	}
   } else {
-	return stringcompare_ (s1, s2);
+	return strcmp (s1, s2);
   }
 }
 
+/**
+   Insert file name and info into binary tree.
+   Return:
+   	True if a new entry was inserted
+   	VALID if a duplicate entry was ignored
+   	False if out of memory
+ */
 static
-void
+FLAG
 tree_insert (nodepoi, s1, s2)
   struct treenode * * nodepoi;
   char * s1;
@@ -1580,23 +1802,30 @@ tree_insert (nodepoi, s1, s2)
   if (* nodepoi) {
 	int cmp = fncompare (s1, (* nodepoi)->s1);
 	if (cmp < 0) {
-		tree_insert (& (* nodepoi)->left, s1, s2);
+		return tree_insert (& (* nodepoi)->left, s1, s2);
 	} else if (cmp > 0) {
-		tree_insert (& (* nodepoi)->right, s1, s2);
+		return tree_insert (& (* nodepoi)->right, s1, s2);
 	} else {
-		/* duplicate entry? */
+		/* duplicate entry */
+		return VALID;
 	}
   } else {
 	* nodepoi = alloc (sizeof (struct treenode));
-	(* nodepoi)->left = 0;
-	(* nodepoi)->right = 0;
-	(* nodepoi)->s1 = s1;
-	(* nodepoi)->s2 = s2;
+	if (* nodepoi) {
+		(* nodepoi)->left = 0;
+		(* nodepoi)->right = 0;
+		(* nodepoi)->s1 = s1;
+		(* nodepoi)->s2 = s2;
+		return True;
+	} else {
+		return False;
+	}
   }
 }
 
 static menuitemtype * filemenu;
 static int itemcount;
+static int filecount;
 
 static
 void
@@ -1608,7 +1837,12 @@ tree_traverse (node, dirs)
 	tree_traverse (node->left, dirs);
 	if (dirs == NOT_VALID ||
 	    (dirs == False) == ! node->s2) {
-		fill_menuitem (& filemenu [itemcount ++], node->s1, node->s2);
+		if (itemcount < filecount) {
+			fill_menuitem (& filemenu [itemcount], node->s1, node->s2);
+		} else if (itemcount == filecount) {
+			ring_bell ();
+		}
+		itemcount ++;
 	}
 	tree_traverse (node->right, dirs);
   }
@@ -1659,7 +1893,6 @@ isin (utf, set)
   }
 }
 
-static
 int
 wildcard (pat, s)
   char * pat;
@@ -1671,12 +1904,20 @@ wildcard (pat, s)
 #endif
 
   if (! * pat) {
-	return ! * s;
-  } else if (* pat == '?') {
-	return * s && wildcard (pat + 1, nextutf8 (s));
+	return
+#ifdef vms
+		* s == ';' ||
+#endif
+		! * s;
   } else if (* pat == '*') {
 	return (* s && wildcard (pat, nextutf8 (s)))
 		|| wildcard (pat + 1, s);
+  } else if (* pat == '?') {
+	return * s && wildcard (pat + 1, nextutf8 (s));
+#ifdef vms
+  } else if (* pat == '%') {
+	return * s && wildcard (pat + 1, nextutf8 (s));
+#else
   } else if (* pat == '[') {
 	char * endset = strchr (pat, ']');
 	if (endset) {
@@ -1693,6 +1934,7 @@ wildcard (pat, s)
 		/* invalid pattern */
 		return 0;
 	}
+#endif
   } else {
 #ifdef case_sensitive
 	return * pat == * s	/* transparently compare UTF-8 bytes */
@@ -1719,9 +1961,134 @@ pathlast (path)
   }
 }
 
-#ifndef S_ISDIR
-#define S_ISDIR(mode)	(mode & S_IFMT) == S_IFDIR
+static
+void
+add_pathsep (dirname, maxlen)
+  char * dirname;
+  int maxlen;
+{
+  if (dirname [0] != '\0') {
+	char last = dirname [strlen (dirname) - 1];
+	if (last != '/'
+#ifdef pc
+		&& last != '\\'
 #endif
+#ifdef vms
+		&& last != ']'
+#endif
+#ifndef unix
+		&& last != ':'
+#endif
+		) {
+		strappend (dirname, "/", maxlen);
+	}
+  }
+}
+
+
+/**
+   display hint "checking directory" after waiting a while
+ */
+#ifdef use_hint_timer
+static struct timeval now_start;
+#endif
+
+void
+check_slow_hint ()
+{
+#ifdef use_hint_timer
+  struct timeval now;
+  gettimeofday (& now, 0);
+  if (now.tv_sec > 0 &&
+	(now.tv_sec - now_start.tv_sec) * 1000
+	+ (now.tv_usec - now_start.tv_usec) / 1000 > 600)
+  {
+		status_msg ("...checking directory...");
+		now.tv_sec = 0;
+  }
+#endif
+}
+
+/* workaround for buggy include files ... */
+#if defined (linux) && defined (i386)
+/* d_type is missing for Linux kernel 2.2 */
+# undef _DIRENT_HAVE_D_TYPE
+#endif
+#if ! defined (DT_DIR) || ! defined (DT_REG)
+# undef _DIRENT_HAVE_D_TYPE
+#endif
+
+/* unsuccessful attempts to speed up readdir/stat on slow network drives: */
+#define dont_use_fstatat
+#define dont_use_rel_stat
+/* new attempt to speed up readdir/stat on slow network drives: */
+#define use_lazy_stat
+
+#define dont_debug_lazy_stat
+
+#ifdef debug_lazy_stat
+# undef _DIRENT_HAVE_D_TYPE
+#endif
+
+#ifdef use_lazy_stat
+static char direcname [maxFILENAMElen];
+
+static
+int
+dummyflagoff (item, i)
+  menuitemtype * item;
+  int i;
+{
+  return 0;
+}
+#endif
+
+/**
+   is_directory () lazy check
+ */
+int
+is_directory (item)
+  menuitemtype * item;
+{
+#ifdef use_lazy_stat
+  if (streq (item->hopitemname, "lazy")) {
+	struct stat fstat_buf;
+	char pathname [maxFILENAMElen];
+# if defined (use_fstatat) || defined (use_rel_stat)
+	strcpy (pathname, "");
+# else
+	strcpy (pathname, direcname);
+	add_pathsep (pathname, maxFILENAMElen);
+# endif
+	strappend (pathname, item->itemname, maxFILENAMElen);
+# ifdef vms
+	if (! strchr (pathname, ';')) {
+		/* pull out dir flag for unaccessible dirs */
+		strappend (pathname, ".DIR;1", maxFILENAMElen);
+	}
+# endif
+# ifdef use_fstatat
+#error fd for fstatat missing in lazy mode
+	if (fstatat (dir_fd, pathname, & fstat_buf, 0) < 0) {
+# else
+	if (stat (pathname, & fstat_buf) < 0) {
+# endif
+		/* No such file or directory ... */
+		item->hopitemname = "";
+		item->itemon = dummyflagoff;
+	} else if (S_ISDIR (fstat_buf.st_mode)) {
+		item->hopitemname = "directory";
+	} else {
+		/* not a directory */
+		item->hopitemname = "";
+		item->itemon = dummyflagoff;
+	}
+/*printf ("is_dir: stat <%s> -> %s\n", pathname, item->hopitemname);*/
+  }
+#endif
+
+  return streq (item->hopitemname, "directory");
+}
 
 /*
  * get_filename () opens a file chooser dialog to choose a file name
@@ -1732,77 +2099,149 @@ get_filename (message, file, directory_chooser)
   char * file;
   FLAG directory_chooser;
 {
-  char dirname [maxFILENAMElen];
+#ifndef use_lazy_stat
+  char direcname [maxFILENAMElen];
+#endif
   char * filter = NIL_PTR;
   char dirpath [maxFILENAMElen];
   FLAG continue_chooser = True;
   int ret = ERRORS;
+  FLAG memerr = False;
 
-  strcpy (dirname, ".");
+  strcpy (direcname, ".");
 
   while (continue_chooser) {
+	DIR * dir;
+#ifdef use_fstatat
+	int dir_fd;
+#endif
 	struct dirent * direntry;
-	int filecount = 0;
 	int fi;
-	DIR * dir = opendir (dirname);
+
+#ifdef use_hint_timer
+	gettimeofday (& now_start, 0);
+#else
+# if defined (msdos) || defined (vms)
+	status_msg ("...checking directory...");
+# endif
+#endif
+
+	filecount = 0;
+
+#ifdef use_rel_stat
+	... = chdir (direcname);
+	dir = opendir (".");
+#else
+	dir = opendir (direcname);
+#endif
+#ifdef use_fstatat
+	dir_fd = open (direcname, 0);
+#endif
 	if (! dir) {
-		error2 ("Cannot open directory: ", dirname);
+		error2 ("Cannot open directory: ", direcname);
 		return ERRORS;
 	}
 
 	dir_root = 0;
-	while ((direntry = readdir (dir))) {
+#ifdef vms
+	if (tree_insert (& dir_root, "..", "directory") == True) {
+		filecount ++;
+	}
+#endif
+	while ((direntry = readdir (dir)) != 0) {
+#ifdef use_hint_timer
+		check_slow_hint ();
+#endif
+
 		if (streq (direntry->d_name, ".")
 		    /* && ! directory_chooser */
 		   ) {
 			/* ignore */
 			/* also for change dir, if absolute cwd is listed */
 		} else if (! filter || wildcard (filter, direntry->d_name)) {
-			char * dirflag = NIL_PTR;
-#ifdef bogus_DIRENT_HAVE_D_TYPE
-			if (direntry->d_type == DT_DIR) {
-				dirflag = "directory";
-			}
+		    char * dirflag = NIL_PTR;
+#ifdef _DIRENT_HAVE_D_TYPE
+/*printf ("d_type %02X %s\n", direntry->d_type, direntry->d_name);*/
+		    if (direntry->d_type == DT_DIR) {
+			dirflag = "directory";
+		    } else if (direntry->d_type == DT_REG) {
+			/* don't mark */
+		    } else
+			/* DT_UNKNOWN or DT_LNK (-> directory?) */
+#endif
+#ifdef use_lazy_stat
+			dirflag = "lazy";
 #else
+		    {
 			struct stat fstat_buf;
 			char pathname [maxFILENAMElen];
-			strcpy (pathname, dirname);
-			if (dirname [strlen (dirname) - 1] != '/'
-#ifdef pc
-			 && dirname [strlen (dirname) - 1] != '\\'
-#endif
-			   ) {
-				strappend (pathname, "/", maxFILENAMElen);
-			}
+# if defined (use_fstatat) || defined (use_rel_stat)
+			strcpy (pathname, "");
+# else
+			strcpy (pathname, direcname);
+			add_pathsep (pathname, maxFILENAMElen);
+# endif
 			strappend (pathname, direntry->d_name, maxFILENAMElen);
+# ifdef vms
+			if (! strchr (pathname, ';')) {
+				/* pull out dir flag for unaccessible dirs */
+				strappend (pathname, ".DIR;1", maxFILENAMElen);
+			}
+# endif
+# ifdef use_fstatat
+			if (fstatat (dir_fd, pathname, & fstat_buf, 0) < 0) {
+# else
+/*printf ("stat %s\n", pathname);*/
 			if (stat (pathname, & fstat_buf) < 0) {
+# endif
 				/* No such file or directory ... */
 			} else if (S_ISDIR (fstat_buf.st_mode)) {
+/*printf ("S_ISDIR %s\n", pathname);*/
 				dirflag = "directory";
 			}
+		    }
 #endif
-			if (! directory_chooser || dirflag) {
-				tree_insert (& dir_root, dupstr (direntry->d_name), dirflag);
-				filecount ++;
-			}
+
+		    if (! directory_chooser || dirflag) {
+				char * fn = dupstr (direntry->d_name);
+				FLAG inserted = False;
+				if (fn) {
+					inserted = tree_insert (& dir_root, fn, dirflag);
+				}
+				if (inserted == True) {
+					filecount ++;
+				} else if (inserted == False && ! memerr) {
+					error ("File menu incomplete - cannot allocate memory for checking files");
+					sleep (1);
+					memerr = True;
+				} else if (inserted == VALID) {
+					/* discarded duplicate entry */
+				}
+		    }
 		}
 	}
 	closedir (dir);
+#ifdef use_rel_stat
+#error not implemented: revert from chdir: go to previous directory
+#endif
+#ifdef use_fstatat
+	close (dir_fd);
+#endif
+
+#if defined (msdos) || defined (vms)
+	clear_status ();
+#endif
+
 	/* reset wildcard filter */
 	filter = NIL_PTR;
 
-	/* adjust # entries */
-	filecount += 1;
-	/* allocate menu structure */
-	filemenu = alloc (filecount * sizeof (menuitemtype));
-	itemcount = 0;
 	/* compose pathname of selected directory for display */
-	if (is_absolute_path (dirname)) {
-		strcpy (dirpath, dirname);
+	if (is_absolute_path (direcname)) {
+		strcpy (dirpath, direcname);
 	} else {
 		char * dnpoi;
 		(void) get_cwd (dirpath);
-		dnpoi = dirname;
+		dnpoi = direcname;
 		while (strisprefix ("..", dnpoi) &&
 			(dnpoi [2] == '\0'
 			 || dnpoi [2] == '/'
@@ -1813,7 +2252,28 @@ get_filename (message, file, directory_chooser)
 		      ) {
 			char * lastslash = pathlast (dirpath);
 			if (lastslash) {
+#ifdef vms
+				if (* lastslash == ']') {
+					* lastslash = 0;
+					lastslash = strrchr (dirpath, '.');
+					if (lastslash) {
+						* lastslash ++ = ']';
+						* lastslash = 0;
+					} else {
+						lastslash = strrchr (dirpath, '[');
+						if (lastslash) {
+							* lastslash = 0;
+						} else {
+							* dirpath = 0;
+						}
+						strappend (dirpath, "[000000]", maxFILENAMElen);
+					}
+				} else {
+					* lastslash = 0;
+				}
+#else
 				* lastslash = 0;
+#endif
 			}
 			dnpoi += 2;
 			if (* dnpoi) {
@@ -1821,13 +2281,34 @@ get_filename (message, file, directory_chooser)
 			}
 		}
 		if (* dnpoi && ! streq (dnpoi, ".")) {
-			strappend (dirpath, "/", maxFILENAMElen);
+			add_pathsep (dirpath, maxFILENAMElen);
 			strappend (dirpath, dnpoi, maxFILENAMElen);
 		}
 		if (! * dirpath) {
 			strcat (dirpath, "/");
 		}
 	}
+
+	/* adjust # entries */
+	filecount += 1;
+	/* allocate menu structure */
+	filemenu = alloc (filecount * sizeof (menuitemtype));
+	memerr = False;
+	while (! filemenu && filecount > 1) {
+		if (! memerr) {
+			memerr = True;
+		}
+		filecount --;
+		filemenu = alloc (filecount * sizeof (menuitemtype));
+	}
+	if (! filemenu) {
+		error ("Cannot allocate memory for file menu");
+		return ERRORS;
+	} else if (memerr) {
+		error ("File menu incomplete - not enough memory for menu");
+		sleep (1);
+	}
+	itemcount = 0;
 	/* pre-fill current working directory */
 	fill_menuitem (& filemenu [itemcount ++], dirpath, "directory");
 	/* fill menu with dir/file names */
@@ -1845,64 +2326,132 @@ get_filename (message, file, directory_chooser)
 		ret = ERRORS;
 		continue_chooser = False;
 	} else if (streq (filemenu [fi].itemname, "~")) {
-		strcpy (dirname, "");
-		strappend (dirname, gethomedir (), maxFILENAMElen);
+		strcpy (direcname, "");
+		strappend (direcname, gethomedir (), maxFILENAMElen);
 	} else {
-	    char * chosenname = filemenu [fi].itemname;
-	    char * dirtag = filemenu [fi].hopitemname;
+	    menuitemtype * chosenitem = & (filemenu [fi]);
+	    char * chosenname = chosenitem->itemname;
 	    char expandedname [maxFILENAMElen];
 	    if (chosenname [0] == '~'
 		&& (chosenname [1] == '/' || chosenname [1] == '\0')) {
 		strcpy (expandedname, "");
 		strappend (expandedname, gethomedir (), maxFILENAMElen);
 		chosenname ++;
+		strip_trailingslash (expandedname);
 		strappend (expandedname, chosenname, maxFILENAMElen);
 		chosenname = expandedname;
 	    }
 
 #ifdef debug_file_chooser
-	    printf ("%s -> %s (%s)", dirname, chosenname, dirtag);
+	    printf ("%s -> %s (%s)", direcname, chosenname, chosenitem->hopitemname);
 #endif
-	    if (streq (dirtag, "*")) {
+	    if (streq (chosenitem->hopitemname, "*")) {
 		/* set wildcard filter */
 		filter = chosenname;
 	    } else {
 		/* compose changed directory/file name */
-		if (streq (chosenname, "..")) {
-			char * lastslash = pathlast (dirname);
+		if (streq (chosenname, "..")
+#ifdef vms
+			|| streq (chosenname, "[-]")
+#endif
+			) {
+			char * lastslash = pathlast (direcname);
 			if ((lastslash && streq (lastslash + 1, ".."))
-			    || streq (dirname, "..")) {
-				strappend (dirname, "/..", maxFILENAMElen);
+			    || streq (direcname, "..")) {
+				strappend (direcname, "/..", maxFILENAMElen);
 			} else if (lastslash) {
-				if (lastslash == dirname) {
+#ifdef vms
+				if (* lastslash == ']') {
+				    if (* (lastslash + 1)) {
+					* (lastslash + 1) = '\0';
+				    } else {
+					* lastslash = 0;
+					lastslash = strrchr (direcname, '.');
+					if (lastslash) {
+						* lastslash ++ = ']';
+						* lastslash = 0;
+					} else {
+						lastslash = strrchr (direcname, '[');
+						if (lastslash) {
+							* lastslash = 0;
+						} else {
+							* direcname = 0;
+						}
+						strappend (direcname, "[000000]", maxFILENAMElen);
+					}
+				    }
+				} else {
+					* lastslash = 0;
+				}
+#else
+				if (lastslash == direcname) {
 					lastslash ++;
 				}
 				* lastslash = 0;
-			} else if (streq (dirname, ".")) {
-				strcpy (dirname, "..");
+#endif
+			} else if (streq (direcname, ".")) {
+				strcpy (direcname, "..");
 			} else {
-				strcpy (dirname, ".");
+				strcpy (direcname, ".");
 			}
 #ifdef pc_better_dont
-			if (dirname [1] == ':' && dirname [2] == '\0') {
-				strcat (dirname, "\\");
+			if (direcname [1] == ':' && direcname [2] == '\0') {
+				strcat (direcname, "\\");
 			}
 #endif
-		} else if (streq (chosenname, ".")) {
+		} else if (streq (chosenname, ".")
+#ifdef vms
+			|| streq (chosenname, "[]")
+#endif
+			) {
 			/* no change */
 		} else {
-			if (streq (dirname, ".") || is_absolute_path (chosenname)) {
-				strcpy (dirname, "");
-			} else {
-				if (dirname [strlen (dirname) - 1] != '/'
-#ifdef pc
-				 && dirname [strlen (dirname) - 1] != '\\'
+#ifdef vms
+			char * dirsuf;
+			char * last;
 #endif
-				   ) {
-					strappend (dirname, "/", maxFILENAMElen);
-				}
+			if (streq (direcname, ".") || is_absolute_path (chosenname)) {
+				strcpy (direcname, "");
+			} else {
+				add_pathsep (direcname, maxFILENAMElen);
 			}
-			strappend (dirname, chosenname, maxFILENAMElen);
+#ifdef vms
+			dirsuf = strstr (chosenname, ".DIR");
+			last = direcname + strlen (direcname);
+			if (last != direcname) {
+				last --;
+			}
+			if (dirsuf && (! dirsuf [4] || dirsuf [4] == ';')) {
+				/* dsk: + THAT.DIR -> dsk:[.THAT]
+				   [dir] + THAT.DIR -> [dir.THAT]
+				*/
+				* dirsuf = '\0';
+
+				if (* last == ']') {
+					* last = '.';
+				} else {
+					strappend (direcname, "[.", maxFILENAMElen);
+				}
+				strappend (direcname, chosenname, maxFILENAMElen);
+				strappend (direcname, "]", maxFILENAMElen);
+			} else if (* last == ']' && * chosenname == '[') {
+				/* [a] + [.b] -> [a.b]
+				   [a] + [] -> [a]
+				   [a] + [-] -> [a.-]
+				*/
+				chosenname ++;
+				if (* chosenname == '.' || * chosenname == ']') {
+					* last = '\0';
+				} else {
+					* last = '.';
+				}
+				strappend (direcname, chosenname, maxFILENAMElen);
+			} else {
+				strappend (direcname, chosenname, maxFILENAMElen);
+			}
+#else
+			strappend (direcname, chosenname, maxFILENAMElen);
+#endif
 		}
 
 #ifdef msdos
@@ -1915,34 +2464,48 @@ get_filename (message, file, directory_chooser)
 		else
 #endif
 		if (directory_chooser) {
-			if (streq (dirtag, "cd")) {
+			if (streq (chosenitem->hopitemname, "cd")) {
 				/* navigate into directory after TAB */
 			} else {
 				/* exit with selected directory name */
-				strcpy (file, dirname);
+				strcpy (file, direcname);
 				ret = FINE;
 				continue_chooser = False;
 			}
-		} else if (streq (dirtag, "directory")) {
+		} else if (is_directory (chosenitem)) {
 			/* navigate into selected directory */
 		} else {
 			/* check whether selected name 
 			   is actually a directory
 			*/
 			struct stat fstat_buf;
-			if (stat (dirname, & fstat_buf) == 0
+#ifdef vms
+			strcpy (expandedname, direcname);
+			if (! strchr (expandedname, ';')) {
+				char * lastslash = pathlast (dirpath);
+				if (lastslash &&
+				    (* lastslash == ']' || * lastslash == ':')) {
+				} else {
+					/* pull out dir info for unaccessible dirs */
+					strappend (expandedname, ".DIR;1", maxFILENAMElen);
+				}
+			}
+			if (stat (expandedname, & fstat_buf) == 0
+#else
+			if (stat (direcname, & fstat_buf) == 0
+#endif
 			 && S_ISDIR (fstat_buf.st_mode)) {
 				/* navigate into directory */
 			} else {
 				/* exit with selected file name */
-				strcpy (file, dirname);
+				strcpy (file, direcname);
 				ret = FINE;
 				continue_chooser = False;
 			}
 		}
 	    }
 #ifdef debug_file_chooser
-	    printf (": %s\n", dirname);
+	    printf (": %s\n", direcname);
 #endif
 	}
 
