@@ -4,7 +4,9 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "charprop.h"
 #include "io.h"
+#include "textfile.h"	/* dont_modify () */
 
 
 /*======================================================================*\
@@ -191,10 +193,10 @@ reverse_method (dir)
 }
 
 /*
- * RS searches using the last search direction and expression.
+ * RES searches using the last search direction and expression.
  */
 void
-RS ()
+RES ()
 {
   if (hop_flag > 0) {
 	hop_flag = 0;
@@ -205,10 +207,10 @@ RS ()
 }
 
 /*
- * RSreverse searches using the last search expression and reverse direction.
+ * RESreverse searches using the last search expression and reverse direction.
  */
 void
-RSreverse ()
+RESreverse ()
 {
   if (hop_flag > 0) {
 	hop_flag = 0;
@@ -329,19 +331,6 @@ change (message, global, confirm)
 	return;
   }
 
-/*
-  if (strchr (exp_buf, '\n') != NIL_PTR) {
-	ring_bell ();
-	status_msg ("Warning: new-line in substitution pattern not fully implemented");
-	sleep (3);
-  }
-  if (strchr (replacement, '\n') != NIL_PTR) {
-	ring_bell ();
-	status_msg ("Warning: repl. with new-line roughly implemented - please save afterwards");
-	sleep (3);
-  }
-*/
-
   if (confirm) {
 	Pushmark ();
   }
@@ -458,13 +447,16 @@ change (message, global, confirm)
   quit_change = quit;
 /* Fix the status line */
   if (subs == 0L && quit == False) {
-	error ("Pattern not found");
+	error ("No substitution");
   } else if (lines >= REPORT_CHANGED_LINES || quit) {
-	build_string (mess_buf, "%s%ld substitutions on %ld lines",
-		quit_change ? "(Replace aborted) " : "", subs, lines);
+	build_string (mess_buf, "%s%ld substitution%s on %ld line%s",
+		quit_change ? "(Replace aborted) " : "",
+		subs, subs > 1 ? "s" : "",
+		lines, lines > 1 ? "s" : "");
 	status_msg (mess_buf);
   } else if (global == NOT_VALID && subs >= REPORT_CHANGED_LINES) {
-	status_line (dec_out (subs), " substitutions");
+	status_line (dec_out (subs),
+			subs > 1 ? " substitutions" :  " substitution");
   } else {
 	clear_status ();
   }
@@ -497,7 +489,7 @@ substitute (line, line_poi,
   char * replacement;	/* Contains replacement pattern */
   int * new_lines_inserted;
 {
-  char text_buffer [MAX_CHARS];
+  char text_buffer [maxLINElen];
   register char * textp = text_buffer;
   register char * subp = replacement;
   char * linep = line->text;
@@ -507,6 +499,7 @@ substitute (line, line_poi,
   char * repl_start;
   char * repl_end;
   int repl_len;
+  lineend_type end_return_type = program->end_line->return_type;
 
   set_modified ();
 
@@ -519,16 +512,16 @@ substitute (line, line_poi,
  * Replace the match by the substitution pattern. Each occurrence of '&' is
  * replaced by the original match. A \ escapes the next character.
  */
-  while (* subp != '\0' && textp < & text_buffer [MAX_CHARS]) {
+  while (* subp != '\0' && textp < & text_buffer [maxLINElen]) {
 	if (* subp == '&') {		/* Replace the original match */
 		amp = program->start_ptr;
 		ampl = program->start_line;
 		/* Do not report an error here if aborting */
-		while (textp < & text_buffer [MAX_CHARS] &&
+		while (textp < & text_buffer [maxLINElen] &&
 			(ampl != program->end_line || amp < program->end_ptr)
 		) {
 			if (* amp == '\n') {
-/*printf ("including original newline\n");*/
+				/* include original newline */
 				if (ampl->return_type != lineend_NONE) {
 					* textp ++ = '\n';
 				}
@@ -558,8 +551,8 @@ substitute (line, line_poi,
   }
   * textp = '\0';
 
-/* Check for line length not exceeding MAX_CHARS */
-  if (length_of (text_buffer) + length_of (program->end_ptr) >= MAX_CHARS) {
+/* Check for line length not exceeding maxLINElen */
+  if (length_of (text_buffer) + length_of (program->end_ptr) >= maxLINElen) {
 	if (program->end_line != program->start_line) {
 		error ("Substitution failed: replacement buffer exceeded");
 	} else {
@@ -600,18 +593,19 @@ substitute (line, line_poi,
 	ring_bell ();
 	error ("Substitution failed: cannot allocate more memory");
 	return NIL_PTR;
-  }
-  else {
+  } else {
 	free_space (line->text);
 	line->text = newtext;
 /*	copy_string (line->text, text_buffer);	*/
 	strncpy (line->text, text_buffer, repl_len);
 	line->text [repl_len] = '\0';
-	update_text_state (line);
+	update_syntax_state (line);
 	linep = line->text + (int) (textp - text_buffer);
   }
 
   while (* repl_end != '\0') {
+	lineend_type old_return_type = line->return_type;
+
 	(* new_lines_inserted) ++;
 	repl_start = repl_end;
 	repl_end = strchr (repl_start, '\n');
@@ -622,12 +616,18 @@ substitute (line, line_poi,
 	}
 	repl_len = (int) (repl_end - repl_start);
 
-	line = line_insert_after (line, repl_start, repl_len, line->return_type);
+	if (old_return_type == lineend_NONE || old_return_type == lineend_NUL) {
+		line->return_type = default_lineend;
+	}
+	line = line_insert_after (line, repl_start, repl_len, old_return_type);
 	if (line == NIL_LINE) {
 		ring_bell ();
 		error ("Substitution failed: cannot allocate more lines");
 		return NIL_PTR;
 	}
+	if (* repl_end == '\0') {
+		line->return_type = end_return_type;
+	}
 	* line_poi = line;
 	linep = line->text + (int) (textp - repl_start);
   }
@@ -671,7 +671,7 @@ LR ()
  * These coordinates are passed to move_to ().
  * Re_search () searches using the last search program and direction.
    Call graph of search procedures:
-RS ------------\
+RES ------------\
 SFW -\		> re_search > match -----------\
 SRV --> search <				\
 	        \				 \
@@ -803,7 +803,9 @@ search_corresponding (expr, method, corresponding)
   wrapped_search = False;
   do {
 	int matched;
-	int nestingweight;
+	/* only used if * corresponding == '#': (avoid -Wmaybe-uninitialized) */
+	int nestingweight = 0;
+
 	do {
 		match_line = search_it (program, method);
 		if (match_line == NIL_LINE) {
@@ -931,6 +933,8 @@ search_it (program, method)
 		status_msg ("Search aborted");
 /*		swallow_dummy_quit_char (); */
 		quit = False;
+	} else if (program->status == REG_ERROR) {
+		status_msg ("Search program corrupted");
 	} else {
 		status_msg ("Pattern not found");
 	}
@@ -1050,12 +1054,12 @@ void
 print_expression (expr)
   int * expr;
 {
-  char s [20];
+  char s [22];
   int ei = 0;
 
   while (ei < MAX_EXP_SIZE) {
 	s [0] = '\0';
-	if (* expr & NORMAL)	strcat (s, "-");
+	if (* expr & NORMAL)	strcat (s, "=");
 	if (* expr & NEGATE)	strcat (s, "!");
 	if (* expr & DOT)	strcat (s, ".");
 	if (* expr & EOLN)	strcat (s, "\\n");
@@ -1063,10 +1067,11 @@ print_expression (expr)
 	if (* expr & BRACKET)	strcat (s, "[");
 	if (* expr & DONE)	strcat (s, "DONE");
 	printf ("%s%04X ", s, * expr);
+
+	ei ++;
 	if (* expr & DONE) break;
 	if (* expr == 0) break;
 	expr ++;
-	ei ++;
   }
   printf (" [exp size %d]\n", ei);
 }
@@ -1109,7 +1114,7 @@ compile (pat_start, program, case_ignore)
   character utfbuf [7];
   character * utfpoi;
 
-  trace_search (("compile "));
+  trace_search (("compiling\n"));
 
 /* Check for begin of line */
   if (* pattern == '^') {
@@ -1126,6 +1131,7 @@ compile (pat_start, program, case_ignore)
 
   for (; ;) {
 	c = * pattern ++;
+	trace_search (("	compile @%02X\n", c));
 	switch (c) {
 	case '.' :
 		* expression ++ = DOT;
@@ -1147,6 +1153,9 @@ compile (pat_start, program, case_ignore)
 		* expression ++ = DONE;
 		trace_expression ((exp_buffer));
 		return finished (program, expression);
+	case '\n' :
+		* expression ++ = (NORMAL | EOLN) + '\n';
+		break;
 	case '\\' :
 		/* If last char, it must! mean a normal '\' */
 		if (* pattern == '\0') {
@@ -1154,11 +1163,14 @@ compile (pat_start, program, case_ignore)
 		} else {
 			c = * pattern ++;
 			if (c == 'n') {
-				* expression ++ = NORMAL + '\n';
+				* expression ++ = (NORMAL | EOLN) + 'n';
 			} else if (c == 'r') {
-				* expression ++ = NORMAL + '\r';
+				* expression ++ = (NORMAL | EOLN) + '\r';
+			} else if (c == 'R') {
+				* expression ++ = (NORMAL | EOLN) + 'R';
 			} else if (c == '0') {
-				* expression ++ = NORMAL + '\0';
+				/* NUL is represented as pseudo-line-end */
+				* expression ++ = (NORMAL | EOLN) + '\0';
 			} else {
 				* expression ++ = NORMAL + c;
 			}
@@ -1356,6 +1368,7 @@ compile (pat_start, program, case_ignore)
 	}
 	if (expression == & exp_buffer [MAX_EXP_SIZE]) {
 		reg_error (too_long);
+		trace_search (("compiled "));
 		trace_expression ((exp_buffer));
 		return FINE;
 	}
@@ -1420,7 +1433,7 @@ match (program, string, method)
 		if (method == FORWARD) {
 			status_uni (">>>>>>>> Search wrapped around end of file >>>>>>>>");
 		} else {
-			status_uni ("<<<<<<<< Search wrapped around end of file <<<<<<<<");
+			status_uni ("<<<<<<<< Search wrapped around top of file <<<<<<<<");
 		}
 		wrapped_search = True;
 		continue;
@@ -1525,12 +1538,13 @@ check_string (program, string, from_line, expression)
   LINE * from_line;
   int * expression;
 {
-  register int opcode;		/* Holds opcode of next expr. atom */
-  character patternchar;	/* Char that must be matched */
-  character * mark;		/* For marking position */
-  int has_star;			/* pattern has a star */
+  register int opcode;		/* holds opcode of next expr. atom */
+  character patternchar;	/* char that must be matched */
   int exprlist_length;
   int ret;
+  int has_star;			/* pattern has a star */
+  /* only used if has_star: (avoid -Wmaybe-uninitialized) */
+  character * mark = (character *) NIL_PTR;	/* for marking position */
 
   if (expression == NIL_INT) {
 	expression = program->result.expression;
@@ -1563,7 +1577,48 @@ check_string (program, string, from_line, expression)
 	}
 	expression ++;		/* Increment expr. */
 
-	switch (opcode) {
+	if (opcode & EOLN) {
+	    if (* string == '\n') {
+		/* handle multi-line matching */
+		lineend_type rettype = from_line->return_type;
+		if ((rettype == lineend_NUL && patternchar == '\0')
+		 || (rettype != lineend_NONE && rettype != lineend_NUL &&
+			  (patternchar == '\n'
+			|| (patternchar == '\r' && rettype == lineend_CRLF)
+			|| (patternchar == 'R' && rettype == lineend_CR)
+			|| (patternchar == 'n' && rettype == lineend_LF)
+			  )
+		    )
+		   )
+		{
+			if (from_line->next == tail) {
+				if (! (* expression & DONE)) {
+					trace_undent ();
+					return NO_MATCH;
+				}
+				/* matching at tail */
+				/* however, we would need to 
+				   forward from_line and string 
+				   to undefined values;
+				   thus not matching below...
+				 */
+			} else {
+				ret = check_string (program, 
+					from_line->next->text, 
+					from_line->next, expression);
+				trace_undent ();
+				return ret;
+			}
+		} else {
+			trace_undent ();
+			return NO_MATCH;
+		}
+	    } else {
+		trace_undent ();
+		return NO_MATCH;
+	    }
+	} else
+	switch (opcode & ~EOLN) {
 	case NORMAL :
 		/* this case only applies to single-byte characters 
 		   as multi-byte characters in the search pattern are 
@@ -1578,37 +1633,12 @@ check_string (program, string, from_line, expression)
 		/*	advance_char (& string);	*/
 		/*	string ++;	???	*/
 		} else {
-			if (! (* string == patternchar
-			       || (* string == '\n' && patternchar == '\r' && from_line->return_type == lineend_CRLF)
-			      )) {
-				trace_undent ();
-				return NO_MATCH;
-			}
-			if (patternchar == '\n' || patternchar == '\r') {
-			    /* handle multi-line matching */
-			    if (from_line->return_type == lineend_NONE
-				|| from_line->return_type == lineend_NUL
-				)
-			    {
-				trace_undent ();
-				return NO_MATCH;
-			    } else {
-				if (from_line->next == tail) {
-					if (! (* expression & DONE)) {
-						trace_undent ();
-						return NO_MATCH;
-					}
-				} else {
-					ret = check_string (program, 
-						from_line->next->text, 
-						from_line->next, expression);
-					trace_undent ();
-					return ret;
-				}
-			    }
-			} else {
+			if (* string == patternchar) {
 			/*	advance_char (& string);	*/
 				string ++;
+			} else {
+				trace_undent ();
+				return NO_MATCH;
 			}
 		}
 		break;
@@ -1686,10 +1716,12 @@ check_string (program, string, from_line, expression)
 		expression += exprlist_length - 1;
 		break;
 	default :
-		/* panic ("Corrupted search program", NIL_PTR); */
+		program->status = REG_ERROR;
 		ring_bell ();
+		/*
 		error ("Corrupted search program");
 		sleep (2);
+		*/
 		trace_undent ();
 		return NO_MATCH;
 	}
@@ -1707,6 +1739,8 @@ check_string (program, string, from_line, expression)
 	 * whether a '$' was given at the end of the expression, or
 	 * the match was found on a null string. (E.g. [a-z]* always
 	 * matches) unless a ^ or $ was included in the pattern.
+	   See also comments above about matching \n or \r on last line 
+	   which fails here (2nd check) as replacement would not work.
 	 */
 	if ((* expression & EOLN) && * string != '\n' && * string != '\0') {
 		trace_undent ();
