@@ -64,12 +64,12 @@ check (char tag)
 		if (valid_cjk (cjkchar, NIL_PTR)) {
 			ok ++;
 		} else {
-			printf ("%c %04X\n", text_encoding_tag, cjkchar);
+			printf ("%c %04lX\n", text_encoding_tag, cjkchar);
 			non ++;
 		}
 		cjk_table_poi ++;
 	}
-	printf ("%c: valid %d, not valid %d\n", text_encoding_tag, ok, non);
+	printf ("%c: valid %ld, not valid %ld\n", text_encoding_tag, ok, non);
 	return non;
 }
 
